/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <ctype.h>

#include "PsxCommon.h"
#include "CDVDiso.h"

#ifdef __VCNET2005__
#pragma warning(disable:4996)//deprecate api annoying messages in vsnet2005beta1
#endif
#define btoi(b)		((b)/16*10 + (b)%16)		/* BCD to u_char */
#define itob(i)		((i)/10*16 + (i)%10)		/* u_char to BCD */

char *nCmdName[0x100]= {
	"CdSync", "CdNop", "CdStandby", "CdStop",
	"CdPause", "CdSeek", "CdRead", "CdReadAudio",
	"DvdRead", "CdGetToc", "", "",
	"CdReadKey", "", "", "",
};

char *sCmdName[0x100]= {
	"", "CdGetDiscType", "CdReadSubQ", "sceCdGetMecaconVersion",
	"", "sceCdTrayReqState", "sceCdTrayReq", "",
	"sceCdReadClock", "sceCdWriteRTC", "sceCdReadNVM", "sceCdWriteNVM",
	"sceCdSetHDMode", "", "", "sceCdPowerOff",
	"", "", "sceCdRI", "sceCdWriteILinkID", /*10*/
	"AudioDigitalOut", "ForbidDVDP", "AutoAdjustCtrl", "sceCdRM",
	"WriteILinkMN", "sceCdForbidRead", "sceCdBootCertify", "sceCdCancelPOffRdy",
	"sceCdBlueLEDCtl", "", "sceRemote2Read", "sceRemote2_7",
	"sceRemote2_6", "sceCdWriteWakeUpTime", "sceCdReadWakeUpTime", "sceCdRcBypassCtl", /*20*/
	"", "", "", "",
	"sceCdNoticeGameStart", "sceCdXBSPowerCtl", "sceCdXLEDCtl", "sceCdBuzzerCtl",
	"", "", "", "",
	"", "sceCdSetMediumRemoval", "sceCdGetMediumRemoval", "sceCdXDVRPReset", /*30*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"CdOpenConfig", "CdReadConfig", "CdWriteConfig", "CdCloseConfig", /*40*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"", "", "", "", /*50*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"", "", "", "", /*60*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"", "", "", "", /*70*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"mechacon_auth_0x80", "mechacon_auth_0x81", "mechacon_auth_0x82", "mechacon_auth_0x83", /*80*/
	"", "", "", "",
	"", "", "", "",
	"", "", "", "mechacon_auth_0x8F",
	"_write_HD_start", "", "", "", /*90*/
	"", "", "", "",
};

#define cdvdReadTime (((PSXCLK / 75) / BIAS) / cdvd.Speed)

#define CDVDREAD_INT(eCycle) PSX_INT(19, eCycle)


/* is cdvd.Status only for NCMDS? (linuzappz) */
enum cdvdStatus {
	CDVD_STATUS_NONE            = 0x00, // not sure ;)
	CDVD_STATUS_SEEK_COMPLETE   = 0x0A,
};


FILE *_cdvdOpenNVM() {
	struct stat buf;
	char file[256];
	char Bios[256];
	FILE *f;
	char *ptr;
	int i;

	strcpy(Bios, Config.BiosDir);
	strcat(Bios, Config.Bios);
	
	sprintf(file, "%s", Bios);
	ptr = file; i = strlen(file);
	while (i > 0) { if (ptr[i] == '.') break; i--; }
	ptr[i+1] = 0;
	strcat(file, "NVM");
	if (stat(file, &buf) == -1) {
		int i;

		SysPrintf("NVM File Not Found , Creating Blank File\n");
		f = fopen(file, "wb");
		if (f == NULL) {
			SysMessage("_cdvdOpenNVM: Error creating %s", file);
			exit(1);
		}

		for (i=0; i<1024; i++) fputc(0, f);
		return f;
	}
	f = fopen(file, "r+b");
	return f;
}

void cdvdReadNVM(u8 *dst, int offset, int bytes) {
	FILE *f;

	f = _cdvdOpenNVM();
	if (f == NULL) return;
	fseek(f, offset, SEEK_SET);
	fread(dst, 1, bytes, f);

	fclose(f);
}

void cdvdWriteNVM(u8 *src, int offset, int bytes) {
	FILE *f;

	f = _cdvdOpenNVM();
	if (f == NULL) return;
	fseek(f, offset, SEEK_SET);
	fwrite(src, 1, bytes, f);

	fclose(f);
}

int  cdvdInterrupt() {
	return 1;
}

int cdvdReadSector() {
	u32 bcr;

#ifdef CDR_LOG
	CDR_LOG("SECTOR %d (BCR %x;%x)\n", cdvd.Sector, HW_DMA3_BCR_H16, HW_DMA3_BCR_L16);
#endif
	bcr = HW_DMA3_BCR_H16 * HW_DMA3_BCR_L16*4;
	if (bcr < cdvd.BlockSize) {
//		SysPrintf("*PCSX2*: cdvdReadSector: bcr < cdvd.BlockSize; %x < %x\n", bcr, cdvd.BlockSize);
		if (HW_DMA3_CHCR & 0x01000000) {
			cdvd.Ready&= ~0x80;
			HW_DMA3_CHCR &= ~0x01000000;
			psxDmaInterrupt(3);
		}
		return -1;
	}

	if (cdvd.BlockSize == 2064) {
		u32 lsn = cdvd.Sector + 0x30000;
		PSXMu8(HW_DMA3_MADR+0) = (u8)(lsn >> 24);
		PSXMu8(HW_DMA3_MADR+1) = (u8)(lsn >> 16);
		PSXMu8(HW_DMA3_MADR+2) = (u8)(lsn >>  8);
		PSXMu8(HW_DMA3_MADR+3) = (u8)(lsn      );
		memcpy(PSXM(HW_DMA3_MADR+12), cdr.pTransfer, 2048);
	} else {
		memcpy(PSXM(HW_DMA3_MADR), cdr.pTransfer, cdvd.BlockSize);
	}
//	SysPrintf("sector %x;%x;%x\n", PSXMu8(HW_DMA3_MADR+0), PSXMu8(HW_DMA3_MADR+1), PSXMu8(HW_DMA3_MADR+2));

	HW_DMA3_BCR_H16-= (cdvd.BlockSize / (HW_DMA3_BCR_L16*4));
	HW_DMA3_MADR+= cdvd.BlockSize;

	return 0;
}

int  cdvdReadInterrupt() {

//	SysPrintf("cdvdReadInterrupt\n");
	if (cdvd.Readed == 0) {
		cdvd.RetryCntP = 0;
		cdvd.Reading = 1;
		cdvd.RErr = CDVDreadTrack(cdvd.Sector, cdvd.ReadMode);
		cdvd.Readed = 1;
		cdvd.Status = CDVD_STATUS_SEEK_COMPLETE;

		CDVDREAD_INT(cdvdReadTime);
		return 0;
	}

	if (cdvd.Reading == 1) {
		if (cdvd.RErr == 0) {
			cdr.pTransfer = CDVDgetBuffer();
		} else cdr.pTransfer = NULL;
		if (cdr.pTransfer == NULL) {
			cdvd.RetryCntP++;
			SysPrintf("READ ERROR %d\n", cdvd.Sector);
			if (cdvd.RetryCntP <= cdvd.RetryCnt) {
				cdvd.RErr = CDVDreadTrack(cdvd.Sector, cdvd.ReadMode);
				CDVDREAD_INT(cdvdReadTime);
				return 0;
			}
		}
		cdvd.Reading = 0;
	}
	if (cdvdReadSector() == -1) {
		CDVDREAD_INT(cdvdReadTime);
		return 0;
	}

	cdvd.Sector++;

	if (--cdvd.nSectors <= 0) {
		psxHu32(0x1070)|= 0x4;
		cdvd.Ready&= ~0x80;
		HW_DMA3_CHCR &= ~0x01000000;
		psxDmaInterrupt(3);
		return 1;
	}

	cdvd.RetryCntP = 0;
	cdvd.Reading = 1;
	cdr.RErr = CDVDreadTrack(cdvd.Sector, cdvd.ReadMode);
	CDVDREAD_INT(cdvdReadTime);
	return 0;
}

void cdvdVsync() {
	cdvd.RTCcount++;
	if (cdvd.RTCcount < 60) return;
	cdvd.RTCcount = 0;

	cdvd.RTC.second++;
	if (cdvd.RTC.second < 60) return;
	cdvd.RTC.second = 0;

	cdvd.RTC.minute++;
	if (cdvd.RTC.minute < 60) return;
	cdvd.RTC.minute = 0;

	cdvd.RTC.hour++;
	if (cdvd.RTC.hour < 24) return;
	cdvd.RTC.hour = 0;

	cdvd.RTC.day++;
	if (cdvd.RTC.day < 30) return;
	cdvd.RTC.day = 0;

	cdvd.RTC.month++;
	if (cdvd.RTC.month < 12) return;
	cdvd.RTC.month = 0;

	cdvd.RTC.year++;
}

#define SetResultSize(size) \
	cdvd.ResultC = size; cdvd.ResultP = 0; \
	cdvd.sDataIn&=~0x40;


u8   cdvdRead04(void) { // NCOMMAND
#ifdef CDR_LOG
	CDR_LOG("cdvdRead04 %x\n", cdvd.nCommand);
#endif
	return cdvd.nCommand;
}

u8   cdvdRead05(void) { // READY
#ifdef CDR_LOG
	CDR_LOG("cdvdRead05 %x\n", cdvd.Ready);
#endif
	return cdvd.Ready;
}

u8   cdvdRead06(void) { // ERROR
#ifdef CDR_LOG
	CDR_LOG("cdvdRead06 %x\n", 0);
#endif
	return 0;
}

u8   cdvdRead07(void) { // BREAK
#ifdef CDR_LOG
	CDR_LOG("cdvdRead07 %x\n", 0);
#endif
	return 0;
}

u8   cdvdRead08(void) { // PWOFF
#ifdef CDR_LOG
	CDR_LOG("cdvdRead08 %x\n", cdvd.PwOff);
#endif
	return cdvd.PwOff;
}

u8   cdvdRead0A(void) { // STATUS
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0A %x\n", cdvd.Status);
#endif
	return cdvd.Status;
}

static u8 ltray;

u8   cdvdRead0B(void) { // TRAY
	u8 tray;
	s32 ret;

	if (cdCaseopen) {
		tray = CDVD_TRAY_OPEN;
	} else {
		ret = CDVDgetTrayStatus();
		if (ret == -1) tray = 0;
		else tray = (u8)ret;
	}
	ret   = tray;
	tray  = ltray;
	ltray = (u8)ret;
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0B %x\n", tray);
#endif
	return tray;
}

u8   cdvdRead0C(void) { // CRT MINUTE
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0C %x\n", itob((u8)(cdvd.Sector/(60*75))));
#endif
	return itob((u8)(cdvd.Sector/(60*75)));
}

u8   cdvdRead0D(void) { // CRT SECOND
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0D %x\n", itob((u8)((cdvd.Sector/75)%60)+2));
#endif
	return itob((u8)((cdvd.Sector/75)%60)+2);
}

u8   cdvdRead0E(void) { // CRT FRAME
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0E %x\n", itob((u8)(cdvd.Sector%75)));
#endif
	return itob((u8)(cdvd.Sector%75));
}

u8   cdvdRead0F(void) { // TYPE
	u8 type;
	s32 ret;

	if (cdCaseopen) {
		type = CDVD_TYPE_DETCT;
		ltray = CDVD_TRAY_OPEN;
	} else {
		ret = CDVDgetType();
		if (ret == -1) type = 0;
		else type = (u8)ret;
	}
#ifdef CDR_LOG
	CDR_LOG("cdvdRead0F %x\n", type);
#endif
	return type;
}

u8   cdvdRead13(void) { // UNKNOWN
#ifdef CDR_LOG
	CDR_LOG("cdvdRead13 %x\n", 4);
#endif
	return 4;
}

u8   cdvdRead16(void) { // SCOMMAND
#ifdef CDR_LOG
	CDR_LOG("cdvdRead16 %x\n", cdvd.sCommand);
#endif
	return cdvd.sCommand;
}

u8   cdvdRead17(void) { // SDATAIN
#ifdef CDR_LOG
	CDR_LOG("cdvdRead17 %x\n", cdvd.sDataIn);
#endif
	return cdvd.sDataIn;
}

u8   cdvdRead18(void) { // SDATAOUT
	u8 ret=0;

	if ((cdvd.sDataIn & 0x40) == 0) {
		if (cdvd.ResultP < cdvd.ResultC) {
			cdvd.ResultP++;
			if (cdvd.ResultP >= cdvd.ResultC) cdvd.sDataIn|= 0x40;
			ret = cdvd.Result[cdvd.ResultP-1];
		}
	}

#ifdef CDR_LOG
	CDR_LOG("cdvdRead18 %x (ResultC=%d, ResultP=%d)\n", ret, cdvd.ResultC, cdvd.ResultP);
#endif

	return ret;
}

u8   cdvdRead20(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead20 %x\n", cdvd.Key[0]);
#endif
	return cdvd.Key[0];
}

u8   cdvdRead21(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead21 %x\n", cdvd.Key[1]);
#endif
	return cdvd.Key[1];
}

u8   cdvdRead22(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead22 %x\n", cdvd.Key[2]);
#endif
	return cdvd.Key[2];
}

u8   cdvdRead23(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead23 %x\n", cdvd.Key[3]);
#endif
	return cdvd.Key[3];
}

u8   cdvdRead24(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead24 %x\n", cdvd.Key[4]);
#endif
	return cdvd.Key[4];
}

u8   cdvdRead28(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead28 %x\n", cdvd.Key[5]);
#endif
	return cdvd.Key[5];
}

u8   cdvdRead29(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead29 %x\n", cdvd.Key[6]);
#endif
	return cdvd.Key[6];
}

u8   cdvdRead2A(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead2A %x\n", cdvd.Key[7]);
#endif
	return cdvd.Key[7];
}

u8   cdvdRead2B(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead2B %x\n", cdvd.Key[8]);
#endif
	return cdvd.Key[8];
}

u8   cdvdRead2C(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead2C %x\n", cdvd.Key[9]);
#endif
	return cdvd.Key[9];
}

u8   cdvdRead30(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead30 %x\n", cdvd.Key[10]);
#endif
	return cdvd.Key[10];
}

u8   cdvdRead31(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead31 %x\n", cdvd.Key[11]);
#endif
	return cdvd.Key[11];
}

u8   cdvdRead32(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead32 %x\n", cdvd.Key[12]);
#endif
	return cdvd.Key[12];
}

u8   cdvdRead33(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead33 %x\n", cdvd.Key[13]);
#endif
	return cdvd.Key[13];
}

u8   cdvdRead34(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead34 %x\n", cdvd.Key[14]);
#endif
	return cdvd.Key[14];
}

u8   cdvdRead38(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead38 %x\n", 0x7);
#endif
	return 0x7;
}

u8   cdvdRead39(void) {
#ifdef CDR_LOG
	CDR_LOG("cdvdRead39 %x\n", 0);
#endif
	return 0;
}

void cdvdReadKey() {
	char str[256];
	char *ptr = str;
	char buf[8];
	int code;

	memset(cdvd.Key, 0, sizeof(cdvd.Key));

	GetPS2ElfName(str);

	while (ptr != str + 255) {
		if (*ptr++ == ':') break;
	}
	if (ptr == str + 255) return;
	if (*ptr == '\\') ptr++;

	// SCES_500.00;1
	cdvd.Key[0] = ptr[3] << 7;
	cdvd.Key[1] = ptr[3] >> 1;
	cdvd.Key[1]|= ptr[2] << 6;
	cdvd.Key[2] = ptr[2] >> 2;
	cdvd.Key[2]|= ptr[1] << 5;
	cdvd.Key[3] = ptr[1] >> 3;
	cdvd.Key[3]|= ptr[0] << 4;
	cdvd.Key[4] = ptr[0] >> 4;

	buf[0] = ptr[5];  buf[1] = ptr[6];
	buf[2] = ptr[7];  buf[3] = ptr[9];
	buf[4] = ptr[10]; buf[5] = 0;
	if (isdigit(buf[0])) {
		code = strtol(buf, (char**)NULL, 0);

		cdvd.Key[4]|=  code << 3;
		cdvd.Key[14]= (code >> 2) & 0xf8;
		cdvd.Key[0]|= (code >> 10) & 0x7f;
	}
	psxHu32(0x1070)|= 0x4;
}

void cdvdWrite04(u8 rt) { // NCOMMAND
	cdvdTN	diskInfo;
	cdvdTD	trackInfo;
	int		i, err;
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite04: NCMD %s (%x) (ParamP = %x)\n", nCmdName[rt], rt, cdvd.ParamP);
#endif
	cdvd.nCommand = rt;
	cdvd.Status = CDVD_STATUS_NONE;
	switch (rt) {
		case 0x00: // CdNop
		case 0x01: // CdNop_
			psxHu32(0x1070)|= 0x4;
			break;
		case 0x05: // CdSeek
			psxHu32(0x1070)|= 0x4;
			break;

		case 0x06: // CdRead
			cdvd.Sector = *(int*)(cdvd.Param);
			cdvd.nSectors = *(int*)(cdvd.Param+4);
			if (cdvd.Param[8] == 0) cdvd.RetryCnt = 0x100;
			else cdvd.RetryCnt = cdvd.Param[8];
			cdvd.SpindlCtrl = cdvd.Param[9];
			switch (cdvd.Param[9]) {
				case 0x01: cdvd.Speed =  1; break;
				case 0x02: cdvd.Speed =  2; break;
				case 0x03: cdvd.Speed =  4; break;
				case 0x04: cdvd.Speed = 12; break;
				default:   cdvd.Speed = 24; break;
			}
			switch (cdvd.Param[10]) {
				case 2: cdvd.ReadMode = CDVD_MODE_2340; cdvd.BlockSize = 2340; break;
				case 1: cdvd.ReadMode = CDVD_MODE_2328; cdvd.BlockSize = 2328; break;
				case 0: cdvd.ReadMode = CDVD_MODE_2048; cdvd.BlockSize = 2048; break;
			}
#ifdef CDR_LOG
//			CDR_LOG("CdRead: %d, nSectors=%d, RetryCnt=%x, Speed=%x(%x), ReadMode=%x(%x) (1074=%x)\n", cdvd.Sector, cdvd.nSectors, cdvd.RetryCnt, cdvd.Speed, cdvd.Param[9], cdvd.ReadMode, cdvd.Param[10], psxHu32(0x1074));
#endif
			SysPrintf("CdRead: %d, nSectors=%d, RetryCnt=%x, Speed=%x(%x), ReadMode=%x(%x) (1074=%x)\n", cdvd.Sector, cdvd.nSectors, cdvd.RetryCnt, cdvd.Speed, cdvd.Param[9], cdvd.ReadMode, cdvd.Param[10], psxHu32(0x1074));

			cdvd.Readed = 0;
			cdvd.Ready|= 0x80;
			CDVDREAD_INT(0x400)
			break;

		case 0x07: // CdReadAudio
			cdvd.Sector = *(int*)(cdvd.Param);
			cdvd.nSectors = *(int*)(cdvd.Param+4);
			if (cdvd.Param[8] == 0) cdvd.RetryCnt = 0x100;
			else cdvd.RetryCnt = cdvd.Param[8];
			cdvd.SpindlCtrl = cdvd.Param[9];
			switch (cdvd.Param[9]) {
				case 0x01: cdvd.Speed =  1; break;
				case 0x02: cdvd.Speed =  2; break;
				case 0x03: cdvd.Speed =  4; break;
				case 0x04: cdvd.Speed = 12; break;
				default:   cdvd.Speed = 24; break;
			}
			switch (cdvd.Param[10]) {
				case 1: cdvd.ReadMode = CDVD_MODE_2368; cdvd.BlockSize = 2368; break;
				case 2:
				case 0: cdvd.ReadMode = CDVD_MODE_2352; cdvd.BlockSize = 2352; break;
			}
			SysPrintf("CdReadAudio: %d, nSectors=%d, RetryCnt=%x, Speed=%x(%x), ReadMode=%x(%x) (1074=%x)\n", cdvd.Sector, cdvd.nSectors, cdvd.RetryCnt, cdvd.Speed, cdvd.Param[9], cdvd.ReadMode, cdvd.Param[10], psxHu32(0x1074));

			cdvd.Readed = 0;
			cdvd.Ready|= 0x80;
//			if (psxRegs.cycle < 0x009d4500) {
			CDVDREAD_INT(0x400)//}
			break;

		case 0x08: // DvdRead
			cdvd.Sector = *(int*)(cdvd.Param);
			cdvd.nSectors = *(int*)(cdvd.Param+4);
			if (cdvd.Param[8] == 0) cdvd.RetryCnt = 0x100;
			else cdvd.RetryCnt = cdvd.Param[8];
			cdvd.SpindlCtrl = cdvd.Param[9];
/*			if (cdvd.Param[9] < 4) cdvd.Speed = cdvd.Param[9];
			else*/ cdvd.Speed = 4;
			cdvd.ReadMode = CDVD_MODE_2048; cdvd.BlockSize = 2064;
#ifdef CDR_LOG
//			CDR_LOG("DvdRead: %d, nSectors=%d, RetryCnt=%x, Speed=%x(%x), ReadMode=%x(%x) (1074=%x)\n", cdvd.Sector, cdvd.nSectors, cdvd.RetryCnt, cdvd.Speed, cdvd.Param[9], cdvd.ReadMode, cdvd.Param[10], psxHu32(0x1074));
#endif
			SysPrintf("DvdRead: %d, nSectors=%d, RetryCnt=%x, Speed=%x(%x), ReadMode=%x(%x) (1074=%x)\n", cdvd.Sector, cdvd.nSectors, cdvd.RetryCnt, cdvd.Speed, cdvd.Param[9], cdvd.ReadMode, cdvd.Param[10], psxHu32(0x1074));
			cdvd.Readed = 0;
			cdvd.Ready|= 0x80;
			CDVDREAD_INT(0x400)
			break;

		case 0x09: // CdGetToc & cdvdman_call19
			//Param[0] is 0 for CdGetToc and any value for cdvdman_call19
			//the code below handles only CdGetToc!
			if (CDVDgetTN(&diskInfo) == -1)	{diskInfo.etrack = 0;diskInfo.strack = 1;}
			PSXMu8(HW_DMA3_MADR+ 2) = 0xA0;
			PSXMu8(HW_DMA3_MADR+ 7) = itob(diskInfo.strack);//Number of FirstTrack
			
			PSXMu8(HW_DMA3_MADR+12) = 0xA1;
			PSXMu8(HW_DMA3_MADR+17) = itob(diskInfo.etrack);//Number of LastTrack

			if (CDVDgetTD(0, &trackInfo) == -1) trackInfo.minute = trackInfo.second = 0;
			PSXMu8(HW_DMA3_MADR+22) = 0xA2;//DiskLength
			PSXMu8(HW_DMA3_MADR+27) = itob(trackInfo.minute);
			PSXMu8(HW_DMA3_MADR+28) = itob(trackInfo.second);

			for (i=diskInfo.strack; i<=diskInfo.etrack; i++){
				err=CDVDgetTD(i, &trackInfo);
				PSXMu8(HW_DMA3_MADR+i*10+30) = trackInfo.type;
				PSXMu8(HW_DMA3_MADR+i*10+32) = err == -1 ? 0 : itob(i);	  //number
				PSXMu8(HW_DMA3_MADR+i*10+37) = itob(trackInfo.minute+0);
				PSXMu8(HW_DMA3_MADR+i*10+38) = itob(trackInfo.second+2);
				PSXMu8(HW_DMA3_MADR+i*10+39) = itob(trackInfo.frame +0);
			}
			psxHu32(0x1070)|= 0x4;
			break;

		case 0x0C: // CdReadKey
			SysPrintf("CDVDReadKey\n");
			cdvdReadKey();
			break;

		default:
			SysPrintf("NCMD Unknown %x\n", rt);
			psxHu32(0x1070)|= 0x4;
			break;
	}
	cdvd.ParamP = 0; cdvd.ParamC = 0;
}

void cdvdWrite05(u8 rt) { // NDATAIN
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite05 %x\n", rt);
#endif
	if (cdvd.ParamP < 32) {
		cdvd.Param[cdvd.ParamP++] = rt;
		cdvd.ParamC++;
	}
}

void cdvdWrite06(u8 rt) { // HOWTO
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite06 %x\n", rt);
#endif
	cdvd.HowTo = rt;
}

void cdvdWrite07(u8 rt) { // BREAK
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite07 %x\n", rt);
#endif
	SysPrintf("*PCSX2*: CDVD BREAK\n");
}

void cdvdWrite08(u8 rt) { // PWOFF
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite08 %x\n", rt);
#endif
}

void cdvdWrite0A(u8 rt) { // STATUS
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite0A %x\n", rt);
#endif
}

void cdvdWrite0F(u8 rt) { // TYPE
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite0F %x\n", rt);
#endif
}

void cdvdWrite16(u8 rt) { // SCOMMAND
	cdvdTN	diskInfo;
	cdvdTD	trackInfo;
	int i, lbn, type, min, sec, frm;

#ifdef CDR_LOG
	CDR_LOG("cdvdWrite16: SCMD %s (%x) (ParamP = %x)\n", sCmdName[rt], rt, cdvd.ParamP);
#endif
	cdvd.sCommand = rt;
	switch (rt) {

//		case 0x01: // GetDiscType - from cdvdman (0:1)
//			SetResultSize(1);
//			cdvd.Result[0] = 0;
//			break;

		case 0x02: // CdReadSubQ  (0:11)
			SetResultSize(11);//all BCD coded!
			cdvd.Result[ 0] = 0;//does not matter?

			if (CDVDgetTN(&diskInfo) == -1){	diskInfo.etrack = 0;diskInfo.strack = 1;}

			for (lbn=0, type=0, i=diskInfo.etrack; i>=diskInfo.strack; i--)
				if (CDVDgetTD(i, &trackInfo) != -1){
					type=trackInfo.type;
					lbn=(trackInfo.minute*60+trackInfo.second)*75+trackInfo.frame;
					if ((u32)lbn<cdvd.Sector)break;
				}
			sec=(cdvd.Sector-lbn)/60;frm=(cdvd.Sector-lbn)%60;min=sec/60;sec=sec%60;
			cdvd.Result[ 1] = type;		// & 0xF == 1		track type
			cdvd.Result[ 2] = itob(i);	// < 0x9A; != 0		track number
			cdvd.Result[ 3] = itob(i);//bcd?// < 0x9A; <0x99cue index
			cdvd.Result[ 4] = itob(min);// < 0x9A;			track minute
			cdvd.Result[ 5] = itob(sec);// < 0x9A; < 60;	track second
			cdvd.Result[ 6] = itob(frm);// < 0x9A; < 75;	track frame
			cdvd.Result[ 7] = 0;		// == 0				[no?] error
			sec=cdvd.Sector/60;frm=cdvd.Sector%60;min=sec/60;sec=sec%60;
			cdvd.Result[ 8] = itob(min);// < 0x9A;			disk minute (same as CDVDreg_CRTMIN?)
			cdvd.Result[ 9] = itob(sec+2);// < 0x9A; < 60;	disk second (same as CDVDreg_CRTSEC?)
			cdvd.Result[10] = itob(frm);// < 0x9A; < 75;	disk frame  (same as CDVDreg_CRTFRAME?)

			break;

		case 0x03: // CdGetMecaconVersion (1:4)
			if(cdvd.Param[0]==0x45) {
				SetResultSize(9);
				cdvd.Result[0] = 0x00;
				cdvdReadNVM(&cdvd.Result[1], 0x1c8, 8);
			} else {
				//SysPrintf("CdGetMecaconVersion param[0]=%x\n", cdvd.Param[0]);
				SetResultSize(4);//incorrect, depends on the input params!
				cdvd.Result[0]=0x00;//i have no clue:p
				cdvd.Result[1]=0xff;//................
				cdvd.Result[2]=0x02;//................
				cdvd.Result[3]=0x03;//................
			}
			break;

		case 0x05: // CdTrayReqState  (0:1)- Returns State of Tray from the previous CdTrayReq
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x06: // CdTrayReq  (1:1)
			SetResultSize(1);
			cdvd.Result[0] = 1;
			break;

		case 0x08: // CdReadRTC (0:8)
			SetResultSize(8);
#ifdef WIN32
			{static SYSTEMTIME st;
			GetSystemTime(&st);    //UCT time instead of local time
			cdvd.RTC.second	=itob((u8)st.wSecond);
			cdvd.RTC.minute	=itob((u8)st.wMinute);
			cdvd.RTC.hour	=itob((u8)st.wHour+8); // add 8 to the hours (the ps2 wants it this way)
			cdvd.RTC.day	=itob((u8)st.wDay);
			cdvd.RTC.month	=itob((u8)st.wMonth);
			cdvd.RTC.year	=itob((u8)(st.wYear-2000));
			}
#endif
			memcpy(cdvd.Result, (u8*)&cdvd.RTC, 8);
			break;


		case 0x09: // sceCdWriteRTC (7:1)
			SetResultSize(1);
			cdvd.Result[0] = 0;
			memcpy((u8*)&cdvd.RTC, cdvd.Param, 7);
			break;

		case 0x0A: // sceCdReadNVM (2:3)
			if (cdvd.Param[1] < 511) {
				SetResultSize(3);
				cdvd.Result[0] = 0;
				cdvdReadNVM(&cdvd.Result[1], cdvd.Param[1]*2, 2);
			} else {
				SetResultSize(1);
				cdvd.Result[0] = 0xff;
			}
			break;

		case 0x0B: // sceCdWriteNVM (4:1)
			SetResultSize(1);
			if (cdvd.Param[1] < 511) {
				cdvd.Result[0] = 0;
				cdvdWriteNVM(&cdvd.Param[2], cdvd.Param[1]*2, 2);
			} else {
				cdvd.Result[0] = 0xff;
			}
			break;

//		case 0x0C: // sceCdSetHDMode (1,1) 
//			break;

		case 0x0F: // sceCdPowerOff (0:1)- Call74 from Xcdvdman 
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x12: // sceCdRI (0:9)
			SetResultSize(9);
			cdvd.Result[0] = 0x00;
			if (BiosVersion  > 345 ) 
				cdvdReadNVM(&cdvd.Result[1], 0x1e0, 8);
			else
				cdvdReadNVM(&cdvd.Result[1], 0x1C0, 8);
			break;

		case 0x13: // sceCdWriteILinkID (8:1)
			if (BiosVersion  > 345 ) 
				cdvdWriteNVM(&cdvd.Result[1], 0x1e0, 8);
			else
				cdvdWriteNVM(&cdvd.Result[1], 0x1C0, 8);
			break;

		case 0x14: // CdCtrlAudioDigitalOut (1:1)
			//parameter can be 2, 0, ...
			SetResultSize(1);
			cdvd.Result[0] = 0;//8 is a flag; not used
			break;

		case 0x15: // sceCdForbidDVDP (0:1) 
			//SysPrintf("sceCdForbidDVDP\n");
			SetResultSize(1);
			cdvd.Result[0] = 5;
			break;

		case 0x16: // AutoAdjustCtrl - from cdvdman (1:1)
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x17: // CdRM (1:9) - from xcdvdman
				SetResultSize(0x9);
				cdvd.Result[0] = 0;
				if (BiosVersion  > 345 ) 
					cdvdReadNVM(&cdvd.Result[1], 0x1b0+cdvd.Param[0], 8);
				else
					cdvdReadNVM(&cdvd.Result[1], 0x1a0+cdvd.Param[0], 8);


			break;

		case 0x18: // _WriteILinkMN (9:1) - from xcdvdman
				SetResultSize(0x9);
				cdvd.Result[0] = 0;
				if (BiosVersion  > 345 ) 
					cdvdWriteNVM(&cdvd.Result[1], 0x1b0+cdvd.Param[0], 8);
				else
					cdvdWriteNVM(&cdvd.Result[1], 0x1a0+cdvd.Param[0], 8);
			break;

//		case 0x19: // sceCdForbidRead (0:1) - from xcdvdman
//			break;

		case 0x1A: // sceCdBootCertify (4:1)
			SetResultSize(1);//on imput there are 4 bytes: 1;?10;J;C for 18000; 1;60;E;C for 39002 from ROMVER
			cdvd.Result[0]=1;//i guess that means okay
			break;

		case 0x1B: // sceCdCancelPOffRdy (0:1) - Call73 from Xcdvdman (1:1)
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x1C: // sceCdBlueLEDCtl (1:1) - Call72 from Xcdvdman
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

//		case 0x1D: // cdvdman_call116 (0:5) - In V10 Bios 
//			break;

		case 0x1E: // sceRemote2Read (0:5) - // 00 14 AA BB CC -> remote key code
			SetResultSize(5);
			cdvd.Result[0] = 0x00;
			cdvd.Result[1] = 0x14;
			cdvd.Result[2] = 0x00;
			cdvd.Result[3] = 0x00;
			cdvd.Result[4] = 0x00;
			break;

//		case 0x1F: // sceRemote2_7 (2:1) - cdvdman_call117
//			break;

		case 0x20: // sceRemote2_6 (0:3)	// 00 01 00
			SetResultSize(3);
			cdvd.Result[0] = 0x00;
			cdvd.Result[1] = 0x01;
			cdvd.Result[2] = 0x00;
			break;

//		case 0x21: // sceCdWriteWakeUpTime (8:1) 
//			break;

		case 0x22: // sceCdReadWakeUpTime (0:10) 
			SetResultSize(10);
			cdvd.Result[0] = 0;
			cdvd.Result[1] = 0;
			cdvd.Result[2] = 0;
			cdvd.Result[3] = 0;
			cdvd.Result[4] = 0;
			cdvd.Result[5] = 0;
			cdvd.Result[6] = 0;
			cdvd.Result[7] = 0;
			cdvd.Result[8] = 0;
			cdvd.Result[9] = 0;
			break;

		case 0x24: // sceCdRCBypassCtrl (1:1) - In V10 Bios 
			SetResultSize(1);
			cdvd.Result[0] = 0;
		break;

//		case 0x25: // cdvdman_call120 (1:1) - In V10 Bios 
//			break;

//		case 0x26: // cdvdman_call128 (0,3) - In V10 Bios 
//			break;

//		case 0x27: // cdvdman_call148 (0:13) - In V10 Bios 
//			break;

//		case 0x28: // cdvdman_call150 (1:1) - In V10 Bios 
//			break;

//		case 0x29: //sceCdNoticeGameStart (1:1)
//			break;

//		case 0x2C: //sceCdXBSPowerCtl (2:2)
//			break;

//		case 0x2D: //sceCdXLEDCtl (2:2)
//			break;

//		case 0x2E: //sceCdBuzzerCtl (0:1)
//			break;

//		case 0x2F: //cdvdman_call167 (16:1)
//			break;

//		case 0x30: //cdvdman_call169 (1:9)
//			break;

//		case 0x31: //sceCdSetMediumRemoval (1:1)
//			break;

//		case 0x32: //sceCdGetMediumRemoval (0:2)
//			break;

//		case 0x33: //sceCdXDVRPReset (1:1)
//			break;

		case 0x36: //cdvdman_call189 (0:15)
			SetResultSize(15);
			cdvd.Result[0] = 0;
			cdvd.Result[1] = 0;
			cdvd.Result[2] = 0;
			cdvd.Result[3] = 0;
			cdvd.Result[4] = 0;
			cdvd.Result[5] = 0;
			cdvd.Result[6] = 0;
			cdvd.Result[7] = 0;
			cdvd.Result[8] = 0;
			cdvd.Result[9] = 0;
			cdvd.Result[10] = 0;
			cdvd.Result[11] = 0;
			cdvd.Result[12] = 0;
			cdvd.Result[13] = 0;
			cdvd.Result[14] = 0;
			break;

		case 0x40: // CdOpenConfig
			cdvd.CIndexA   = cdvd.Param[0];
			cdvd.CIndexB   = cdvd.Param[1];
			cdvd.CIndexBlk = cdvd.Param[2];
			cdvd.CIndex    = 0;
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x41: // CdReadConfig
			SetResultSize(16);
			if (cdvd.CIndexBlk) {
//				int i, j;

/*				for (i=0, j=0; i<15; i++){
					cdvd.Result[i] = cdvd.Config[cdvd.CIndexA][cdvd.CIndexB].data[cdvd.CIndex][i];
					j+= cdvd.Config[cdvd.CIndexA][cdvd.CIndexB].data[cdvd.CIndex][i];
				}
				cdvd.Result[15] = j;*/
				if (BiosVersion  > 345 ){ 
				cdvdReadNVM(&cdvd.Result[0], 0x270+cdvd.CIndexB*64+cdvd.CIndex*16, 16);
				} else{
					cdvdReadNVM(&cdvd.Result[0], 0x2c0+cdvd.CIndexB*64+cdvd.CIndex*16, 16);}

				cdvd.CIndex++;//<config_blk
			} else {
				memset(cdvd.Result, 0, 16);
			}
			break;

		case 0x42: // CdWriteConfig
			SetResultSize(1);
			if (cdvd.CIndexBlk) {
/*				int i, j;

				for (i=0, j=0; i<15; i++){
					cdvd.Config[cdvd.CIndexA][cdvd.CIndexB].data[cdvd.CIndex][i] = cdvd.Param[i];
					j+= cdvd.Config[cdvd.CIndexA][cdvd.CIndexB].data[cdvd.CIndex][i];
				}*/
//				cdvd.Result[15] = j;
				if (BiosVersion  > 345 ){ 
				cdvdWriteNVM(&cdvd.Param[0], 0x270+cdvd.CIndexB*64+cdvd.CIndex*16, 16);
				}else {
					cdvdWriteNVM(&cdvd.Param[0], 0x2c0+cdvd.CIndexB*64+cdvd.CIndex*16, 16);}
				cdvd.CIndex++;//<config_blk
			} else {
//				memset(cdvd.Result, 0, 16);
			}
			cdvd.Result[0] = 0;
//			memcpy(cdvd.Result, cdvd.Config[0][0], 16);
			break;

		case 0x43: // CdCloseConfig
			SetResultSize(1);
			cdvd.Result[0] = 0;
			break;

		case 0x80: // secrman: __mechacon_auth_0x80
			SetResultSize(1);//in:1
			cdvd.Result[0] = 0;
			break;

		case 0x81: // secrman: __mechacon_auth_0x81
			SetResultSize(1);//in:1
			cdvd.Result[0] = 0;
			break;

		case 0x82: // secrman: __mechacon_auth_0x82
			SetResultSize(1);//in:16
			cdvd.Result[0] = 0;
			break;

		case 0x83: // secrman: __mechacon_auth_0x83
			SetResultSize(1);//in:8
			cdvd.Result[0] = 0;
			break;

		case 0x84: // secrman: __mechacon_auth_0x84
			SetResultSize(1+8+4);//in:0
			cdvd.Result[0] = 0;
			
			cdvd.Result[1] = 0x21;
			cdvd.Result[2] = 0xdc;
			cdvd.Result[3] = 0x31;
			cdvd.Result[4] = 0x96;
			cdvd.Result[5] = 0xce;
			cdvd.Result[6] = 0x72;
			cdvd.Result[7] = 0xe0;
			cdvd.Result[8] = 0xc8;

			cdvd.Result[9]  = 0x69;
			cdvd.Result[10] = 0xda;
			cdvd.Result[11] = 0x34;
			cdvd.Result[12] = 0x9b;
			break;

		case 0x85: // secrman: __mechacon_auth_0x85
			SetResultSize(1+4+8);//in:0
			cdvd.Result[0] = 0;
			
			cdvd.Result[1] = 0xeb;
			cdvd.Result[2] = 0x01;
			cdvd.Result[3] = 0xc7;
			cdvd.Result[4] = 0xa9;

			cdvd.Result[ 5] = 0x3f;
			cdvd.Result[ 6] = 0x9c;
			cdvd.Result[ 7] = 0x5b;
			cdvd.Result[ 8] = 0x19;
			cdvd.Result[ 9] = 0x31;
			cdvd.Result[10] = 0xa0;
			cdvd.Result[11] = 0xb3;
			cdvd.Result[12] = 0xa3;
			break;

		case 0x86: // secrman: __mechacon_auth_0x86
			SetResultSize(1);//in:16
			cdvd.Result[0] = 0;
			break;

		case 0x87: // secrman: __mechacon_auth_0x87
			SetResultSize(1);//in:8
			cdvd.Result[0] = 0;
			break;

		case 0x8d: // ???
			SetResultSize(1);//in:0
			cdvd.Result[0] = 0; // 0 complete ; 1 busy ; 0x80 error
			break;

		case 0x88: // secrman: __mechacon_auth_0x88	//for now it is the same; so, fall;)
		case 0x8F: // secrman: __mechacon_auth_0x8F
			SetResultSize(1);//in:0
			cdvd.Result[0] = 0; // 0 complete ; 1 busy ; 0x80 error
			break;

		case 0x90: // _write_HD_start ???
			SetResultSize(1);//in:0
			cdvd.Result[0] = 0; // 0 complete ; 1 busy ; 0x80 error
			break;

		case 0x91: // ???
			SetResultSize(1);//in:0
			cdvd.Result[0] = 0; // 0 complete ; 1 busy ; 0x80 error
			break;

		default:
			SysPrintf("SCMD Unknown %x\n", rt);
			SetResultSize(1);//in:0
			cdvd.Result[0] = 0; // 0 complete ; 1 busy ; 0x80 error
			break;
	}
	cdvd.ParamP = 0; cdvd.ParamC = 0;
}

void cdvdWrite17(u8 rt) { // SDATAIN
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite17 %x\n", rt);
#endif
	if (cdvd.ParamP < 32) {
		cdvd.Param[cdvd.ParamP++] = rt;
		cdvd.ParamC++;
	}
}

void cdvdWrite18(u8 rt) { // SDATAOUT
#ifdef CDR_LOG
	CDR_LOG("cdvdWrite18 %x\n", rt);
#endif
}

void cdvdReset() {
	memset(&cdvd, 0, sizeof(cdvd));
	cdvd.sDataIn = 0x40;
	cdvd.Ready   = 0x4e;
	cdvd.Config[0][0].data[0][0] = 0x80;
	cdvd.Config[0][0].data[0][1] = 0xFF;
	cdvd.Config[0][0].data[0][2] = 0xFF;
	cdvd.Config[0][0].data[0][3] = 0xFF;
	cdvd.Config[0][1].data[0][0] = 0xFF;
	cdvd.Config[0][1].data[1][0] = 0x1D;
	cdCaseopen = 0;
}

int  cdvdFreeze(gzFile f, int Mode) {
	gzfreeze(&cdvd, sizeof(cdvd));
	if (Mode == FREEZE_LOAD) {
		if (cdvd.Reading) {
			cdvd.RErr = CDVDreadTrack(cdvd.Sector, cdvd.ReadMode);
		}
	}

	return 0;
}

