/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "Common.h"
#include "VU0.h"
#include "VUmicro.h"
/*****************************************/
/*          NEW FLAGS                    */ //By asadr. Thnkx F|RES :p
/*****************************************/

#define		COP2_FLOAT_MAX	2147483648.0
#define		COP2_FLOAT_MIN	-2147483648.0


__inline void vuUpdateDI(VURegs * VU) {
	u32 Flag_S = 0;
	u32 Flag_I = 0;
	u32 Flag_D = 0;

	/*
	FLAG D - I
	*/
	if ( (VU->VI[REG_MAC_FLAG].UL & 0x00F0 ) != 0 ) Flag_S = 1;
	Flag_I = (VU->VI[REG_STATUS_FLAG].UL >> 4) & 0x1;
	Flag_D = (VU->VI[REG_STATUS_FLAG].UL >> 5) & 0x1;

	VU->VI[REG_STATUS_FLAG].UL|= (Flag_I | Flag_S) << 10;
	VU->VI[REG_STATUS_FLAG].UL|= (Flag_D | Flag_S) << 11;
}

void VU_UPDATE_STAT_REG_DI0(VURegs * VU) {
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI1(VURegs * VU) {
	VU->VI[REG_STATUS_FLAG].UL|= 0x20; 
	VU->VI[REG_STATUS_FLAG].UL&=~0x10; 
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI2(VURegs * VU) {
	VU->VI[REG_STATUS_FLAG].UL|= 0x10; 
	VU->VI[REG_STATUS_FLAG].UL&=~0x20; 
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI3(VURegs * VU) {
	VU->VI[REG_STATUS_FLAG].UL&=~0x20; 
	VU->VI[REG_STATUS_FLAG].UL&=~0x10; 
	vuUpdateDI(VU);
}

float VU_MACx_UPDATE(VURegs * VU, double x) {
	if (x == 0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0008; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0008;
	if (x <  0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0080; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0080;
	if (x <  COP2_FLOAT_MIN) VU->VI[REG_MAC_FLAG].UL|= 0x0800;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0800;
	if (x >  COP2_FLOAT_MAX) VU->VI[REG_MAC_FLAG].UL|= 0x8000;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x8000;

	return (float)x;
}

void VU_MACx_CLEAR(VURegs * VU) {
	VU->VI[REG_MAC_FLAG].UL&= ~0x8888;
}

float VU_MACy_UPDATE(VURegs * VU, double y) {
	if (y == 0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0004; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0004;
	if (y <  0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0040; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0040;
	if (y <  COP2_FLOAT_MIN) VU->VI[REG_MAC_FLAG].UL|= 0x0400;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0400;
	if (y >  COP2_FLOAT_MAX) VU->VI[REG_MAC_FLAG].UL|= 0x4000;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x4000;

	return (float)y;
}

void VU_MACy_CLEAR(VURegs * VU) {
	VU->VI[REG_MAC_FLAG].UL&= ~0x4444;
}

float VU_MACz_UPDATE(VURegs * VU, double z) {
	if (z == 0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0002; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0002;
	if (z <  0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0020; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0020;
	if (z <  COP2_FLOAT_MIN) VU->VI[REG_MAC_FLAG].UL|= 0x0200;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0200;
	if (z >  COP2_FLOAT_MAX) VU->VI[REG_MAC_FLAG].UL|= 0x2000;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x2000;

	return (float)z;
}

void VU_MACz_CLEAR(VURegs * VU) {
	VU->VI[REG_MAC_FLAG].UL&= ~0x2222;
}

float VU_MACw_UPDATE(VURegs * VU, double w) {
	if (w == 0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0001; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0001;
	if (w <  0.0)            VU->VI[REG_MAC_FLAG].UL|= 0x0010; 
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0010;
	if (w <  COP2_FLOAT_MIN) VU->VI[REG_MAC_FLAG].UL|= 0x0100;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x0100;
	if (w >  COP2_FLOAT_MAX) VU->VI[REG_MAC_FLAG].UL|= 0x1000;
	else                     VU->VI[REG_MAC_FLAG].UL&=~0x1000;

	return (float)w;
}

void VU_MACw_CLEAR(VURegs * VU) {
	VU->VI[REG_MAC_FLAG].UL&= ~0x1111;
}

void VU_STAT_UPDATE(VURegs * VU) {
	int newZ=0, newS=0, newU=0, newO=0;

	if (VU->VI[REG_MAC_FLAG].UL & 0x000F) newZ = 1;
	if (VU->VI[REG_MAC_FLAG].UL & 0x00F0) newS = 1;
	if (VU->VI[REG_MAC_FLAG].UL & 0x0F00) newU = 1;
	if (VU->VI[REG_MAC_FLAG].UL & 0xF000) newO = 1;

	VU->VI[REG_STATUS_FLAG].UL|= ((VU->VI[REG_STATUS_FLAG].UL & 0x1) | newZ) << 6;
	VU->VI[REG_STATUS_FLAG].UL|= ((VU->VI[REG_STATUS_FLAG].UL & 0x2) | newS) << 7;
	VU->VI[REG_STATUS_FLAG].UL|= ((VU->VI[REG_STATUS_FLAG].UL & 0x4) | newU) << 8;
	VU->VI[REG_STATUS_FLAG].UL|= ((VU->VI[REG_STATUS_FLAG].UL & 0x8) | newO) << 9;

	VU->VI[REG_STATUS_FLAG].UL|= newZ << 0;
	VU->VI[REG_STATUS_FLAG].UL|= newS << 1;
	VU->VI[REG_STATUS_FLAG].UL|= newU << 2;
	VU->VI[REG_STATUS_FLAG].UL|= newO << 3;
}

