/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef __WIN32__
#pragma warning(disable:4244)
#endif

#include <math.h>
#include <stdlib.h>
#include "Common.h"
#include "Debug.h"
#include "R5900.h"
#include "VUmicro.h"
#include "VUflags.h"
#include "VUops.h"

//Lower/Upper instructions can use that..
#define _Ft_ ((VU->code >> 16) & 0x1F)  // The rt part of the instruction register 
#define _Fs_ ((VU->code >> 11) & 0x1F)  // The rd part of the instruction register 
#define _Fd_ ((VU->code >>  6) & 0x1F)  // The sa part of the instruction register

#define _X ((VU->code>>24) & 0x1)
#define _Y ((VU->code>>23) & 0x1)
#define _Z ((VU->code>>22) & 0x1)
#define _W ((VU->code>>21) & 0x1)

#define _Fsf_ ((VU->code >> 21) & 0x03)
#define _Ftf_ ((VU->code >> 23) & 0x03)

#define _Imm11_		(s32)(VU->code & 0x400 ? 0xfffffc00 | (VU->code & 0x3ff) : VU->code & 0x3ff)
#define _UImm11_	(s32)(VU->code & 0x7ff)


VECTOR RDzero;
/******************************/
/*   VU Upper instructions    */
/******************************/

void _vuABS(VURegs * VU) {
    if (_Ft_ == 0) return;
	if (_X){ VU->VF[_Ft_].f.x = (float)fabs(VU->VF[_Fs_].f.x); }
	if (_Y){ VU->VF[_Ft_].f.y = (float)fabs(VU->VF[_Fs_].f.y); }
	if (_Z){ VU->VF[_Ft_].f.z = (float)fabs(VU->VF[_Fs_].f.z); }
	if (_W){ VU->VF[_Ft_].f.w = (float)fabs(VU->VF[_Fs_].f.w); }
}/*Reworked from define to function. asadr*/


void _vuADD(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/


void _vuADDi(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + VU->VI[REG_I].F);} else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + VU->VI[REG_I].F);} else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + VU->VI[REG_I].F);} else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + VU->VI[REG_I].F);} else VU_MACw_CLEAR(VU);
 	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDq(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + VU->VI[REG_Q].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + VU->VI[REG_Q].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + VU->VI[REG_Q].F); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + VU->VI[REG_Q].F); } else VU_MACw_CLEAR(VU);
 	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/


void _vuADDx(VURegs * VU) {
	float ftx;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftx=VU->VF[_Ft_].f.x;
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + ftx); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + ftx); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + ftx); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + ftx); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDy(VURegs * VU) {
	float fty;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	fty=VU->VF[_Ft_].f.y;
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + fty);} else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + fty);} else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + fty);} else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + fty);} else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDz(VURegs * VU) {
	float ftz;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftz=VU->VF[_Ft_].f.z;
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + ftz); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + ftz); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + ftz); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + ftz); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDw(VURegs * VU) {
	float ftw;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftw=VU->VF[_Ft_].f.w;
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + ftw); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + ftw); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + ftw); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + ftw); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDA(VURegs * VU) {
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAi(VURegs * VU) {
	float ti = VU->VI[REG_I].F;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + ti); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + ti); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + ti); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + ti); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAq(VURegs * VU) {
	float tf = VU->VI[REG_Q].F;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + tf); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + tf); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + tf); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + tf); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAx(VURegs * VU) {
	float tx = VU->VF[_Ft_].f.x;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + tx); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + tx); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + tx); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + tx); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAy(VURegs * VU) {
	float ty = VU->VF[_Ft_].f.y;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + ty); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + ty); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + ty); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + ty); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAz(VURegs * VU) {
	float tz = VU->VF[_Ft_].f.z;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + tz); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + tz); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + tz); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + tz); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/

void _vuADDAw(VURegs * VU) {
	float tw = VU->VF[_Ft_].f.w;
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x + tw); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y + tw); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z + tw); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w + tw); } else VU_MACw_CLEAR(VU);
	VU_STAT_UPDATE(VU);
}/*Reworked from define to function. asadr*/


void _vuSUB(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VF[_Ft_].f.x);  } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VF[_Ft_].f.y);  } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VF[_Ft_].f.z);  } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VF[_Ft_].f.w);  } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow



void _vuSUBi(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VI[REG_I].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VI[REG_I].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VI[REG_I].F); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VI[REG_I].F); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBq(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VI[REG_Q].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VI[REG_Q].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VI[REG_Q].F); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VI[REG_Q].F); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow


void _vuSUBx(VURegs * VU) {
	float ftx;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftx=VU->VF[_Ft_].f.x;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - ftx); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - ftx); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - ftx); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - ftx); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow


void _vuSUBy(VURegs * VU) {
	float fty;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	fty=VU->VF[_Ft_].f.y;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - fty); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - fty); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - fty); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - fty); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow


void _vuSUBz(VURegs * VU) {
	float ftz;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
 	ftz=VU->VF[_Ft_].f.z;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - ftz); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - ftz); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - ftz); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - ftz); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow


void _vuSUBw(VURegs * VU) {
	float ftw;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    ftw=VU->VF[_Ft_].f.w;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - ftw); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - ftw); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - ftw); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - ftw); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow


void _vuSUBA(VURegs * VU) {
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAi(VURegs * VU) {
	
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VI[REG_I].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VI[REG_I].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VI[REG_I].F); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VI[REG_I].F); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAq(VURegs * VU) {
    
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - VU->VI[REG_Q].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - VU->VI[REG_Q].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - VU->VI[REG_Q].F); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - VU->VI[REG_Q].F); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAx(VURegs * VU) {
	float tx = VU->VF[_Ft_].f.x;
    
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - tx); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - tx); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - tx); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - tx); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAy(VURegs * VU) {
	float ty = VU->VF[_Ft_].f.y;
    
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - ty); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - ty); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - ty); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - ty); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAz(VURegs * VU) {
	float tz = VU->VF[_Ft_].f.z;
    
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - tz); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - tz); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - tz); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - tz); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow

void _vuSUBAw(VURegs * VU) {
	float tw = VU->VF[_Ft_].f.w;
    
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x - tw); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y - tw); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z - tw); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w - tw); } else VU_MACw_CLEAR(VU);
 VU_STAT_UPDATE(VU);
}//updated 10/05/03 shadow



void _vuMUL(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */

/* No need to presave I reg in ti. asadr */
void _vuMULi(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VI[REG_I].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VI[REG_I].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VI[REG_I].F); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VI[REG_I].F); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */

void _vuMULq(VURegs * VU) {
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
 	
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VI[REG_Q].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VI[REG_Q].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VI[REG_Q].F); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VI[REG_Q].F); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

void _vuMULx(VURegs * VU) {
	float ftx;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
 	ftx=VU->VF[_Ft_].f.x;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * ftx); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * ftx); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * ftx); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * ftx); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */


void _vuMULy(VURegs * VU) {
	float fty;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
 	fty=VU->VF[_Ft_].f.y;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * fty); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * fty); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * fty); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * fty); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */

void _vuMULz(VURegs * VU) {
	float ftz;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
 	ftz=VU->VF[_Ft_].f.z;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * ftz); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * ftz); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * ftz); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * ftz); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */

void _vuMULw(VURegs * VU) {
	float ftw;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftw=VU->VF[_Ft_].f.w;
    
	if (_X){ dst->f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * ftw); } else VU_MACx_CLEAR(VU);
	if (_Y){ dst->f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * ftw); } else VU_MACy_CLEAR(VU);
	if (_Z){ dst->f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * ftw); } else VU_MACz_CLEAR(VU);
	if (_W){ dst->f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * ftw); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */


void _vuMULA(VURegs * VU) {
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU);
}/* last update 8/05/03 shadow */

/* No need to presave I reg in ti. asadr */
void _vuMULAi(VURegs * VU) {
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VI[REG_I].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VI[REG_I].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VI[REG_I].F); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VI[REG_I].F); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

/* No need to presave Q reg in ti. asadr */
void _vuMULAq(VURegs * VU) { 
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VI[REG_Q].F); } else VU_MACx_CLEAR(VU);
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VI[REG_Q].F); } else VU_MACy_CLEAR(VU);
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VI[REG_Q].F); } else VU_MACz_CLEAR(VU);
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VI[REG_Q].F); } else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

/* No need to presave X reg in ti. asadr */
void _vuMULAx(VURegs * VU) { 
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x); } else VU_MACx_CLEAR(VU); 
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.x); } else VU_MACy_CLEAR(VU); 
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.x); } else VU_MACz_CLEAR(VU); 
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.x); } else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */


void _vuMULAy(VURegs * VU) { 
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.y); } else VU_MACx_CLEAR(VU); 
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y); } else VU_MACy_CLEAR(VU); 
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.y); } else VU_MACz_CLEAR(VU); 
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.y); } else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

void _vuMULAz(VURegs * VU) { 
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.z); } else VU_MACx_CLEAR(VU); 
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.z); } else VU_MACy_CLEAR(VU); 
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z); } else VU_MACz_CLEAR(VU); 
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.z); } else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

void _vuMULAw(VURegs * VU) { 
	if (_X){ VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.w); } else VU_MACx_CLEAR(VU); 
	if (_Y){ VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.w); } else VU_MACy_CLEAR(VU); 
	if (_Z){ VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.w); } else VU_MACz_CLEAR(VU); 
	if (_W){ VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w); } else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 8/05/03 shadow */

void _vuMADD(VURegs * VU) { 
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */


void _vuMADDi(VURegs * VU) { 
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * VU->VI[REG_I].F ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * VU->VI[REG_I].F ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * VU->VI[REG_I].F ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * VU->VI[REG_I].F ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

/* No need to presave . asadr */
void _vuMADDq(VURegs * VU) { 
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * VU->VI[REG_Q].F  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * VU->VI[REG_Q].F  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * VU->VI[REG_Q].F  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * VU->VI[REG_Q].F  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

void _vuMADDx(VURegs * VU) { 
	float ftx;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
 	ftx=VU->VF[_Ft_].f.x;
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * ftx  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * ftx  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * ftx  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * ftx  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

void _vuMADDy(VURegs * VU) { 
	float fty;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
 	fty=VU->VF[_Ft_].f.y;
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * fty  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * fty  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * fty  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * fty  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

void _vuMADDz(VURegs * VU) { 
	float ftz;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
 	ftz=VU->VF[_Ft_].f.z;
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * ftz ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * ftz ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * ftz ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * ftz ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

void _vuMADDw(VURegs * VU) { 
	float ftw;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_]; 	
 	ftw=VU->VF[_Ft_].f.w;
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  + ( VU->VF[_Fs_].f.x * ftw  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  + ( VU->VF[_Fs_].f.y * ftw  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  + ( VU->VF[_Fs_].f.z * ftw  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  + ( VU->VF[_Fs_].f.w * ftw  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 10/05/03 shadow */

void _vuMADDA(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  10/05/03 shadow*/

void _vuMADDAi(VURegs * VU) { 
	float ti = VU->VI[REG_I].F;

    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * ti)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * ti)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * ti)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * ti)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  10/05/03 shadow*/

void _vuMADDAq(VURegs * VU) { 
	float tq = VU->VI[REG_Q].F;
    
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * tq)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * tq)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * tq)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * tq)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last update  10/05/03 shadow*/

void _vuMADDAx(VURegs * VU) {
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.x)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.x)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.x)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last update 11/05/03 shadow*/

void _vuMADDAy(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.y)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.y)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.y)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last update  11/05/03 shadow*/

void _vuMADDAz(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.z)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.z)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.z)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last update  11/05/03 shadow*/

void _vuMADDAw(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x + ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.w)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y + ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.w)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z + ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.w)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w + ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last update  11/05/03 shadow*/

void _vuMSUB(VURegs * VU) { 
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */

void _vuMSUBi(VURegs * VU) { 
	float ti = VU->VI[REG_I].F;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * ti  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * ti  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * ti  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * ti  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */

void _vuMSUBq(VURegs * VU) { 
	float tq = VU->VI[REG_Q].F;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * tq  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * tq  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * tq  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * tq  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */


void _vuMSUBx(VURegs * VU) { 
	float ftx;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftx=VU->VF[_Ft_].f.x;
    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * ftx  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * ftx  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * ftx  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * ftx  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */


void _vuMSUBy(VURegs * VU) { 
	float fty;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	fty=VU->VF[_Ft_].f.y;
    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * fty  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * fty  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * fty  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * fty  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */


void _vuMSUBz(VURegs * VU) { 
	float ftz;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftz=VU->VF[_Ft_].f.z;
    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * ftz  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * ftz  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * ftz  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * ftz  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */

void _vuMSUBw(VURegs * VU) { 
	float ftw;
	VECTOR * dst;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];
	ftw=VU->VF[_Ft_].f.w;
    
    if (_X) dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x  - ( VU->VF[_Fs_].f.x * ftw  ) ); else VU_MACx_CLEAR(VU); 
    if (_Y) dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y  - ( VU->VF[_Fs_].f.y * ftw  ) ); else VU_MACy_CLEAR(VU); 
    if (_Z) dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z  - ( VU->VF[_Fs_].f.z * ftw  ) ); else VU_MACz_CLEAR(VU); 
    if (_W) dst->f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w  - ( VU->VF[_Fs_].f.w * ftw  ) ); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/* last update 11/05/03 shadow */


void _vuMSUBA(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.x)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.y)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.z)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * VU->VF[_Ft_].f.w)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMSUBAi(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * VU->VI[REG_I].F)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * VU->VI[REG_I].F)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * VU->VI[REG_I].F)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * VU->VI[REG_I].F)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMSUBAq(VURegs * VU) { 
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * VU->VI[REG_Q].F)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * VU->VI[REG_Q].F)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * VU->VI[REG_Q].F)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * VU->VI[REG_Q].F)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMSUBAx(VURegs * VU) { 
	float tx = VU->VF[_Ft_].f.x;
	
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * tx)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * tx)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * tx)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * tx)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMSUBAy(VURegs * VU) { 
	float ty = VU->VF[_Ft_].f.y;
    
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * ty)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * ty)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * ty)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * ty)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMSUBAz(VURegs * VU) { 
	float tz = VU->VF[_Ft_].f.z;
    
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * tz)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * tz)); else VU_MACy_CLEAR(VU); 
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * tz)); else VU_MACz_CLEAR(VU); 
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * tz)); else VU_MACw_CLEAR(VU); 
    VU_STAT_UPDATE(VU);
}/*last updated  11/05/03 shadow*/

void _vuMSUBAw(VURegs * VU) { 
	float tw = VU->VF[_Ft_].f.w;
    
    if (_X) VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - ( VU->VF[_Fs_].f.x * tw)); else VU_MACx_CLEAR(VU); 
    if (_Y) VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - ( VU->VF[_Fs_].f.y * tw)); else VU_MACy_CLEAR(VU);
    if (_Z) VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - ( VU->VF[_Fs_].f.z * tw)); else VU_MACz_CLEAR(VU);
    if (_W) VU->ACC.f.w = VU_MACw_UPDATE(VU, (double)VU->ACC.f.w - ( VU->VF[_Fs_].f.w * tw)); else VU_MACw_CLEAR(VU);
    VU_STAT_UPDATE(VU); 
}/*last updated  11/05/03 shadow*/

void _vuMAX(VURegs * VU) { 
	if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > VU->VF[_Ft_].f.x) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = VU->VF[_Ft_].f.x; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > VU->VF[_Ft_].f.y) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = VU->VF[_Ft_].f.y; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > VU->VF[_Ft_].f.z) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = VU->VF[_Ft_].f.z; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > VU->VF[_Ft_].f.w) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = VU->VF[_Ft_].f.w; 
	} 
}//checked 13/05/03 shadow

void _vuMAXi(VURegs * VU) { 
		if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = VU->VI[REG_I].F; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = VU->VI[REG_I].F; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = VU->VI[REG_I].F; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = VU->VI[REG_I].F; 
	} 
}//checked 13/05/03 shadow

void _vuMAXx(VURegs * VU) { 
	float ftx;
	if (_Fd_ == 0) return; 
 
	ftx=VU->VF[_Ft_].f.x;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > ftx) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = ftx; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > ftx) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftx; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > ftx) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftx; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > ftx) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftx; 
	} 
}
//checked 13/05/03 shadow

void _vuMAXy(VURegs * VU) { 
	float fty;
	if (_Fd_ == 0) return; 
 
	fty=VU->VF[_Ft_].f.y;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > fty) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = fty; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > fty) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = fty; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > fty) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = fty; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > fty) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = fty; 
	} 
}//checked 13/05/03 shadow

void _vuMAXz(VURegs * VU) { 
	float ftz;
	if (_Fd_ == 0) return; 
 
	ftz=VU->VF[_Ft_].f.z;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > ftz) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = ftz; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > ftz) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftz; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > ftz) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftz; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > ftz) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftz; 
	} 
}

void _vuMAXw(VURegs * VU) { 
	float ftw;
	if (_Fd_ == 0) return; 
 
	ftw=VU->VF[_Ft_].f.w;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x > ftw) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = ftw; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y > ftw) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftw; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z > VU->VF[_Ft_].f.w) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftw; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w > VU->VF[_Ft_].f.w) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftw; 
	} 
}

void _vuMINI(VURegs * VU) { 
	if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x < VU->VF[_Ft_].f.x) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = VU->VF[_Ft_].f.x; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < VU->VF[_Ft_].f.y) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = VU->VF[_Ft_].f.y; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < VU->VF[_Ft_].f.z) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = VU->VF[_Ft_].f.z; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < VU->VF[_Ft_].f.w) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = VU->VF[_Ft_].f.w; 
	} 
}

void _vuMINIi(VURegs * VU) { 
	if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x < VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = VU->VI[REG_I].F; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = VU->VI[REG_I].F; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = VU->VI[REG_I].F; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < VU->VI[REG_I].F) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = VU->VI[REG_I].F; 
	} 
}

void _vuMINIx(VURegs * VU) { 
	float ftx;
	if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	ftx=VU->VF[_Ft_].f.x;
	if (_X) { 
		if (VU->VF[_Fs_].f.x < ftx) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else
			VU->VF[_Fd_].f.x = ftx; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < ftx) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftx; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < ftx) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftx; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < ftx) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftx; 
	} 
}//checked 13/05/03 shadow

void _vuMINIy(VURegs * VU) { 
	float fty;
	if (_Fd_ == 0) return; 
 
	/* ft is bc */ 
	fty=VU->VF[_Ft_].f.y;
	if (_X) { 
		if (VU->VF[_Fs_].f.x < fty) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else
			VU->VF[_Fd_].f.x = fty; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < fty) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = fty; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < fty) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = fty; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < fty) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = fty; 
	} 
}//checked 13/05/03 shadow

void _vuMINIz(VURegs * VU) { 
	float ftz;
	if (_Fd_ == 0) return; 
 
	ftz=VU->VF[_Ft_].f.z;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x < ftz) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = ftz; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < ftz) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftz; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < ftz) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftz; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < ftz) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftz; 
	} 
}//checked 13/05/03 shadow

void _vuMINIw(VURegs * VU) { 
	float ftw;
	if (_Fd_ == 0) return; 
 
	ftw=VU->VF[_Ft_].f.w;
	/* ft is bc */ 
	if (_X) { 
		if (VU->VF[_Fs_].f.x < ftw) 
			VU->VF[_Fd_].f.x = VU->VF[_Fs_].f.x; 
		else 
			VU->VF[_Fd_].f.x = ftw; 
	} 
 
	if (_Y) { 
		if (VU->VF[_Fs_].f.y < ftw) 
			VU->VF[_Fd_].f.y = VU->VF[_Fs_].f.y; 
		else 
			VU->VF[_Fd_].f.y = ftw; 
	} 
 
	if (_Z) { 
		if (VU->VF[_Fs_].f.z < ftw) 
			VU->VF[_Fd_].f.z = VU->VF[_Fs_].f.z; 
		else 
			VU->VF[_Fd_].f.z = ftw; 
	} 
 
	if (_W) { 
		if (VU->VF[_Fs_].f.w < ftw) 
			VU->VF[_Fd_].f.w = VU->VF[_Fs_].f.w; 
		else 
			VU->VF[_Fd_].f.w = ftw; 
	} 
}//checked 13/05/03 shadow

void _vuOPMULA(VURegs * VU) { 
    VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.z); 
    VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.x); 
    VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.y); 
    VU_STAT_UPDATE(VU); 
}/*last updated  8/05/03 shadow*/

void _vuOPMSUB(VURegs * VU) { 
	VECTOR * dst;
	float ftx, fty, ftz;
	float fsx, fsy, fsz;
    if (_Fd_ == 0) dst = &RDzero;
	else dst = &VU->VF[_Fd_];

	ftx = VU->VF[_Ft_].f.x; fty = VU->VF[_Ft_].f.y; ftz = VU->VF[_Ft_].f.z;
	fsx = VU->VF[_Fs_].f.x; fsy = VU->VF[_Fs_].f.y; fsz = VU->VF[_Fs_].f.z;
    dst->f.x = VU_MACx_UPDATE(VU, (double)VU->ACC.f.x - fsy * ftz); 
    dst->f.y = VU_MACy_UPDATE(VU, (double)VU->ACC.f.y - fsz * ftx); 
    dst->f.z = VU_MACz_UPDATE(VU, (double)VU->ACC.f.z - fsx * fty); 
    VU_STAT_UPDATE(VU); 
}/*last updated  8/05/03 shadow*/

void _vuNOP(VURegs * VU) { 
}

void _vuFTOI0(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].SL[0] = (s32)VU->VF[_Fs_].f.x; 
	if (_Y) VU->VF[_Ft_].SL[1] = (s32)VU->VF[_Fs_].f.y; 
	if (_Z) VU->VF[_Ft_].SL[2] = (s32)VU->VF[_Fs_].f.z; 
	if (_W) VU->VF[_Ft_].SL[3] = (s32)VU->VF[_Fs_].f.w; 
}

void _vuFTOI4(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].SL[0] = float_to_int4(VU->VF[_Fs_].f.x); 
	if (_Y) VU->VF[_Ft_].SL[1] = float_to_int4(VU->VF[_Fs_].f.y); 
	if (_Z) VU->VF[_Ft_].SL[2] = float_to_int4(VU->VF[_Fs_].f.z); 
	if (_W) VU->VF[_Ft_].SL[3] = float_to_int4(VU->VF[_Fs_].f.w); 
}

void _vuFTOI12(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].SL[0] = float_to_int12(VU->VF[_Fs_].f.x); 
	if (_Y) VU->VF[_Ft_].SL[1] = float_to_int12(VU->VF[_Fs_].f.y); 
	if (_Z) VU->VF[_Ft_].SL[2] = float_to_int12(VU->VF[_Fs_].f.z); 
	if (_W) VU->VF[_Ft_].SL[3] = float_to_int12(VU->VF[_Fs_].f.w); 
}

void _vuFTOI15(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].SL[0] = float_to_int15(VU->VF[_Fs_].f.x); 
	if (_Y) VU->VF[_Ft_].SL[1] = float_to_int15(VU->VF[_Fs_].f.y); 
	if (_Z) VU->VF[_Ft_].SL[2] = float_to_int15(VU->VF[_Fs_].f.z); 
	if (_W) VU->VF[_Ft_].SL[3] = float_to_int15(VU->VF[_Fs_].f.w); 
}

void _vuITOF0(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].f.x = (float)VU->VF[_Fs_].SL[0]; 
	if (_Y) VU->VF[_Ft_].f.y = (float)VU->VF[_Fs_].SL[1]; 
	if (_Z) VU->VF[_Ft_].f.z = (float)VU->VF[_Fs_].SL[2]; 
	if (_W) VU->VF[_Ft_].f.w = (float)VU->VF[_Fs_].SL[3]; 
}

void _vuITOF4(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].f.x = int4_to_float(VU->VF[_Fs_].SL[0]); 
	if (_Y) VU->VF[_Ft_].f.y = int4_to_float(VU->VF[_Fs_].SL[1]); 
	if (_Z) VU->VF[_Ft_].f.z = int4_to_float(VU->VF[_Fs_].SL[2]); 
	if (_W) VU->VF[_Ft_].f.w = int4_to_float(VU->VF[_Fs_].SL[3]); 
}

void _vuITOF12(VURegs * VU) { 
	if (_Ft_ == 0) return; 
	if (_X) VU->VF[_Ft_].f.x = int12_to_float(VU->VF[_Fs_].SL[0]); 
	if (_Y) VU->VF[_Ft_].f.y = int12_to_float(VU->VF[_Fs_].SL[1]); 
	if (_Z) VU->VF[_Ft_].f.z = int12_to_float(VU->VF[_Fs_].SL[2]); 
	if (_W) VU->VF[_Ft_].f.w = int12_to_float(VU->VF[_Fs_].SL[3]); 
}

void _vuITOF15(VURegs * VU) { 
	if (_Ft_ == 0) return; 
 	if (_X) VU->VF[_Ft_].f.x = int15_to_float(VU->VF[_Fs_].SL[0]); 
	if (_Y) VU->VF[_Ft_].f.y = int15_to_float(VU->VF[_Fs_].SL[1]); 
	if (_Z) VU->VF[_Ft_].f.z = int15_to_float(VU->VF[_Fs_].SL[2]); 
	if (_W) VU->VF[_Ft_].f.w = int15_to_float(VU->VF[_Fs_].SL[3]); 
}

/* Different type of clipping by presaving w. asadr */
void _vuCLIP(VURegs * VU) { 
	float value = fabs(VU->VF[_Ft_].f.w);

	VU->VI[REG_CLIP_FLAG].UL<<= 6; 
    if ( VU->VF[_Fs_].f.x > +value ) VU->VI[REG_CLIP_FLAG].UL|= 0x01;
	if ( VU->VF[_Fs_].f.x < -value ) VU->VI[REG_CLIP_FLAG].UL|= 0x02;
	if ( VU->VF[_Fs_].f.y > +value ) VU->VI[REG_CLIP_FLAG].UL|= 0x04;
	if ( VU->VF[_Fs_].f.y < -value ) VU->VI[REG_CLIP_FLAG].UL|= 0x08;
	if ( VU->VF[_Fs_].f.z > +value ) VU->VI[REG_CLIP_FLAG].UL|= 0x10;
	if ( VU->VF[_Fs_].f.z < -value ) VU->VI[REG_CLIP_FLAG].UL|= 0x20;
}/*last update 11/05/03 shadow */

/******************************/ 
/*   VU Lower instructions    */ 
/******************************/ 

void _vuDIV(VURegs * VU) { 
    if ( VU->VF[_Ft_].F[_Ftf_] == 0.0 ) { 
		 if ( VU->VF[_Fs_].F[_Fsf_] == 0.0 ) 
			 VU_UPDATE_STAT_REG_DI2(VU);
		 else 
			 VU_UPDATE_STAT_REG_DI1(VU);
	} else {
		VU_UPDATE_STAT_REG_DI0(VU);
		VU->VI[REG_Q].F =  VU->VF[_Fs_].F[_Fsf_] / VU->VF[_Ft_].F[_Ftf_]; 
	} 
}/*last update  8/05/03 shadow*/

void _vuSQRT(VURegs * VU) { 
	if ( VU->VF[_Ft_].F[_Ftf_] < 0.0 ) 
		VU_UPDATE_STAT_REG_DI2(VU);
	else VU_UPDATE_STAT_REG_DI3(VU);
	VU->VI[REG_Q].F = (float)sqrt(fabs(VU->VF[_Ft_].F[_Ftf_])) ; 
}/*last update  8/05/03 shadow*/


/* Eminent Bug - Dvisior == 0 Check Missing ( D Flag Not Set ) */
/* REFIXED....ASADR */
void _vuRSQRT(VURegs * VU) { 
	float temp; 

	if ( VU->VF[_Ft_].F[_Ftf_] < 0.0 ) {
		VU_UPDATE_STAT_REG_DI2(VU);
	} else { 
		temp = (float)sqrt(fabs(VU->VF[_Ft_].F[_Ftf_] ) ); 
		if ( temp == 0.0 ) { 
			VU_UPDATE_STAT_REG_DI1(VU); 
		}else{ 
			VU_UPDATE_STAT_REG_DI3(VU); 
			VU->VI[REG_Q].F =  ( VU->VF[_Fs_].F[_Fsf_] / temp);
		} 
	} 
}/*last update 11/05/03 shadow */


void _vuIADDI(VURegs * VU) { 
	s16 imm = ((VU->code >> 6) & 0x1f); 
	imm = ((imm & 0x10 ? 0xfff0 : 0) | (imm & 0xf));
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].SS[0] = VU->VI[_Fs_].SS[0] + imm;
}//last checked 17/05/03 shadow NOTE: not quite sure about that

void _vuIADDIU(VURegs * VU) { 
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].US[0] = VU->VI[_Fs_].US[0] + (((VU->code >> 10) & 0x7800) | (VU->code & 0x7ff)); 
}//last checked 17/05/03 shadow

void _vuIADD(VURegs * VU) { 
	if(_Fd_ == 0) return; 
	VU->VI[_Fd_].US[0] = VU->VI[_Fs_].US[0] + VU->VI[_Ft_].US[0]; 
}//last checked 17/05/03 shadow

void _vuIAND(VURegs * VU) { 
	if(_Fd_ == 0) return; 
	VU->VI[_Fd_].US[0] = VU->VI[_Fs_].US[0] & VU->VI[_Ft_].US[0]; 
}//last checked 17/05/03 shadow

void _vuIOR(VURegs * VU) { 
	if(_Fd_ == 0) return; 
	VU->VI[_Fd_].US[0] = VU->VI[_Fs_].US[0] | VU->VI[_Ft_].US[0]; 
}

void _vuISUB(VURegs * VU) { 
	if(_Fd_ == 0) return; 
	VU->VI[_Fd_].US[0] = VU->VI[_Fs_].US[0] - VU->VI[_Ft_].US[0]; 
}

void _vuISUBIU(VURegs * VU) { 
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].US[0] = VU->VI[_Fs_].US[0] - (((VU->code & 0x1E00000) << 11) | (VU->code & 0x7ff)); 
}

void _vuMOVE(VURegs * VU) { 
	if(_Ft_ == 0) return; 
 
	if (_X) VU->VF[_Ft_].UL[0] = VU->VF[_Fs_].UL[0]; 
	if (_Y) VU->VF[_Ft_].UL[1] = VU->VF[_Fs_].UL[1]; 
	if (_Z) VU->VF[_Ft_].UL[2] = VU->VF[_Fs_].UL[2]; 
	if (_W) VU->VF[_Ft_].UL[3] = VU->VF[_Fs_].UL[3]; 
}//last checked 17/05/03 shadow

void _vuMFIR(VURegs * VU) { 
	if (_Ft_ == 0) return; 

	if (_X) VU->VF[_Ft_].SL[0] = (s32)VU->VI[_Fs_].SS[0];
	if (_Y) VU->VF[_Ft_].SL[1] = (s32)VU->VI[_Fs_].SS[0]; 
	if (_Z) VU->VF[_Ft_].SL[2] = (s32)VU->VI[_Fs_].SS[0]; 
	if (_W) VU->VF[_Ft_].SL[3] = (s32)VU->VI[_Fs_].SS[0]; 
}

// Big bug!!! mov from fs to ft not ft to fs. asadr
void _vuMTIR(VURegs * VU) { 
	if(_Ft_ == 0) return;
    VU->VI[_Ft_].US[0] =  *(u16*)&VU->VF[_Fs_].F[_Fsf_]; 
} 


void _vuMR32(VURegs * VU) {
	float tx = VU->VF[_Fs_].f.x;
	if (_Ft_ == 0) return;

	if (_X) VU->VF[_Ft_].f.x = VU->VF[_Fs_].f.y; 
	if (_Y) VU->VF[_Ft_].f.y = VU->VF[_Fs_].f.z; 
	if (_Z) VU->VF[_Ft_].f.z = VU->VF[_Fs_].f.w; 
	if (_W) VU->VF[_Ft_].f.w = tx; 
}//last updated 23/10/03 linuzappz

void _vuLQ(VURegs * VU) { 
	s16 imm; 
	u16 addr;
	u32 *ptr; 

	if (_Ft_ == 0) return; 
 
	imm = (VU->code & 0x400) ? (VU->code & 0x3ff) | 0xfc00 : (VU->code & 0x3ff); 
	addr = (imm + VU->VI[_Fs_].SS[0]) * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("LQ error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
 	ptr = (u32*)(VU->Mem + addr);
	if (_X) VU->VF[_Ft_].UL[0] = ptr[0]; 
	if (_Y) VU->VF[_Ft_].UL[1] = ptr[1]; 
	if (_Z) VU->VF[_Ft_].UL[2] = ptr[2]; 
	if (_W) VU->VF[_Ft_].UL[3] = ptr[3]; 
}

void _vuLQD( VURegs * VU ) { 
	u32 addr;
	u32 *ptr; 

	if (_Fs_ != 0) VU->VI[_Fs_].US[0]--; 
	if (_Ft_ == 0) return;

	addr = VU->VI[_Fs_].US[0] * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("LQD error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
	ptr = (u32*)(VU->Mem + addr);
	if (_X) VU->VF[_Ft_].UL[0] = ptr[0]; 
	if (_Y) VU->VF[_Ft_].UL[1] = ptr[1]; 
	if (_Z) VU->VF[_Ft_].UL[2] = ptr[2]; 
	if (_W) VU->VF[_Ft_].UL[3] = ptr[3]; 
}

void _vuLQI(VURegs * VU) { 
	if (_Ft_) {
		u32 addr;
		u32 *ptr;

		addr = VU->VI[_Fs_].US[0] * 16;
		if (VU->maxmem == -1) {
			addr&= 0x3fff;
		} else {
			if (addr > VU->maxmem) {
#ifdef CPU_LOG
				CPU_LOG("LQI error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
				return;
			}
		}

		ptr = (u32*)(VU->Mem + addr);
		if (_X) VU->VF[_Ft_].UL[0] = ptr[0]; 
		if (_Y) VU->VF[_Ft_].UL[1] = ptr[1]; 
		if (_Z) VU->VF[_Ft_].UL[2] = ptr[2]; 
		if (_W) VU->VF[_Ft_].UL[3] = ptr[3]; 
   }
   if (_Fs_ != 0) VU->VI[_Fs_].US[0]++; 
}

/* addr is now signed. Asadr */
void _vuSQ(VURegs * VU) { 
	s16 imm; 
	u16 addr;
	u32 *ptr; 
 
	imm = (VU->code & 0x400) ? (VU->code & 0x3ff) | 0xfc00 : (VU->code & 0x3ff); 
	addr = (imm + VU->VI[_Ft_].SS[0]) * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("SQ error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
 	ptr = (u32*)(VU->Mem + addr); 
	if (_X) ptr[0] = VU->VF[_Fs_].UL[0]; 
	if (_Y) ptr[1] = VU->VF[_Fs_].UL[1]; 
	if (_Z) ptr[2] = VU->VF[_Fs_].UL[2]; 
	if (_W) ptr[3] = VU->VF[_Fs_].UL[3]; 
}

void _vuSQD(VURegs * VU) { 
	u32 addr;
	u32 *ptr; 
 
	if(_Ft_ != 0) VU->VI[_Ft_].US[0]--; 
	addr = VU->VI[_Ft_].US[0] * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("SQD error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
 	ptr = (u32*)(VU->Mem + addr);
	if (_X) ptr[0] = VU->VF[_Fs_].UL[0]; 
	if (_Y) ptr[1] = VU->VF[_Fs_].UL[1]; 
	if (_Z) ptr[2] = VU->VF[_Fs_].UL[2]; 
	if (_W) ptr[3] = VU->VF[_Fs_].UL[3]; 
}

void _vuSQI(VURegs * VU) { 
	u32 addr;
	u32 *ptr; 
 
	addr = VU->VI[_Ft_].US[0] * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("SQI error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
	ptr = (u32*)(VU->Mem + addr); 
	if (_X) ptr[0] = VU->VF[_Fs_].UL[0];
	if (_Y) ptr[1] = VU->VF[_Fs_].UL[1];
	if (_Z) ptr[2] = VU->VF[_Fs_].UL[2];
	if (_W) ptr[3] = VU->VF[_Fs_].UL[3];
	if(_Ft_ != 0) VU->VI[_Ft_].US[0]++; 
}

/* addr now signed. asadr */
void _vuILW(VURegs * VU) { 
	s16 imm; 
	u16 addr;
	u16 *ptr; 
	if (_Ft_ == 0) return; 
 
 	imm = (VU->code & 0x400) ? (VU->code & 0x3ff) | 0xfc00 : (VU->code & 0x3ff); 
	addr = (imm + VU->VI[_Fs_].SS[0]) * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("ILW error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
 	ptr = (u16*)(VU->Mem + addr); 
	if (_X) VU->VI[_Ft_].US[0] = ptr[0]; 
	if (_Y) VU->VI[_Ft_].US[0] = ptr[2]; 
	if (_Z) VU->VI[_Ft_].US[0] = ptr[4]; 
	if (_W) VU->VI[_Ft_].US[0] = ptr[6]; 
}

void _vuISW(VURegs * VU) { 
	s16 imm; 
	u16 addr;
	u16 *ptr; 
 
 	imm = (VU->code & 0x400) ? (VU->code & 0x3ff) | 0xfc00 : (VU->code & 0x3ff); 
	addr = (imm + VU->VI[_Fs_].SS[0]) * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("ISW error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}

	ptr = (u16*)(VU->Mem + addr); 
	if (_X) ptr[0] = VU->VI[_Ft_].US[0]; 
	if (_Y) ptr[2] = VU->VI[_Ft_].US[0]; 
	if (_Z) ptr[4] = VU->VI[_Ft_].US[0]; 
	if (_W) ptr[6] = VU->VI[_Ft_].US[0]; 
}

void _vuILWR(VURegs * VU) {
	u32 addr;
	u16 *ptr; 
	if (_Ft_ == 0) return; 
 
	addr = VU->VI[_Fs_].US[0] * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("ILWR error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}
 	ptr = (u16*)(VU->Mem + addr); 
	if (_X) VU->VI[_Ft_].US[0] = ptr[0]; 
	if (_Y) VU->VI[_Ft_].US[0] = ptr[2]; 
	if (_Z) VU->VI[_Ft_].US[0] = ptr[4]; 
	if (_W) VU->VI[_Ft_].US[0] = ptr[6]; 
}

void _vuISWR(VURegs * VU) { 
	u32 addr;
	u16 *ptr; 
 
	addr = VU->VI[_Fs_].US[0] * 16;
	if (VU->maxmem == -1) {
		addr&= 0x3fff;
	} else
	if (addr > VU->maxmem) {
#ifdef CPU_LOG
		CPU_LOG("ISWR error, overflow: %x > %x\n", addr, VU->maxmem);
#endif
		return;
	}

	ptr = (u16*)(VU->Mem + addr); 
	if (_X) ptr[0] = VU->VI[_Ft_].US[0]; 
	if (_Y) ptr[2] = VU->VI[_Ft_].US[0]; 
	if (_Z) ptr[4] = VU->VI[_Ft_].US[0]; 
	if (_W) ptr[6] = VU->VI[_Ft_].US[0]; 
}

void _vuLOI(VURegs * VU) { 
#ifdef CPU_LOG
	CPU_LOG("VU->LOI opcode unhandled %x\n",cpuRegs.pc); 
#endif
}

void _vuRINIT(VURegs * VU) { 
	VU->VI[REG_R].F = VU->VF[_Fs_].F[_Fsf_]; 
}


void _vuRGET(VURegs * VU) { 
	if (_Ft_ == 0) return;
 
	if (_X) VU->VF[_Ft_].UL[0] = VU->VI[REG_R].UL;
	if (_Y) VU->VF[_Ft_].UL[1] = VU->VI[REG_R].UL;
	if (_Z) VU->VF[_Ft_].UL[2] = VU->VI[REG_R].UL;
	if (_W) VU->VF[_Ft_].UL[3] = VU->VI[REG_R].UL;
}

void _vuRNEXT(VURegs * VU) { 
	if (_Ft_ == 0) return;
	VU->VI[REG_R].F = 1.0f+(float)(1.0*rand()/(RAND_MAX+1.0));
	if (_X) VU->VF[_Ft_].UL[0] = VU->VI[REG_R].UL;
	if (_Y) VU->VF[_Ft_].UL[1] = VU->VI[REG_R].UL;
	if (_Z) VU->VF[_Ft_].UL[2] = VU->VI[REG_R].UL;
	if (_W) VU->VF[_Ft_].UL[3] = VU->VI[REG_R].UL;
}

void _vuRXOR(VURegs * VU) { 
	VU->VI[REG_R].F = (int)VU->VI[REG_R].F ^ (int)VU->VF[_Fs_].F[_Fsf_]; 
}

void _vuWAITQ(VURegs * VU) { 
}

// Fixed
void _vuFSAND(VURegs * VU) { 
	u16 imm = 0;
	imm = ((VU->code >> 21 ) & 0x1) << 0xb | (VU->code & 0x7ff);
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].UL = VU->VI[REG_STATUS_FLAG].UL & imm; 
}

void _vuFSEQ(VURegs * VU) { 
	u16 imm = 0;
	imm = ((VU->code >> 21 ) & 0x1) << 0xb | (VU->code & 0x7ff);
	if(_Ft_ == 0) return; 
	if(VU->VI[REG_STATUS_FLAG].UL == imm) VU->VI[_Ft_].UL = 1; 
	else VU->VI[_Ft_].UL = 0; 
}

// Fixed
void _vuFSOR(VURegs * VU) { 
	u16 imm = 0;
	imm = ((VU->code >> 21 ) & 0x1) << 0xb | (VU->code & 0x7ff);
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].UL = VU->VI[REG_STATUS_FLAG].UL | imm; 
}

void _vuFSSET(VURegs * VU) { 
	u16 imm = 0;
	imm = ((VU->code >> 21 ) & 0x1) << 0x5 | ((VU->code >> 6) & 0x1F);
	VU->VI[REG_STATUS_FLAG].UL = (imm & 0x3f) << 0x6 | (VU->VI[REG_STATUS_FLAG].UL & 0x3F); 
}

void _vuFMAND(VURegs * VU) { 
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].US[0] = VU->VI[_Fs_].US[0] & VU->VI[REG_MAC_FLAG].US[0]; 
}

void _vuFMEQ(VURegs * VU) { 
	if(_Ft_ == 0) return; 
	if(VU->VI[REG_MAC_FLAG].US[0] == VU->VI[_Fs_].US[0]){ 
	VU->VI[_Ft_].US[0] =1;} else { VU->VI[_Ft_].US[0] =0; } 
}

void _vuFMOR(VURegs * VU) { 
	if(_Ft_ == 0) return; 
	VU->VI[_Ft_].US[0] = VU->VI[REG_MAC_FLAG].US[0] | VU->VI[_Fs_].US[0]; 
}

void _vuFCAND(VURegs * VU) {
	if(VU->VI[REG_CLIP_FLAG].UL & (VU->code & 0xFFFFFF)) VU->VI[1].UL = 1; 
	else VU->VI[1].UL = 0; 
}

void _vuFCEQ(VURegs * VU) { 
	if(VU->VI[REG_CLIP_FLAG].UL == (VU->code & 0xFFFFFF)) VU->VI[1].UL = 1; 
	else VU->VI[1].UL = 0; 
}

void _vuFCOR(VURegs * VU) { 
	u32 hold = VU->VI[REG_CLIP_FLAG].UL | ( VU->code & 0xFFFFFF); 
	if(hold == 0xFFFFFF) VU->VI[1].UL = 1; 
	else VU->VI[1].UL = 0; 
}

void _vuFCSET(VURegs * VU) { 
	VU->VI[REG_CLIP_FLAG].UL = (u32) (VU->code & 0xFFFFFF); 
}

void _vuFCGET(VURegs * VU) { 
	if(_Ft_ == 0) return;
	VU->VI[_Ft_].US[0] = VU->VI[REG_CLIP_FLAG].UL & 0x0FFF;
}

s32 _branchAddr(VURegs * VU) {
	s32 bpc = VU->VI[REG_TPC].SL + _Imm11_ * 8; 
	if (bpc < 0) {
		bpc = VU->VI[REG_TPC].SL + _UImm11_ * 8; 
	}
	if (VU == &VU1) {
		bpc&= 0x3fff;
	} else {
		bpc&= 0x0fff;
	}

	return bpc;
}

void _vuIBEQ(VURegs * VU) { 
	if (VU->VI[_Ft_].US[0] == VU->VI[_Fs_].US[0]) { 
		s32 bpc = _branchAddr(VU);

		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuIBGEZ(VURegs * VU) { 
	if (VU->VI[_Fs_].SS[0] >= 0) { 
		s32 bpc = _branchAddr(VU);
 
		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuIBGTZ(VURegs * VU) { 
	if (VU->VI[_Fs_].SS[0] > 0) { 
		s32 bpc = _branchAddr(VU);
 
		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuIBLEZ(VURegs * VU) { 
	if (VU->VI[_Fs_].SS[0] <= 0) { 
		s32 bpc = _branchAddr(VU);

		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuIBLTZ(VURegs * VU) { 
	if (VU->VI[_Fs_].SS[0] < 0) { 
		s32 bpc = _branchAddr(VU);

		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuIBNE(VURegs * VU) { 
	if (VU->VI[_Ft_].US[0] != VU->VI[_Fs_].US[0]) { 
		s32 bpc = _branchAddr(VU);
 
		VU->vuExec(VU);
	 	VU->VI[REG_TPC].UL = bpc; 
	} 
}

void _vuB(VURegs * VU) { 
	s32 bpc = _branchAddr(VU);
    
	VU->vuExec(VU);
	VU->VI[REG_TPC].UL = bpc; 
}

void _vuBAL(VURegs * VU) { 
	s32 bpc = _branchAddr(VU);

	if (_Ft_) {
		VU->VI[_Ft_].US[0] = VU->VI[REG_TPC].UL + 8;
	}

	VU->vuExec(VU);
 	VU->VI[REG_TPC].UL = bpc; 
}

void _vuJR(VURegs * VU) { 
	u32 bpc = VU->VI[_Fs_].US[0] & ~0x7; 

	VU->vuExec(VU);
 	VU->VI[REG_TPC].UL = bpc; 
}

void _vuJALR(VURegs * VU) { 
//    u32 bpc = VU->VI[_Fs_].US[0] & ~0x7; 
    u32 bpc = VU->VI[_Fs_].US[0] * 8; /* Aura uses it this way, wth? (linuz) */
	if (_Ft_) {
		VU->VI[_Ft_].US[0] = VU->VI[REG_TPC].UL + 8;
	}
 
	VU->vuExec(VU);
 	VU->VI[REG_TPC].UL = bpc; 
}

void _vuMFP(VURegs * VU) { 
	if (_Ft_ == 0) return; 
 
	if (_X) VU->VF[_Ft_].f.x = VU->VI[REG_P].F; 
	if (_Y) VU->VF[_Ft_].f.y = VU->VI[REG_P].F; 
	if (_Z) VU->VF[_Ft_].f.z = VU->VI[REG_P].F; 
	if (_W) VU->VF[_Ft_].f.w = VU->VI[REG_P].F; 
}

void _vuWAITP(VURegs * VU) { 
}

void _vuESADD(VURegs * VU) { 
	VU->VI[REG_P].F = VU->VF[_Fs_].f.x * VU->VF[_Fs_].f.x + VU->VF[_Fs_].f.y * VU->VF[_Fs_].f.y + VU->VF[_Fs_].f.z * VU->VF[_Fs_].f.z; 
}

void _vuERSADD(VURegs * VU) { 
	VU->VI[REG_P].F = (VU->VF[_Fs_].f.x * VU->VF[_Fs_].f.x) + (VU->VF[_Fs_].f.y * VU->VF[_Fs_].f.y) + (VU->VF[_Fs_].f.z * VU->VF[_Fs_].f.z); 
	if (VU->VI[REG_P].F != 0.0)
		VU->VI[REG_P].F = 1.0f / VU->VI[REG_P].F; 
}

/* Fixed. Could have caused crash due to value being -ve for sqrt *asadr */
void _vuELENG(VURegs * VU) { 
	float temp = VU->VF[_Fs_].f.x * VU->VF[_Fs_].f.x + VU->VF[_Fs_].f.y * VU->VF[_Fs_].f.y + VU->VF[_Fs_].f.z * VU->VF[_Fs_].f.z; 
	if(temp >= 0){ 
		VU->VI[REG_P].F = (float) sqrt(temp); 
	} 
}

/* Fixed. Could have caused crash due to divisor being = 0 *asadr */
void _vuERLENG(VURegs * VU) { 
	float temp = VU->VF[_Fs_].f.x * VU->VF[_Fs_].f.x + VU->VF[_Fs_].f.y * VU->VF[_Fs_].f.y + VU->VF[_Fs_].f.z * VU->VF[_Fs_].f.z; 
	if(temp >=0){ 
		if(sqrt(temp) != 0){ 
			VU->VI[REG_P].F = 1.0f / (float) sqrt(temp); 
		} 
	} 
}

/* Fixed. Could have caused crash due to divisor being = 0 *asadr */
void _vuEATANxy(VURegs * VU) { 
	if(VU->VF[_Fs_].f.x != 0) { 
		VU->VI[REG_P].F = atan(VU->VF[_Fs_].f.y / VU->VF[_Fs_].f.x); 
	} 
}

/* Fixed. Could have caused crash due to divisor being = 0 *asadr */
void _vuEATANxz(VURegs * VU) { 
	if(VU->VF[_Fs_].f.x != 0) { 
		VU->VI[REG_P].F = atan(VU->VF[_Fs_].f.z / VU->VF[_Fs_].f.x); 
	} 
}

void _vuESUM(VURegs * VU) { 
	VU->VI[REG_P].F = VU->VF[_Fs_].f.x + VU->VF[_Fs_].f.y + VU->VF[_Fs_].f.z + VU->VF[_Fs_].f.w; 
}

/* Fixed. Could have caused crash due to divisor being = 0 *asadr */
void _vuERCPR(VURegs * VU) { 
	if(VU->VF[_Fs_].F[_Fsf_] != 0){
		VU->VI[REG_P].F = 1.0f / (float)VU->VF[_Fs_].F[_Fsf_]; 
	}
}

/* Fixed. Could have caused crash due to Value being -ve for sqrt *asadr */
void _vuESQRT(VURegs * VU) { 
	if(VU->VF[_Fs_].F[_Fsf_] >= 0){
		VU->VI[REG_P].F = (float)sqrt(VU->VF[_Fs_].F[_Fsf_]); 
	}
}

/* Fixed. Could have caused crash due to divisor being = 0 *asadr */
void _vuERSQRT(VURegs * VU) { 
	if(VU->VF[_Fs_].F[_Fsf_] >= 0){
		if(sqrt(VU->VF[_Fs_].F[_Fsf_])){
		VU->VI[REG_P].F = 1.0f / (float)sqrt(VU->VF[_Fs_].F[_Fsf_]); 
		}
	}
}

void _vuESIN(VURegs * VU) { 
	VU->VI[REG_P].F = (float)sin(VU->VF[_Fs_].F[_Fsf_]); 
}

void _vuEATAN(VURegs * VU) { 
	VU->VI[REG_P].F = (float)atan(VU->VF[_Fs_].F[_Fsf_]); 
}

void _vuEEXP(VURegs * VU) { 
	VU->VI[REG_P].F = (float)exp(-(VU->VF[_Fs_].F[_Fsf_])); 
}

void _vuXITOP(VURegs * VU) { 
	if (_Ft_ == 0) return;
	VU->VI[_Ft_].US[0] = VU->vifRegs->itop;
}

void _vuXGKICK(VURegs * VU) {
	u16 addr = (VU->VI[_Fs_].US[0]*16) & 0x3fff;
	GSgifTransfer1(VU->Mem, addr);
}

void _vuXTOP(VURegs * VU) {
	if(_Ft_ == 0) return;
	VU->VI[_Ft_].US[0] = (u16)VU->vifRegs->top;
}
