/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <string.h>

#include "Common.h"
#include "ix86/ix86.h"
#include "Vif.h"
#include "VU0.h"
#include "VUmicro.h"

//#define VIF_OPTIMIZE

VIFregisters *_vifRegs;
int cl = 0;
int wl = 0;

__inline static int _limit( int a, int max ) 
{
	return ( a > max ? max : a );
}

#define _UNPACKpart( offnum, func )                         \
	if ( ( size > 0 ) && ( _vifRegs->offset == offnum ) ) { \
		func;                                               \
		size--;                                             \
		_vifRegs->offset++;                                  \
	}


void UNPACK_S_32(u32 *dest, u32 *data) {
#ifndef VIF_OPTIMIZE
	dest[0] = data[0];
	dest[1] = data[0];
	dest[2] = data[0];
	dest[3] = data[0];

	dest[4] = data[1];
	dest[5] = data[1];
	dest[6] = data[1];
	dest[7] = data[1];

	dest[8] = data[2];
	dest[9] = data[2];
	dest[10] = data[2];
	dest[11] = data[2];
#elif defined( __WIN32__ )

   if ( hasStreamingSIMDExtensions )
   {
      _asm
      {
         mov esi, data
         mov edi, dest
         movq mm0, [esi]
         movq mm1, [esi]
         movd mm2, [esi + 8]
         punpckldq  mm0, mm0     // mm0 = data[0]|data[0]
         punpckhdq  mm1, mm1     // mm1 = data[1]|data[1]
         punpckldq  mm2, mm2     // mm2 = data[2]|data[2]
         movntq [edi], mm0
         movntq [edi + 8], mm0
         movntq [edi + 16], mm1
         movntq [edi + 24], mm1
         movntq [edi + 32], mm2
         movntq [edi + 40], mm2
         emms
      }
   }
   else
   {
      _asm
      {
         mov esi, data
         mov edi, dest
         movq mm0, [esi]
         movq mm1, [esi]
         movd mm2, [esi + 8]
         punpckldq  mm0, mm0     // mm0 = data[0]|data[0]
         punpckhdq  mm1, mm1     // mm1 = data[1]|data[1]
         punpckldq  mm2, mm2     // mm2 = data[2]|data[2]
         movq [edi], mm0
         movq [edi + 8], mm0
         movq [edi + 16], mm1
         movq [edi + 24], mm1
         movq [edi + 32], mm2
         movq [edi + 40], mm2
         emms
      }
   }

#endif
}

void UNPACK_S_32_O( u32 *dest, u32 *data ) {
	dest[0] = data[0] + _vifRegs->r0;
	dest[1] = data[0] + _vifRegs->r1;
	dest[2] = data[0] + _vifRegs->r2;
	dest[3] = data[0] + _vifRegs->r3;

	dest[4] = data[1] + _vifRegs->r0;
	dest[5] = data[1] + _vifRegs->r1;
	dest[6] = data[1] + _vifRegs->r2;
	dest[7] = data[1] + _vifRegs->r3;

	dest[8] = data[2] + _vifRegs->r0;
	dest[9] = data[2] + _vifRegs->r1;
	dest[10] = data[2] + _vifRegs->r2;
	dest[11] = data[2] + _vifRegs->r3;
}

void UNPACK_S_32_D( u32 *dest, u32 *data ) {
	dest[0] = data[0] + _vifRegs->r0*2;
	dest[1] = data[0] + _vifRegs->r1*2;
	dest[2] = data[0] + _vifRegs->r2*2;
	dest[3] = data[0] + _vifRegs->r3*2;

	dest[4] = data[1] + _vifRegs->r0*2;
	dest[5] = data[1] + _vifRegs->r1*2;
	dest[6] = data[1] + _vifRegs->r2*2;
	dest[7] = data[1] + _vifRegs->r3*2;

	dest[8] = data[2] + _vifRegs->r0*2;
	dest[9] = data[2] + _vifRegs->r1*2;
	dest[10] = data[2] + _vifRegs->r2*2;
	dest[11] = data[2] + _vifRegs->r3*2;
}

void UNPACK_S_32part(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  
                *dest++ = *data;
				*dest++ = *data;
				*dest++ = *data;
				*dest++ = *data++; );

	_UNPACKpart(1,  
                *dest++ = *data;
				*dest++ = *data;
				*dest++ = *data;
				*dest++ = *data++; );

	_UNPACKpart(2,  
                *dest++ = *data;
				*dest++ = *data;
				*dest++ = *data;
				*dest++ = *data++; );
}

void UNPACK_S_32part_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  
                *dest++ = *data + _vifRegs->r0;
				*dest++ = *data + _vifRegs->r1;
				*dest++ = *data + _vifRegs->r2;
				*dest++ = *data++ + _vifRegs->r3; );

	_UNPACKpart(1,  
                *dest++ = *data + _vifRegs->r0;
				*dest++ = *data + _vifRegs->r1;
				*dest++ = *data + _vifRegs->r2;
				*dest++ = *data++ + _vifRegs->r3; );

	_UNPACKpart(2,  
                *dest++ = *data + _vifRegs->r0;
				*dest++ = *data + _vifRegs->r1;
				*dest++ = *data + _vifRegs->r2;
				*dest++ = *data++ + _vifRegs->r3; );
}

void UNPACK_S_32part_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  
                *dest++ = *data + _vifRegs->r0*2;
				*dest++ = *data + _vifRegs->r1*2;
				*dest++ = *data + _vifRegs->r2*2;
				*dest++ = *data++ + _vifRegs->r3*2; );

	_UNPACKpart(1,  
                *dest++ = *data + _vifRegs->r0*2;
				*dest++ = *data + _vifRegs->r1*2;
				*dest++ = *data + _vifRegs->r2*2;
				*dest++ = *data++ + _vifRegs->r3*2; );

	_UNPACKpart(2,  
                *dest++ = *data + _vifRegs->r0*2;
				*dest++ = *data + _vifRegs->r1*2;
				*dest++ = *data + _vifRegs->r2*2;
				*dest++ = *data++ + _vifRegs->r3*2; );
}

#define _UNPACK_S_16(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0]; \
	dest[1]  = sdata[0]; \
	dest[2]  = sdata[0]; \
	dest[3]  = sdata[0]; \
 \
	dest[4]  = sdata[1]; \
	dest[5]  = sdata[1]; \
	dest[6]  = sdata[1]; \
	dest[7]  = sdata[1]; \
 \
	dest[8]  = sdata[2]; \
	dest[9]  = sdata[2]; \
	dest[10] = sdata[2]; \
	dest[11] = sdata[2];

void UNPACK_S_16s( u32 *dest, u32 *data ) {
	_UNPACK_S_16( s16 );
}

void UNPACK_S_16u( u32 *dest, u32 *data ) {
	_UNPACK_S_16( u16 );
}

#define _UNPACK_S_16_O(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0; \
	dest[1]  = sdata[0] + _vifRegs->r1; \
	dest[2]  = sdata[0] + _vifRegs->r2; \
	dest[3]  = sdata[0] + _vifRegs->r3; \
 \
	dest[4]  = sdata[1] + _vifRegs->r0; \
	dest[5]  = sdata[1] + _vifRegs->r1; \
	dest[6]  = sdata[1] + _vifRegs->r2; \
	dest[7]  = sdata[1] + _vifRegs->r3; \
 \
	dest[8]  = sdata[2] + _vifRegs->r0; \
	dest[9]  = sdata[2] + _vifRegs->r1; \
	dest[10] = sdata[2] + _vifRegs->r2; \
	dest[11] = sdata[2] + _vifRegs->r3;

void UNPACK_S_16s_O( u32 *dest, u32 *data ) {
	_UNPACK_S_16_O( s16 );
}

void UNPACK_S_16u_O( u32 *dest, u32 *data ) {
	_UNPACK_S_16_O( u16 );
}

#define _UNPACK_S_16_D(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0*2; \
	dest[1]  = sdata[0] + _vifRegs->r1*2; \
	dest[2]  = sdata[0] + _vifRegs->r2*2; \
	dest[3]  = sdata[0] + _vifRegs->r3*2; \
 \
	dest[4]  = sdata[1] + _vifRegs->r0*2; \
	dest[5]  = sdata[1] + _vifRegs->r1*2; \
	dest[6]  = sdata[1] + _vifRegs->r2*2; \
	dest[7]  = sdata[1] + _vifRegs->r3*2; \
 \
	dest[8]  = sdata[2] + _vifRegs->r0*2; \
	dest[9]  = sdata[2] + _vifRegs->r1*2; \
	dest[10] = sdata[2] + _vifRegs->r2*2; \
	dest[11] = sdata[2] + _vifRegs->r3*2;

void UNPACK_S_16s_D( u32 *dest, u32 *data ) {
	_UNPACK_S_16_D( s16 );
}

void UNPACK_S_16u_D( u32 *dest, u32 *data ) {
	_UNPACK_S_16_D( u16 );
}

#define _UNPACK_S_16part(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata++; \
					*dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata; \
					*dest++ = *sdata; );

void UNPACK_S_16spart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part(s16);
}

void UNPACK_S_16upart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part(u16);
}

#define _UNPACK_S_16part_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata + _vifRegs->r0; \
					*dest++ = *sdata + _vifRegs->r1; \
					*dest++ = *sdata + _vifRegs->r2; \
					*dest++ = *sdata++ + _vifRegs->r3; \
					*dest++ = *sdata + _vifRegs->r0; \
					*dest++ = *sdata + _vifRegs->r1; \
					*dest++ = *sdata + _vifRegs->r2; \
					*dest++ = *sdata++ + _vifRegs->r3; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata + _vifRegs->r0; \
					*dest++ = *sdata + _vifRegs->r1; \
					*dest++ = *sdata + _vifRegs->r2; \
					*dest++ = *sdata + _vifRegs->r3; );

void UNPACK_S_16spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part_O(s16);
}

void UNPACK_S_16upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part_O(u16);
}

#define _UNPACK_S_16part_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata + _vifRegs->r0*2; \
					*dest++ = *sdata + _vifRegs->r1*2; \
					*dest++ = *sdata + _vifRegs->r2*2; \
					*dest++ = *sdata++ + _vifRegs->r3*2; \
					*dest++ = *sdata + _vifRegs->r0*2; \
					*dest++ = *sdata + _vifRegs->r1*2; \
					*dest++ = *sdata + _vifRegs->r2*2; \
					*dest++ = *sdata++ + _vifRegs->r3*2; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata + _vifRegs->r0*2; \
					*dest++ = *sdata + _vifRegs->r1*2; \
					*dest++ = *sdata + _vifRegs->r2*2; \
					*dest++ = *sdata + _vifRegs->r3*2; );

void UNPACK_S_16spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part_D(s16);
}

void UNPACK_S_16upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16part_D(u16);
}

#define _UNPACK_S_8(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0]; \
	dest[1]  = cdata[0]; \
	dest[2]  = cdata[0]; \
	dest[3]  = cdata[0]; \
 \
	dest[4]  = cdata[1]; \
	dest[5]  = cdata[1]; \
	dest[6]  = cdata[1]; \
	dest[7]  = cdata[1]; \
 \
	dest[8]  = cdata[2]; \
	dest[9]  = cdata[2]; \
	dest[10] = cdata[2]; \
	dest[11] = cdata[2];

void UNPACK_S_8s(u32 *dest, u32 *data) {
	_UNPACK_S_8(s8);
}

void UNPACK_S_8u(u32 *dest, u32 *data) {
	_UNPACK_S_8(u8);
}

void UNPACK_S_8spart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8(s8);
}

void UNPACK_S_8upart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8(u8);
}

#define _UNPACK_S_8_O(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0; \
	dest[1]  = cdata[0] + _vifRegs->r1; \
	dest[2]  = cdata[0] + _vifRegs->r2; \
	dest[3]  = cdata[0] + _vifRegs->r3; \
 \
	dest[4]  = cdata[1] + _vifRegs->r0; \
	dest[5]  = cdata[1] + _vifRegs->r1; \
	dest[6]  = cdata[1] + _vifRegs->r2; \
	dest[7]  = cdata[1] + _vifRegs->r3; \
 \
	dest[8]  = cdata[2] + _vifRegs->r0; \
	dest[9]  = cdata[2] + _vifRegs->r1; \
	dest[10] = cdata[2] + _vifRegs->r2; \
	dest[11] = cdata[2] + _vifRegs->r3;

void UNPACK_S_8s_O(u32 *dest, u32 *data) {
	_UNPACK_S_8_O(s8);
}

void UNPACK_S_8u_O(u32 *dest, u32 *data) {
	_UNPACK_S_8_O(u8);
}

void UNPACK_S_8spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8_O(s8);
}

void UNPACK_S_8upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8_O(u8);
}

#define _UNPACK_S_8_D(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0*2; \
	dest[1]  = cdata[0] + _vifRegs->r1*2; \
	dest[2]  = cdata[0] + _vifRegs->r2*2; \
	dest[3]  = cdata[0] + _vifRegs->r3*2; \
 \
	dest[4]  = cdata[1] + _vifRegs->r0*2; \
	dest[5]  = cdata[1] + _vifRegs->r1*2; \
	dest[6]  = cdata[1] + _vifRegs->r2*2; \
	dest[7]  = cdata[1] + _vifRegs->r3*2; \
 \
	dest[8]  = cdata[2] + _vifRegs->r0*2; \
	dest[9]  = cdata[2] + _vifRegs->r1*2; \
	dest[10] = cdata[2] + _vifRegs->r2*2; \
	dest[11] = cdata[2] + _vifRegs->r3*2;

void UNPACK_S_8s_D(u32 *dest, u32 *data) {
	_UNPACK_S_8_D(s8);
}

void UNPACK_S_8u_D(u32 *dest, u32 *data) {
	_UNPACK_S_8_D(u8);
}

void UNPACK_S_8spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8_D(s8);
}

void UNPACK_S_8upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8_D(u8);
}

void UNPACK_V2_32( u32 *dest, u32 *data ) {
	dest[0] = data[0];
	dest[1] = data[1];
	dest[4] = data[2];
	dest[5] = data[3];
	dest[8] = data[4];
	dest[9] = data[5];
}

void UNPACK_V2_32_O( u32 *dest, u32 *data ) {
	dest[0] = data[0] + _vifRegs->r0;
	dest[1] = data[1] + _vifRegs->r1;
	dest[4] = data[2] + _vifRegs->r0;
	dest[5] = data[3] + _vifRegs->r1;
	dest[8] = data[4] + _vifRegs->r0;
	dest[9] = data[5] + _vifRegs->r1;
}

void UNPACK_V2_32_D( u32 *dest, u32 *data ) {
	dest[0] = data[0] + _vifRegs->r0*2;
	dest[1] = data[1] + _vifRegs->r1*2;
	dest[4] = data[2] + _vifRegs->r0*2;
	dest[5] = data[3] + _vifRegs->r1*2;
	dest[8] = data[4] + _vifRegs->r0*2;
	dest[9] = data[5] + _vifRegs->r1*2;
}

void UNPACK_V2_32part( u32 *dest, u32 *data, int size ) {
	_UNPACKpart(0,  *dest++ = *data++; );
	_UNPACKpart(1,  *dest++ = *data++; );
	_UNPACKpart(3,  *dest++ = *data++; );
	_UNPACKpart(4,  *dest++ = *data++; );
	_UNPACKpart(5,  *dest++ = *data++; );
	_UNPACKpart(6,  *dest = *data;  );
}

void UNPACK_V2_32part_O( u32 *dest, u32 *data, int size ) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(6,  *dest = *data + _vifRegs->r1;  );
}

void UNPACK_V2_32part_D( u32 *dest, u32 *data, int size ) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(6,  *dest = *data + _vifRegs->r1*2;  );
}

#define _UNPACK_V2_16(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0]; \
	dest[1]  = sdata[1]; \
	dest[4]  = sdata[2]; \
	dest[5]  = sdata[3]; \
	dest[8]  = sdata[4]; \
	dest[9]  = sdata[5];

void UNPACK_V2_16s(u32 *dest, u32 *data) {
	_UNPACK_V2_16(s16);
}

void UNPACK_V2_16u(u32 *dest, u32 *data) {
	_UNPACK_V2_16(u16);
}

#define _UNPACK_V2_16_O(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0; \
	dest[1]  = sdata[1] + _vifRegs->r1; \
	dest[4]  = sdata[2] + _vifRegs->r0; \
	dest[5]  = sdata[3] + _vifRegs->r1; \
	dest[8]  = sdata[4] + _vifRegs->r0; \
	dest[9]  = sdata[5] + _vifRegs->r1;

void UNPACK_V2_16s_O(u32 *dest, u32 *data) {
	_UNPACK_V2_16_O(s16);
}

void UNPACK_V2_16u_O(u32 *dest, u32 *data) {
	_UNPACK_V2_16_O(u16);
}

#define _UNPACK_V2_16_D(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0*2; \
	dest[1]  = sdata[1] + _vifRegs->r1*2; \
	dest[4]  = sdata[2] + _vifRegs->r0*2; \
	dest[5]  = sdata[3] + _vifRegs->r1*2; \
	dest[8]  = sdata[4] + _vifRegs->r0*2; \
	dest[9]  = sdata[5] + _vifRegs->r1*2;

void UNPACK_V2_16s_D(u32 *dest, u32 *data) {
	_UNPACK_V2_16_D(s16);
}

void UNPACK_V2_16u_D(u32 *dest, u32 *data) {
	_UNPACK_V2_16_D(u16);
}

#define _UNPACK_V2_16part(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
	_UNPACKpart(1,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
	_UNPACKpart(2,  *dest++ = *sdata++; \
					*dest++   = *sdata; );

void UNPACK_V2_16spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part(s16);
}

void UNPACK_V2_16upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part(u16);
}

#define _UNPACK_V2_16part_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
	_UNPACKpart(1,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++   = *sdata + _vifRegs->r1; );

void UNPACK_V2_16spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part_O(s16);
}

void UNPACK_V2_16upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part_O(u16);
}

#define _UNPACK_V2_16part_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
	_UNPACKpart(1,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++   = *sdata + _vifRegs->r1*2; );

void UNPACK_V2_16spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part_D(s16);
}

void UNPACK_V2_16upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16part_D(u16);
}

#define _UNPACK_V2_8(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0]; \
	dest[1]  = cdata[1]; \
	dest[4]  = cdata[2]; \
	dest[5]  = cdata[3]; \
	dest[8]  = cdata[4]; \
	dest[9]  = cdata[5];

void UNPACK_V2_8s(u32 *dest, u32 *data) {
	_UNPACK_V2_8(s8);
}

void UNPACK_V2_8u(u32 *dest, u32 *data) {
	_UNPACK_V2_8(u8);
}

#define _UNPACK_V2_8_O(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0; \
	dest[1]  = cdata[1] + _vifRegs->r1; \
	dest[4]  = cdata[2] + _vifRegs->r0; \
	dest[5]  = cdata[3] + _vifRegs->r1; \
	dest[8]  = cdata[4] + _vifRegs->r0; \
	dest[9]  = cdata[5] + _vifRegs->r1;

void UNPACK_V2_8s_O(u32 *dest, u32 *data) {
	_UNPACK_V2_8_O(s8);
}

void UNPACK_V2_8u_O(u32 *dest, u32 *data) {
	_UNPACK_V2_8_O(u8);
}

#define _UNPACK_V2_8_D(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0*2; \
	dest[1]  = cdata[1] + _vifRegs->r1*2; \
	dest[4]  = cdata[2] + _vifRegs->r0*2; \
	dest[5]  = cdata[3] + _vifRegs->r1*2; \
	dest[8]  = cdata[4] + _vifRegs->r0*2; \
	dest[9]  = cdata[5] + _vifRegs->r1*2;

void UNPACK_V2_8s_D(u32 *dest, u32 *data) {
	_UNPACK_V2_8_D(s8);
}

void UNPACK_V2_8u_D(u32 *dest, u32 *data) {
	_UNPACK_V2_8_D(u8);
}

#define _UNPACK_V2_8part(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++; \
					*dest   = *cdata++; dest+= 3; \
					*dest++ = *cdata++; \
					*dest   = *cdata++; dest+= 3; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++; \
					*dest   = *cdata; );

void UNPACK_V2_8spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part(s8);
}

void UNPACK_V2_8upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part(u8);
}

#define _UNPACK_V2_8part_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest   = *cdata++ + _vifRegs->r1; dest+= 3; \
					*dest++ = *cdata++ + _vifRegs->r0; \
					*dest   = *cdata++ + _vifRegs->r1; dest+= 3; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest   = *cdata + _vifRegs->r1; );

void UNPACK_V2_8spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part_O(s8);
}

void UNPACK_V2_8upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part_O(u8);
}

#define _UNPACK_V2_8part_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest   = *cdata++ + _vifRegs->r1*2; dest+= 3; \
					*dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest   = *cdata++ + _vifRegs->r1*2; dest+= 3; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest   = *cdata + _vifRegs->r1*2; );

void UNPACK_V2_8spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part_D(s8);
}

void UNPACK_V2_8upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8part_D(u8);
}

void UNPACK_V3_32(u32 *dest, u32 *data) {
	dest[0]  = data[0];
	dest[1]  = data[1];
	dest[2]  = data[2];
	dest[4]  = data[3];
	dest[5]  = data[4];
	dest[6]  = data[5];
	dest[8]  = data[6];
	dest[9]  = data[7];
	dest[10] = data[8];
}

void UNPACK_V3_32_O(u32 *dest, u32 *data) {
	dest[0]  = data[0] + _vifRegs->r0;
	dest[1]  = data[1] + _vifRegs->r1;
	dest[2]  = data[2] + _vifRegs->r2;
	dest[4]  = data[3] + _vifRegs->r0;
	dest[5]  = data[4] + _vifRegs->r1;
	dest[6]  = data[5] + _vifRegs->r2;
	dest[8]  = data[6] + _vifRegs->r0;
	dest[9]  = data[7] + _vifRegs->r1;
	dest[10] = data[8] + _vifRegs->r2;
}

void UNPACK_V3_32_D(u32 *dest, u32 *data) {
	dest[0]  = data[0] + _vifRegs->r0*2;
	dest[1]  = data[1] + _vifRegs->r1*2;
	dest[2]  = data[2] + _vifRegs->r2*2;
	dest[4]  = data[3] + _vifRegs->r0*2;
	dest[5]  = data[4] + _vifRegs->r1*2;
	dest[6]  = data[5] + _vifRegs->r2*2;
	dest[8]  = data[6] + _vifRegs->r0*2;
	dest[9]  = data[7] + _vifRegs->r1*2;
	dest[10] = data[8] + _vifRegs->r2*2;
}

void UNPACK_V3_32part(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++; );
	_UNPACKpart(1,  *dest++ = *data++; );
	_UNPACKpart(2,  *dest++ = *data++; );
    _UNPACKpart(3,  *dest++ = *data++; );
	_UNPACKpart(4,  *dest++ = *data++; );
	_UNPACKpart(5,  *dest++ = *data++; );
	_UNPACKpart(6,  *dest++ = *data++; );
	_UNPACKpart(7,  *dest++ = *data++; );
	_UNPACKpart(8,  *dest = *data; );
}

void UNPACK_V3_32part_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(2,  *dest++ = *data++ + _vifRegs->r2; );
    _UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r2; );
	_UNPACKpart(6,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(7,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(8,  *dest = *data + _vifRegs->r2; );
}

void UNPACK_V3_32part_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(2,  *dest++ = *data++ + _vifRegs->r2*2; );
    _UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r2*2; );
	_UNPACKpart(6,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(7,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(8,  *dest = *data + _vifRegs->r2*2; );
}

#define _UNPACK_V3_16(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0]; \
	dest[1]  = sdata[1]; \
	dest[2]  = sdata[2]; \
	dest[4]  = sdata[3]; \
	dest[5]  = sdata[4]; \
	dest[6]  = sdata[5]; \
	dest[8]  = sdata[6]; \
	dest[9]  = sdata[7]; \
	dest[10] = sdata[8];

void UNPACK_V3_16s(u32 *dest, u32 *data) {
	_UNPACK_V3_16(s16);
}

void UNPACK_V3_16u(u32 *dest, u32 *data) {
	_UNPACK_V3_16(u16);
}

#define _UNPACK_V3_16_O(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0; \
	dest[1]  = sdata[1] + _vifRegs->r1; \
	dest[2]  = sdata[2] + _vifRegs->r2; \
	dest[4]  = sdata[3] + _vifRegs->r0; \
	dest[5]  = sdata[4] + _vifRegs->r1; \
	dest[6]  = sdata[5] + _vifRegs->r2; \
	dest[8]  = sdata[6] + _vifRegs->r0; \
	dest[9]  = sdata[7] + _vifRegs->r1; \
	dest[10] = sdata[8] + _vifRegs->r2;

void UNPACK_V3_16s_O(u32 *dest, u32 *data) {
	_UNPACK_V3_16_O(s16);
}

void UNPACK_V3_16u_O(u32 *dest, u32 *data) {
	_UNPACK_V3_16_O(u16);
}

#define _UNPACK_V3_16_D(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0*2; \
	dest[1]  = sdata[1] + _vifRegs->r1*2; \
	dest[2]  = sdata[2] + _vifRegs->r2*2; \
	dest[4]  = sdata[3] + _vifRegs->r0*2; \
	dest[5]  = sdata[4] + _vifRegs->r1*2; \
	dest[6]  = sdata[5] + _vifRegs->r2*2; \
	dest[8]  = sdata[6] + _vifRegs->r0*2; \
	dest[9]  = sdata[7] + _vifRegs->r1*2; \
	dest[10] = sdata[8] + _vifRegs->r2*2;

void UNPACK_V3_16s_D(u32 *dest, u32 *data) {
	_UNPACK_V3_16_D(s16);
}

void UNPACK_V3_16u_D(u32 *dest, u32 *data) {
	_UNPACK_V3_16_D(u16);
}

#define _UNPACK_V3_16part(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(1,  *dest   = *sdata++; dest+= 2; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++; \
					*dest   = *sdata++; dest+= 2; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(4,  *dest   = *sdata;   );

void UNPACK_V3_16spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part(s16);
}

void UNPACK_V3_16upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part(u16);
}

#define _UNPACK_V3_16part_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(1,  *dest   = *sdata++ + _vifRegs->r2; dest+= 2; \
					*dest++ = *sdata++ + _vifRegs->r0; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r1; \
					*dest   = *sdata++ + _vifRegs->r2; dest+= 2; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(4,  *dest   = *sdata + _vifRegs->r2;   );

void UNPACK_V3_16spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part_O(s16);
}

void UNPACK_V3_16upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part_O(u16);
}

#define _UNPACK_V3_16part_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(1,  *dest   = *sdata++ + _vifRegs->r2*2; dest+= 2; \
					*dest++ = *sdata++ + _vifRegs->r0*2; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r1*2; \
					*dest   = *sdata++ + _vifRegs->r2*2; dest+= 2; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(4,  *dest   = *sdata + _vifRegs->r2*2;   );

void UNPACK_V3_16spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part_D(s16);
}

void UNPACK_V3_16upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16part_D(u16);
}

#define _UNPACK_V3_8(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0]; \
	dest[1]  = cdata[1]; \
	dest[2]  = cdata[2]; \
	dest[4]  = cdata[3]; \
	dest[5]  = cdata[4]; \
	dest[6]  = cdata[5]; \
	dest[8]  = cdata[6]; \
	dest[9]  = cdata[7]; \
	dest[10] = cdata[8];

void UNPACK_V3_8s(u32 *dest, u32 *data) {
	_UNPACK_V3_8(s8);
}

void UNPACK_V3_8u(u32 *dest, u32 *data) {
	_UNPACK_V3_8(u8);
}

#define _UNPACK_V3_8_O(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0; \
	dest[1]  = cdata[1] + _vifRegs->r1; \
	dest[2]  = cdata[2] + _vifRegs->r2; \
	dest[4]  = cdata[3] + _vifRegs->r0; \
	dest[5]  = cdata[4] + _vifRegs->r1; \
	dest[6]  = cdata[5] + _vifRegs->r2; \
	dest[8]  = cdata[6] + _vifRegs->r0; \
	dest[9]  = cdata[7] + _vifRegs->r1; \
	dest[10] = cdata[8] + _vifRegs->r2;

void UNPACK_V3_8s_O(u32 *dest, u32 *data) {
	_UNPACK_V3_8_O(s8);
}

void UNPACK_V3_8u_O(u32 *dest, u32 *data) {
	_UNPACK_V3_8_O(u8);
}

#define _UNPACK_V3_8_D(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0*2; \
	dest[1]  = cdata[1] + _vifRegs->r1*2; \
	dest[2]  = cdata[2] + _vifRegs->r2*2; \
	dest[4]  = cdata[3] + _vifRegs->r0*2; \
	dest[5]  = cdata[4] + _vifRegs->r1*2; \
	dest[6]  = cdata[5] + _vifRegs->r2*2; \
	dest[8]  = cdata[6] + _vifRegs->r0*2; \
	dest[9]  = cdata[7] + _vifRegs->r1*2; \
	dest[10] = cdata[8] + _vifRegs->r2*2;

void UNPACK_V3_8s_D(u32 *dest, u32 *data) {
	_UNPACK_V3_8_D(s8);
}

void UNPACK_V3_8u_D(u32 *dest, u32 *data) {
	_UNPACK_V3_8_D(u8);
}

#define _UNPACK_V3_8part(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest   = *cdata++; dest+= 2; \
					*dest++ = *cdata++; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++; \
					*dest   = *cdata++; dest+= 2; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; ); \
 \
	_UNPACKpart(2,  *dest   = *cdata;   );

void UNPACK_V3_8spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part(s8);
}

void UNPACK_V3_8upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part(u8);
}

#define _UNPACK_V3_8part_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest++ = *cdata++ + _vifRegs->r1; \
					*dest   = *cdata++ + _vifRegs->r2; dest+= 2; \
					*dest++ = *cdata++ + _vifRegs->r0; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r1; \
					*dest   = *cdata++ + _vifRegs->r2; dest+= 2; \
					*dest++ = *cdata++ + _vifRegs->r0; \
					*dest++ = *cdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(2,  *dest   = *cdata + _vifRegs->r2;   );

void UNPACK_V3_8spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part_O(s8);
}

void UNPACK_V3_8upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part_O(u8);
}

#define _UNPACK_V3_8part_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest++ = *cdata++ + _vifRegs->r1*2; \
					*dest   = *cdata++ + _vifRegs->r2*2; dest+= 2; \
					*dest++ = *cdata++ + _vifRegs->r0*2; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r1*2; \
					*dest   = *cdata++ + _vifRegs->r2*2; dest+= 2; \
					*dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest++ = *cdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(2,  *dest   = *cdata + _vifRegs->r2*2;   );

void UNPACK_V3_8spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part_D(s8);
}

void UNPACK_V3_8upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8part_D(u8);
}

void UNPACK_V4_32( u32 *dest, u32 *data ) {
	memcpy(dest, data, 12*4);
}

void UNPACK_V4_32_O( u32 *dest, u32 *data ) {
	dest[0]  = data[0] + _vifRegs->r0;
	dest[1]  = data[1] + _vifRegs->r1;
	dest[2]  = data[2] + _vifRegs->r2;
	dest[3]  = data[3] + _vifRegs->r3;
	dest[4]  = data[4] + _vifRegs->r0;
	dest[5]  = data[5] + _vifRegs->r1;
	dest[6]  = data[6] + _vifRegs->r2;
	dest[7]  = data[7] + _vifRegs->r3;
	dest[8]  = data[8] + _vifRegs->r0;
	dest[9]  = data[9] + _vifRegs->r1;
	dest[10] = data[10] + _vifRegs->r2;
	dest[11] = data[11] + _vifRegs->r3;
}

void UNPACK_V4_32_D( u32 *dest, u32 *data ) {
	dest[0]  = data[0] + _vifRegs->r0*2;
	dest[1]  = data[1] + _vifRegs->r1*2;
	dest[2]  = data[2] + _vifRegs->r2*2;
	dest[3]  = data[3] + _vifRegs->r3*2;
	dest[4]  = data[4] + _vifRegs->r0*2;
	dest[5]  = data[5] + _vifRegs->r1*2;
	dest[6]  = data[6] + _vifRegs->r2*2;
	dest[7]  = data[7] + _vifRegs->r3*2;
	dest[8]  = data[8] + _vifRegs->r0*2;
	dest[9]  = data[9] + _vifRegs->r1*2;
	dest[10] = data[10] + _vifRegs->r2*2;
	dest[11] = data[11] + _vifRegs->r3*2;
}

void UNPACK_V4_32part(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++; );
	_UNPACKpart(1,  *dest++ = *data++; );
	_UNPACKpart(2,  *dest++ = *data++; );
	_UNPACKpart(3,  *dest++ = *data++; );

	_UNPACKpart(4,  *dest++ = *data++; );
	_UNPACKpart(5,  *dest++ = *data++; );
	_UNPACKpart(6,  *dest++ = *data++; );
	_UNPACKpart(7,  *dest++ = *data++; );

	_UNPACKpart(8,  *dest++ = *data++; );
	_UNPACKpart(9,  *dest++ = *data++; );
	_UNPACKpart(10, *dest++ = *data++; );
	_UNPACKpart(11, *dest   = *data;   );
}

void UNPACK_V4_32part_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(2,  *dest++ = *data++ + _vifRegs->r2; );
	_UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r3; );

	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(6,  *dest++ = *data++ + _vifRegs->r2; );
	_UNPACKpart(7,  *dest++ = *data++ + _vifRegs->r3; );

	_UNPACKpart(8,  *dest++ = *data++ + _vifRegs->r0; );
	_UNPACKpart(9,  *dest++ = *data++ + _vifRegs->r1; );
	_UNPACKpart(10, *dest++ = *data++ + _vifRegs->r2; );
	_UNPACKpart(11, *dest   = *data + _vifRegs->r3;   );
}

void UNPACK_V4_32part_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(1,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(2,  *dest++ = *data++ + _vifRegs->r2*2; );
	_UNPACKpart(3,  *dest++ = *data++ + _vifRegs->r3*2; );

	_UNPACKpart(4,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(5,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(6,  *dest++ = *data++ + _vifRegs->r2*2; );
	_UNPACKpart(7,  *dest++ = *data++ + _vifRegs->r3*2; );

	_UNPACKpart(8,  *dest++ = *data++ + _vifRegs->r0*2; );
	_UNPACKpart(9,  *dest++ = *data++ + _vifRegs->r1*2; );
	_UNPACKpart(10, *dest++ = *data++ + _vifRegs->r2*2; );
	_UNPACKpart(11, *dest   = *data + _vifRegs->r3*2;   );
}

#define _UNPACK_V4_16(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0]; \
	dest[1]  = sdata[1]; \
	dest[2]  = sdata[2]; \
	dest[3]  = sdata[3]; \
 \
	dest[4]  = sdata[4]; \
	dest[5]  = sdata[5]; \
	dest[6]  = sdata[6]; \
	dest[7]  = sdata[7]; \
 \
	dest[8]  = sdata[8]; \
	dest[9]  = sdata[9]; \
	dest[10] = sdata[10]; \
	dest[11] = sdata[11];

void UNPACK_V4_16s(u32 *dest, u32 *data) {
	_UNPACK_V4_16(s16);
}

void UNPACK_V4_16u(u32 *dest, u32 *data) {
	_UNPACK_V4_16(u16);
}

#define _UNPACK_V4_16_O(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0; \
	dest[1]  = sdata[1] + _vifRegs->r1; \
	dest[2]  = sdata[2] + _vifRegs->r2; \
	dest[3]  = sdata[3] + _vifRegs->r3; \
 \
	dest[4]  = sdata[4] + _vifRegs->r0; \
	dest[5]  = sdata[5] + _vifRegs->r1; \
	dest[6]  = sdata[6] + _vifRegs->r2; \
	dest[7]  = sdata[7] + _vifRegs->r3; \
 \
	dest[8]  = sdata[8] + _vifRegs->r0; \
	dest[9]  = sdata[9] + _vifRegs->r1; \
	dest[10] = sdata[10] + _vifRegs->r2; \
	dest[11] = sdata[11] + _vifRegs->r3;

void UNPACK_V4_16s_O(u32 *dest, u32 *data) {
	_UNPACK_V4_16_O(s16);
}

void UNPACK_V4_16u_O(u32 *dest, u32 *data) {
	_UNPACK_V4_16_O(s16);
}

#define _UNPACK_V4_16_D(format) \
	format *sdata = (format*)data; \
	dest[0]  = sdata[0] + _vifRegs->r0*2; \
	dest[1]  = sdata[1] + _vifRegs->r1*2; \
	dest[2]  = sdata[2] + _vifRegs->r2*2; \
	dest[3]  = sdata[3] + _vifRegs->r3*2; \
 \
	dest[4]  = sdata[4] + _vifRegs->r0*2; \
	dest[5]  = sdata[5] + _vifRegs->r1*2; \
	dest[6]  = sdata[6] + _vifRegs->r2*2; \
	dest[7]  = sdata[7] + _vifRegs->r3*2; \
 \
	dest[8]  = sdata[8] + _vifRegs->r0*2; \
	dest[9]  = sdata[9] + _vifRegs->r1*2; \
	dest[10] = sdata[10] + _vifRegs->r2*2; \
	dest[11] = sdata[11] + _vifRegs->r3*2;

void UNPACK_V4_16s_D(u32 *dest, u32 *data) {
	_UNPACK_V4_16_D(s16);
}

void UNPACK_V4_16u_D(u32 *dest, u32 *data) {
	_UNPACK_V4_16_D(s16);
}

#define _UNPACK_V4_16part(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(4,  *dest++ = *sdata++; \
					*dest++ = *sdata++; ); \
 \
	_UNPACKpart(5,  *dest++ = *sdata++; \
					*dest   = *sdata  ; );

void UNPACK_V4_16spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part(s16);
}

void UNPACK_V4_16upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part(u16);
}

#define _UNPACK_V4_16part_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata++ + _vifRegs->r2; \
					*dest++ = *sdata++ + _vifRegs->r3; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++ + _vifRegs->r2; \
					*dest++ = *sdata++ + _vifRegs->r3; ); \
 \
	_UNPACKpart(4,  *dest++ = *sdata++ + _vifRegs->r0; \
					*dest++ = *sdata++ + _vifRegs->r1; ); \
 \
	_UNPACKpart(5,  *dest++ = *sdata++ + _vifRegs->r2; \
					*dest   = *sdata   + _vifRegs->r3; );

void UNPACK_V4_16spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part_O(s16);
}

void UNPACK_V4_16upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part_O(u16);
}

#define _UNPACK_V4_16part_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(1,  *dest++ = *sdata++ + _vifRegs->r2*2; \
					*dest++ = *sdata++ + _vifRegs->r3*2; ); \
 \
	_UNPACKpart(2,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(3,  *dest++ = *sdata++ + _vifRegs->r2*2; \
					*dest++ = *sdata++ + _vifRegs->r3*2; ); \
 \
	_UNPACKpart(4,  *dest++ = *sdata++ + _vifRegs->r0*2; \
					*dest++ = *sdata++ + _vifRegs->r1*2; ); \
 \
	_UNPACKpart(5,  *dest++ = *sdata++ + _vifRegs->r2*2; \
					*dest   = *sdata   + _vifRegs->r3*2; );

void UNPACK_V4_16spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part_D(s16);
}

void UNPACK_V4_16upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16part_D(u16);
}

#define _UNPACK_V4_8(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0]; \
	dest[1]  = cdata[1]; \
	dest[2]  = cdata[2]; \
	dest[3]  = cdata[3]; \
 \
	dest[4]  = cdata[4]; \
	dest[5]  = cdata[5]; \
	dest[6]  = cdata[6]; \
	dest[7]  = cdata[7]; \
 \
	dest[8]  = cdata[8]; \
	dest[9]  = cdata[9]; \
	dest[10] = cdata[10]; \
	dest[11] = cdata[11];

void UNPACK_V4_8s(u32 *dest, u32 *data) {
	_UNPACK_V4_8(s8);
}

void UNPACK_V4_8u(u32 *dest, u32 *data) {
	_UNPACK_V4_8(u8);
}

#define _UNPACK_V4_8_O(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0; \
	dest[1]  = cdata[1] + _vifRegs->r1; \
	dest[2]  = cdata[2] + _vifRegs->r2; \
	dest[3]  = cdata[3] + _vifRegs->r3; \
 \
	dest[4]  = cdata[4] + _vifRegs->r0; \
	dest[5]  = cdata[5] + _vifRegs->r1; \
	dest[6]  = cdata[6] + _vifRegs->r2; \
	dest[7]  = cdata[7] + _vifRegs->r3; \
 \
	dest[8]  = cdata[8] + _vifRegs->r0; \
	dest[9]  = cdata[9] + _vifRegs->r1; \
	dest[10] = cdata[10] + _vifRegs->r2; \
	dest[11] = cdata[11] + _vifRegs->r3;

void UNPACK_V4_8s_O(u32 *dest, u32 *data) {
	_UNPACK_V4_8_O(s8);
}

void UNPACK_V4_8u_O(u32 *dest, u32 *data) {
	_UNPACK_V4_8_O(u8);
}

#define _UNPACK_V4_8_D(format) \
	format *cdata = (format*)data; \
	dest[0]  = cdata[0] + _vifRegs->r0*2; \
	dest[1]  = cdata[1] + _vifRegs->r1*2; \
	dest[2]  = cdata[2] + _vifRegs->r2*2; \
	dest[3]  = cdata[3] + _vifRegs->r3*2; \
 \
	dest[4]  = cdata[4] + _vifRegs->r0*2; \
	dest[5]  = cdata[5] + _vifRegs->r1*2; \
	dest[6]  = cdata[6] + _vifRegs->r2*2; \
	dest[7]  = cdata[7] + _vifRegs->r3*2; \
 \
	dest[8]  = cdata[8] + _vifRegs->r0*2; \
	dest[9]  = cdata[9] + _vifRegs->r1*2; \
	dest[10] = cdata[10] + _vifRegs->r2*2; \
	dest[11] = cdata[11] + _vifRegs->r3*2;

void UNPACK_V4_8s_D(u32 *dest, u32 *data) {
	_UNPACK_V4_8_D(s8);
}

void UNPACK_V4_8u_D(u32 *dest, u32 *data) {
	_UNPACK_V4_8_D(u8);
}

#define _UNPACK_V4_8part(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; ); \
 \
	_UNPACKpart(2,  *dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest++ = *cdata++; \
					*dest   = *cdata  ; );

void UNPACK_V4_8spart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part(s8);
}

void UNPACK_V4_8upart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part(u8);
}

#define _UNPACK_V4_8part_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest++ = *cdata++ + _vifRegs->r1; \
					*dest++ = *cdata++ + _vifRegs->r2; \
					*dest++ = *cdata++ + _vifRegs->r3; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest++ = *cdata++ + _vifRegs->r1; \
					*dest++ = *cdata++ + _vifRegs->r2; \
					*dest++ = *cdata++ + _vifRegs->r3; ); \
 \
	_UNPACKpart(2,  *dest++ = *cdata++ + _vifRegs->r0; \
					*dest++ = *cdata++ + _vifRegs->r1; \
					*dest++ = *cdata++ + _vifRegs->r2; \
					*dest   = *cdata   + _vifRegs->r3; );

void UNPACK_V4_8spart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part_O(s8);
}

void UNPACK_V4_8upart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part_O(u8);
}

#define _UNPACK_V4_8part_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest++ = *cdata++ + _vifRegs->r1*2; \
					*dest++ = *cdata++ + _vifRegs->r2*2; \
					*dest++ = *cdata++ + _vifRegs->r3*2; ); \
 \
	_UNPACKpart(1,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest++ = *cdata++ + _vifRegs->r1*2; \
					*dest++ = *cdata++ + _vifRegs->r2*2; \
					*dest++ = *cdata++ + _vifRegs->r3*2; ); \
 \
	_UNPACKpart(2,  *dest++ = *cdata++ + _vifRegs->r0*2; \
					*dest++ = *cdata++ + _vifRegs->r1*2; \
					*dest++ = *cdata++ + _vifRegs->r2*2; \
					*dest   = *cdata   + _vifRegs->r3*2; );

void UNPACK_V4_8spart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part_D(s8);
}

void UNPACK_V4_8upart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8part_D(u8);
}

void UNPACK_V4_5(u32 *dest, u32 *data) {
	u32 rgba;

	rgba = data[0] & 0xffff; /* RGBA1 */
	dest[0] = (rgba & 0x1f) << 3;
	dest[1] = ((rgba >> 5)  & 0x1f) << 3;
	dest[2] = ((rgba >> 10) & 0x1f) << 3;
	dest[3] = ((rgba >> 15) & 0x1) << 7;

	rgba = (data[0] >> 16); /* RGBA2 */
	dest[4] = (rgba & 0x1f) << 3;
	dest[5] = ((rgba >> 5)  & 0x1f) << 3;
	dest[6] = ((rgba >> 10) & 0x1f) << 3;
	dest[7] = ((rgba >> 15) & 0x1) << 7;

	rgba = data[1] & 0xffff; /* RGBA3 */
	dest[8] = (rgba & 0x1f) << 3;
	dest[9] = ((rgba >> 5)  & 0x1f) << 3;
	dest[10] = ((rgba >> 10) & 0x1f) << 3;
	dest[11] = ((rgba >> 15) & 0x1) << 7;
}

void UNPACK_V4_5part(u32 *dest, u32 *data, int size) {
	u32 rgba;

	_UNPACKpart(0, 
		rgba = *data & 0xffff; /* RGBA1 */
		dest[0] = (rgba & 0x1f) << 3;
		dest[1] = ((rgba >> 5)  & 0x1f) << 3;
		dest[2] = ((rgba >> 10) & 0x1f) << 3;
		dest[3] = ((rgba >> 15) & 0x1) << 7;

		rgba = (*data++ >> 16); /* RGBA2 */
		dest[4] = (rgba & 0x1f) << 3;
		dest[5] = ((rgba >> 5)  & 0x1f) << 3;
		dest[6] = ((rgba >> 10) & 0x1f) << 3;
		dest[7] = ((rgba >> 15) & 0x1) << 7;
		dest+= 8;
	);

	_UNPACKpart(1, 
		rgba = *data & 0xffff; /* RGBA3 */
		dest[8] = (rgba & 0x1f) << 3;
		dest[9] = ((rgba >> 5)  & 0x1f) << 3;
		dest[10] = ((rgba >> 10) & 0x1f) << 3;
		dest[11] = ((rgba >> 15) & 0x1) << 7;
	);
}

// Tables for the optimized functions below (_writeX, _writeY, _writeZ, _writeW)
int maskDisplacementX[ 256 ] =
{
   0,   8, 16, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
   24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24
};

int maskDisplacementY[ 256 ] =
{
    2, 10, 18, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
   26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26, 26
};

int maskDisplacementZ[ 256 ] =
{
    4, 12, 20, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28,
   28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28
};

int maskDisplacementW[ 256 ] =
{
    6, 14, 22, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
   30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 30
};

static void _writeX( u32 *dest, u32 data ) 
{
#ifndef VIF_OPTIMIZE

   int n;
	switch ( cl )
   {
		case 0:  n =  0; break;
		case 1:  n =  8; break;
		case 2:  n = 16; break;
		default: n = 24; break;
	}
	switch ( ( _vifRegs->mask >> n ) & 0x3 )
   {
		case 0: *dest = data;         break;
		case 1: *dest = _vifRegs->r0; break;
		case 2: 
			switch ( wl ) 
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#else

   switch ( ( _vifRegs->mask >> maskDisplacementX[ cl ] ) & 0x3 )
   {
		case 0: *dest = data;         break;
		case 1: *dest = _vifRegs->r0; break;
		case 2: 
			switch ( wl ) 
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#endif
}

static void _writeY( u32 *dest, u32 data ) 
{
#ifndef VIF_OPTIMIZE

   int n;
	switch ( cl )
   {
		case 0:  n =  2; break;
		case 1:  n = 10; break;
		case 2:  n = 18; break;
		default: n = 26; break;
	}
	switch ( ( _vifRegs->mask >> n ) & 0x3 )
   {
		case 0: *dest = data; break;
		case 1: *dest = _vifRegs->r1; break;
		case 2: 
			switch ( wl )
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#else

   switch ( ( _vifRegs->mask >> maskDisplacementY[ cl ] ) & 0x3 )
   {
		case 0: *dest = data;         break;
		case 1: *dest = _vifRegs->r1; break;
		case 2: 
			switch ( wl ) 
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#endif
}

static void _writeZ( u32 *dest, u32 data )
{
#ifndef VIF_OPTIMIZE

	int n;
	switch ( cl )
   {
		case 0:  n =  4; break;
		case 1:  n = 12; break;
		case 2:  n = 20; break;
		default: n = 28; break;
	}
	switch ( ( _vifRegs->mask >> n ) & 0x3 )
   {
		case 0: *dest = data; break;
		case 1: *dest = _vifRegs->r2; break;
		case 2: 
			switch ( wl )
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#else

   switch ( ( _vifRegs->mask >> maskDisplacementZ[ cl ] ) & 0x3 )
   {
		case 0: *dest = data;         break;
		case 1: *dest = _vifRegs->r2; break;
		case 2: 
			switch ( wl )
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#endif
}

static void _writeW( u32 *dest, u32 data ) 
{
#ifndef VIF_OPTIMIZE

	int n;
	switch ( cl )
   {
		case 0:  n =  6; break;
		case 1:  n = 14; break;
		case 2:  n = 22; break;
		default: n = 30; break;
	}
	switch ( ( _vifRegs->mask >> n ) & 0x3 )
   {
		case 0: *dest = data; break;
		case 1: *dest = _vifRegs->r3; break;
		case 2: 
			switch ( wl )
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#else

   switch ( ( _vifRegs->mask >> maskDisplacementW[ cl ] ) & 0x3 )
   {
		case 0: *dest = data;         break;
		case 1: *dest = _vifRegs->r3; break;
		case 2: 
			switch ( wl ) 
         {
				case 0:  *dest = _vifRegs->c0; break;
				case 1:  *dest = _vifRegs->c1; break;
				case 2:  *dest = _vifRegs->c2; break;
				default: *dest = _vifRegs->c3; break;
			}
			break;
	}

#endif
}

void UNPACK_S_32M(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0]);
	_writeY(dest+1,  data[0]);
	_writeZ(dest+2,  data[0]);
	_writeW(dest+3,  data[0]);

	_writeX(dest+4,  data[1]);
	_writeY(dest+5,  data[1]);
	_writeZ(dest+6,  data[1]);
	_writeW(dest+7,  data[1]);

	_writeX(dest+8,  data[2]);
	_writeY(dest+9,  data[2]);
	_writeZ(dest+10, data[2]);
	_writeW(dest+11, data[2]);
}

void UNPACK_S_32M_O(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0);
	_writeY(dest+1,  data[0] + _vifRegs->r1);
	_writeZ(dest+2,  data[0] + _vifRegs->r2);
	_writeW(dest+3,  data[0] + _vifRegs->r3);

	_writeX(dest+4,  data[1] + _vifRegs->r0);
	_writeY(dest+5,  data[1] + _vifRegs->r1);
	_writeZ(dest+6,  data[1] + _vifRegs->r2);
	_writeW(dest+7,  data[1] + _vifRegs->r3);

	_writeX(dest+8,  data[2] + _vifRegs->r0);
	_writeY(dest+9,  data[2] + _vifRegs->r1);
	_writeZ(dest+10, data[2] + _vifRegs->r2);
	_writeW(dest+11, data[2] + _vifRegs->r3);
}

void UNPACK_S_32M_D(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0*2);
	_writeY(dest+1,  data[0] + _vifRegs->r1*2);
	_writeZ(dest+2,  data[0] + _vifRegs->r2*2);
	_writeW(dest+3,  data[0] + _vifRegs->r3*2);

	_writeX(dest+4,  data[1] + _vifRegs->r0*2);
	_writeY(dest+5,  data[1] + _vifRegs->r1*2);
	_writeZ(dest+6,  data[1] + _vifRegs->r2*2);
	_writeW(dest+7,  data[1] + _vifRegs->r3*2);

	_writeX(dest+8,  data[2] + _vifRegs->r0*2);
	_writeY(dest+9,  data[2] + _vifRegs->r1*2);
	_writeZ(dest+10, data[2] + _vifRegs->r2*2);
	_writeW(dest+11, data[2] + _vifRegs->r3*2);
}

void UNPACK_S_32Mpart(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data);
					_writeY(dest++, *data);
					_writeZ(dest++, *data);
					_writeW(dest++, *data++); );

	_UNPACKpart(1,  _writeX(dest++, *data);
					_writeY(dest++, *data);
					_writeZ(dest++, *data);
					_writeW(dest++, *data++); );

	_UNPACKpart(2,  _writeX(dest++, *data);
					_writeY(dest++, *data);
					_writeZ(dest++, *data);
					_writeW(dest++, *data); );
}

void UNPACK_S_32Mpart_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data + _vifRegs->r0);
					_writeY(dest++, *data + _vifRegs->r1);
					_writeZ(dest++, *data + _vifRegs->r2);
					_writeW(dest++, *data++ + _vifRegs->r3); );

	_UNPACKpart(1,  _writeX(dest++, *data + _vifRegs->r0);
					_writeY(dest++, *data + _vifRegs->r1);
					_writeZ(dest++, *data + _vifRegs->r2);
					_writeW(dest++, *data++ + _vifRegs->r3); );

	_UNPACKpart(2,  _writeX(dest++, *data + _vifRegs->r0);
					_writeY(dest++, *data + _vifRegs->r1);
					_writeZ(dest++, *data + _vifRegs->r2);
					_writeW(dest++, *data + _vifRegs->r3); );
}

void UNPACK_S_32Mpart_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data + _vifRegs->r0*2);
					_writeY(dest++, *data + _vifRegs->r1*2);
					_writeZ(dest++, *data + _vifRegs->r2*2);
					_writeW(dest++, *data++ + _vifRegs->r3*2); );

	_UNPACKpart(1,  _writeX(dest++, *data + _vifRegs->r0*2);
					_writeY(dest++, *data + _vifRegs->r1*2);
					_writeZ(dest++, *data + _vifRegs->r2*2);
					_writeW(dest++, *data++ + _vifRegs->r3*2); );

	_UNPACKpart(2,  _writeX(dest++, *data + _vifRegs->r0*2);
					_writeY(dest++, *data + _vifRegs->r1*2);
					_writeZ(dest++, *data + _vifRegs->r2*2);
					_writeW(dest++, *data + _vifRegs->r3*2); );
}

#define _UNPACK_S_16M(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0]); \
	_writeY(dest+1,  sdata[0]); \
	_writeZ(dest+2,  sdata[0]); \
	_writeW(dest+3,  sdata[0]); \
 \
	_writeX(dest+4,  sdata[1]); \
	_writeY(dest+5,  sdata[1]); \
	_writeZ(dest+6,  sdata[1]); \
	_writeW(dest+7,  sdata[1]); \
 \
	_writeX(dest+8,  sdata[2]); \
	_writeY(dest+9,  sdata[2]); \
	_writeZ(dest+10, sdata[2]); \
	_writeW(dest+11, sdata[2]);

void UNPACK_S_16sM(u32 *dest, u32 *data) {
	_UNPACK_S_16M(s16);
}

void UNPACK_S_16uM(u32 *dest, u32 *data) {
	_UNPACK_S_16M(u16);
}

#define _UNPACK_S_16M_O(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  sdata[0] + _vifRegs->r1); \
	_writeZ(dest+2,  sdata[0] + _vifRegs->r2); \
	_writeW(dest+3,  sdata[0] + _vifRegs->r3); \
 \
	_writeX(dest+4,  sdata[1] + _vifRegs->r0); \
	_writeY(dest+5,  sdata[1] + _vifRegs->r1); \
	_writeZ(dest+6,  sdata[1] + _vifRegs->r2); \
	_writeW(dest+7,  sdata[1] + _vifRegs->r3); \
 \
	_writeX(dest+8,  sdata[2] + _vifRegs->r0); \
	_writeY(dest+9,  sdata[2] + _vifRegs->r1); \
	_writeZ(dest+10, sdata[2] + _vifRegs->r2); \
	_writeW(dest+11, sdata[2] + _vifRegs->r3);

void UNPACK_S_16sM_O(u32 *dest, u32 *data) {
	_UNPACK_S_16M_O(s16);
}

void UNPACK_S_16uM_O(u32 *dest, u32 *data) {
	_UNPACK_S_16M_O(u16);
}

#define _UNPACK_S_16M_D(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  sdata[0] + _vifRegs->r1*2); \
	_writeZ(dest+2,  sdata[0] + _vifRegs->r2*2); \
	_writeW(dest+3,  sdata[0] + _vifRegs->r3*2); \
 \
	_writeX(dest+4,  sdata[1] + _vifRegs->r0*2); \
	_writeY(dest+5,  sdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+6,  sdata[1] + _vifRegs->r2*2); \
	_writeW(dest+7,  sdata[1] + _vifRegs->r3*2); \
 \
	_writeX(dest+8,  sdata[2] + _vifRegs->r0*2); \
	_writeY(dest+9,  sdata[2] + _vifRegs->r1*2); \
	_writeZ(dest+10, sdata[2] + _vifRegs->r2*2); \
	_writeW(dest+11, sdata[2] + _vifRegs->r3*2);

void UNPACK_S_16sM_D(u32 *dest, u32 *data) {
	_UNPACK_S_16M_D(s16);
}

void UNPACK_S_16uM_D(u32 *dest, u32 *data) {
	_UNPACK_S_16M_D(u16);
}

#define _UNPACK_S_16Mpart(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata); \
					_writeY(dest++, *sdata); \
					_writeZ(dest++, *sdata); \
					_writeW(dest++, *sdata++); \
					_writeX(dest++, *sdata); \
					_writeY(dest++, *sdata); \
					_writeZ(dest++, *sdata); \
					_writeW(dest++, *sdata++); ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *sdata); \
					_writeY(dest++, *sdata); \
					_writeZ(dest++, *sdata); \
					_writeW(dest  , *sdata); );

void UNPACK_S_16sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart(s16);
}

void UNPACK_S_16uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart(u16);
}

#define _UNPACK_S_16Mpart_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata + _vifRegs->r0); \
					_writeY(dest++, *sdata + _vifRegs->r1); \
					_writeZ(dest++, *sdata + _vifRegs->r2); \
					_writeW(dest++, *sdata++ + _vifRegs->r3); \
					_writeX(dest++, *sdata + _vifRegs->r0); \
					_writeY(dest++, *sdata + _vifRegs->r1); \
					_writeZ(dest++, *sdata + _vifRegs->r2); \
					_writeW(dest++, *sdata++ + _vifRegs->r3); ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *sdata + _vifRegs->r0); \
					_writeY(dest++, *sdata + _vifRegs->r1); \
					_writeZ(dest++, *sdata + _vifRegs->r2); \
					_writeW(dest  , *sdata + _vifRegs->r3); );

void UNPACK_S_16sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart_O(s16);
}

void UNPACK_S_16uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart_O(u16);
}

#define _UNPACK_S_16Mpart_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata + _vifRegs->r0*2); \
					_writeY(dest++, *sdata + _vifRegs->r1*2); \
					_writeZ(dest++, *sdata + _vifRegs->r2*2); \
					_writeW(dest++, *sdata++ + _vifRegs->r3*2); \
					_writeX(dest++, *sdata + _vifRegs->r0*2); \
					_writeY(dest++, *sdata + _vifRegs->r1*2); \
					_writeZ(dest++, *sdata + _vifRegs->r2*2); \
					_writeW(dest++, *sdata++ + _vifRegs->r3*2); ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *sdata + _vifRegs->r0*2); \
					_writeY(dest++, *sdata + _vifRegs->r1*2); \
					_writeZ(dest++, *sdata + _vifRegs->r2*2); \
					_writeW(dest  , *sdata + _vifRegs->r3*2); );

void UNPACK_S_16sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart_D(s16);
}

void UNPACK_S_16uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_16Mpart_D(u16);
}

#define _UNPACK_S_8M(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0]); \
	_writeY(dest+1,  cdata[0]); \
	_writeZ(dest+2,  cdata[0]); \
	_writeW(dest+3,  cdata[0]); \
 \
	_writeX(dest+4,  cdata[1]); \
	_writeY(dest+5,  cdata[1]); \
	_writeZ(dest+6,  cdata[1]); \
	_writeW(dest+7,  cdata[1]); \
 \
	_writeX(dest+8,  cdata[2]); \
	_writeY(dest+9,  cdata[2]); \
	_writeZ(dest+10, cdata[2]); \
	_writeW(dest+11, cdata[2]);

void UNPACK_S_8sM(u32 *dest, u32 *data) {
	_UNPACK_S_8M(s8);
}

void UNPACK_S_8uM(u32 *dest, u32 *data) {
	_UNPACK_S_8M(u8);
}

void UNPACK_S_8sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M(s8);
}

void UNPACK_S_8uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M(u8);
}

#define _UNPACK_S_8M_O(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  cdata[0] + _vifRegs->r1); \
	_writeZ(dest+2,  cdata[0] + _vifRegs->r2); \
	_writeW(dest+3,  cdata[0] + _vifRegs->r3); \
 \
	_writeX(dest+4,  cdata[1] + _vifRegs->r0); \
	_writeY(dest+5,  cdata[1] + _vifRegs->r1); \
	_writeZ(dest+6,  cdata[1] + _vifRegs->r2); \
	_writeW(dest+7,  cdata[1] + _vifRegs->r3); \
 \
	_writeX(dest+8,  cdata[2] + _vifRegs->r0); \
	_writeY(dest+9,  cdata[2] + _vifRegs->r1); \
	_writeZ(dest+10, cdata[2] + _vifRegs->r2); \
	_writeW(dest+11, cdata[2] + _vifRegs->r3);

void UNPACK_S_8sM_O(u32 *dest, u32 *data) {
	_UNPACK_S_8M_O(s8);
}

void UNPACK_S_8uM_O(u32 *dest, u32 *data) {
	_UNPACK_S_8M_O(u8);
}

void UNPACK_S_8sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M_O(s8);
}

void UNPACK_S_8uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M_O(u8);
}

#define _UNPACK_S_8M_D(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  cdata[0] + _vifRegs->r1*2); \
	_writeZ(dest+2,  cdata[0] + _vifRegs->r2*2); \
	_writeW(dest+3,  cdata[0] + _vifRegs->r3*2); \
 \
	_writeX(dest+4,  cdata[1] + _vifRegs->r0*2); \
	_writeY(dest+5,  cdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+6,  cdata[1] + _vifRegs->r2*2); \
	_writeW(dest+7,  cdata[1] + _vifRegs->r3*2); \
 \
	_writeX(dest+8,  cdata[2] + _vifRegs->r0*2); \
	_writeY(dest+9,  cdata[2] + _vifRegs->r1*2); \
	_writeZ(dest+10, cdata[2] + _vifRegs->r2*2); \
	_writeW(dest+11, cdata[2] + _vifRegs->r3*2);

void UNPACK_S_8sM_D(u32 *dest, u32 *data) {
	_UNPACK_S_8M_D(s8);
}

void UNPACK_S_8uM_D(u32 *dest, u32 *data) {
	_UNPACK_S_8M_D(u8);
}

void UNPACK_S_8sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M_D(s8);
}

void UNPACK_S_8uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_S_8M_D(u8);
}

void UNPACK_V2_32M(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0]);
	_writeY(dest+1,  data[1]);
	_writeZ(dest+2,  0);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[2]);
	_writeY(dest+5,  data[3]);
	_writeZ(dest+6,  0);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[4]);
	_writeY(dest+9,  data[5]);
	_writeZ(dest+10,  0);
	_writeW(dest+11,  0);
}

void UNPACK_V2_32M_O(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0);
	_writeY(dest+1,  data[1] + _vifRegs->r1);
	_writeZ(dest+2,  0);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[2] + _vifRegs->r0);
	_writeY(dest+5,  data[3] + _vifRegs->r1);
	_writeZ(dest+6,  0);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[4] + _vifRegs->r0);
	_writeY(dest+9,  data[5] + _vifRegs->r1);
	_writeZ(dest+10,  0);
	_writeW(dest+11,  0);
}

void UNPACK_V2_32M_D(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0*2);
	_writeY(dest+1,  data[1] + _vifRegs->r1*2);
	_writeZ(dest+2,  0);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[2] + _vifRegs->r0*2);
	_writeY(dest+5,  data[3] + _vifRegs->r1*2);
	_writeZ(dest+6,  0);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[4] + _vifRegs->r0*2);
	_writeY(dest+9,  data[5] + _vifRegs->r1*2);
	_writeZ(dest+10,  0);
	_writeW(dest+11,  0);
}

void UNPACK_V2_32Mpart(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++); );
	_UNPACKpart(1, _writeY(dest++, *data++); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(2, _writeX(dest++, *data++); );
	_UNPACKpart(3, _writeY(dest++, *data++); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(4, _writeX(dest++, *data++); );
	_UNPACKpart(5, _writeY(dest++, *data++); );
	_writeZ(dest++,  0);
	_writeW(dest,  0);
}

void UNPACK_V2_32Mpart_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++ + _vifRegs->r0); );
	_UNPACKpart(1, _writeY(dest++, *data++ + _vifRegs->r1); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(2, _writeX(dest++, *data++ + _vifRegs->r0); );
	_UNPACKpart(3, _writeY(dest++, *data++ + _vifRegs->r1); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(4, _writeX(dest++, *data++ + _vifRegs->r0); );
	_UNPACKpart(5, _writeY(dest++, *data++ + _vifRegs->r1); );
	_writeZ(dest++,  0);
	_writeW(dest,  0);
}

void UNPACK_V2_32Mpart_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++ + _vifRegs->r0*2); );
	_UNPACKpart(1, _writeY(dest++, *data++ + _vifRegs->r1*2); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(2, _writeX(dest++, *data++ + _vifRegs->r0*2); );
	_UNPACKpart(3, _writeY(dest++, *data++ + _vifRegs->r1*2); );
	_writeZ(dest++,  0);
	_writeW(dest++,  0);
	_UNPACKpart(4, _writeX(dest++, *data++ + _vifRegs->r0*2); );
	_UNPACKpart(5, _writeY(dest++, *data++ + _vifRegs->r1*2); );
	_writeZ(dest++,  0);
	_writeW(dest,  0);
}

#define _UNPACK_V2_16M(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0]); \
	_writeY(dest+1,  sdata[1]); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[2]); \
	_writeY(dest+5,  sdata[3]); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[4]); \
	_writeY(dest+9,  sdata[5]); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_16sM(u32 *dest, u32 *data) {
	_UNPACK_V2_16M(s16);
}

void UNPACK_V2_16uM(u32 *dest, u32 *data) {
	_UNPACK_V2_16M(u16);
}

#define _UNPACK_V2_16M_O(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[2] + _vifRegs->r0); \
	_writeY(dest+5,  sdata[3] + _vifRegs->r1); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[4] + _vifRegs->r0); \
	_writeY(dest+9,  sdata[5] + _vifRegs->r1); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_16sM_O(u32 *dest, u32 *data) {
	_UNPACK_V2_16M_O(s16);
}

void UNPACK_V2_16uM_O(u32 *dest, u32 *data) {
	_UNPACK_V2_16M_O(u16);
}

#define _UNPACK_V2_16M_D(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[2] + _vifRegs->r0*2); \
	_writeY(dest+5,  sdata[3] + _vifRegs->r1*2); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[4] + _vifRegs->r0*2); \
	_writeY(dest+9,  sdata[5] + _vifRegs->r1*2); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_16sM_D(u32 *dest, u32 *data) {
	_UNPACK_V2_16M_D(s16);
}

void UNPACK_V2_16uM_D(u32 *dest, u32 *data) {
	_UNPACK_V2_16M_D(u16);
}

#define _UNPACK_V2_16Mpart(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata++); \
					_writeY(dest++, *sdata++); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(1,  _writeX(dest++, *sdata++); \
					_writeY(dest++, *sdata++); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(2,  _writeX(dest++, *sdata++); \
					_writeY(dest++, *sdata++); ); \
	_writeZ(dest++,  0); \
	_writeW(dest,  0); 

void UNPACK_V2_16sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart(s16);
}

void UNPACK_V2_16uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart(u16);
}

#define _UNPACK_V2_16Mpart_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata++ + _vifRegs->r0); \
					_writeY(dest++, *sdata++ + _vifRegs->r1); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(1,  _writeX(dest++, *sdata++ + _vifRegs->r0); \
					_writeY(dest++, *sdata++ + _vifRegs->r1); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(2,  _writeX(dest++, *sdata++ + _vifRegs->r0); \
					_writeY(dest++, *sdata++ + _vifRegs->r1); ); \
	_writeZ(dest++,  0); \
	_writeW(dest,  0); 

void UNPACK_V2_16sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart_O(s16);
}

void UNPACK_V2_16uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart_O(u16);
}

#define _UNPACK_V2_16Mpart_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
					_writeY(dest++, *sdata++ + _vifRegs->r1*2); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(1,  _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
					_writeY(dest++, *sdata++ + _vifRegs->r1*2); ); \
	_writeZ(dest++,  0); \
	_writeW(dest++,  0); \
	_UNPACKpart(2,  _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
					_writeY(dest++, *sdata++ + _vifRegs->r1*2); ); \
	_writeZ(dest++,  0); \
	_writeW(dest,  0); 

void UNPACK_V2_16sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart_D(s16);
}

void UNPACK_V2_16uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_16Mpart_D(u16);
}

#define _UNPACK_V2_8M(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0]); \
	_writeY(dest+1,  cdata[1]); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  cdata[2]); \
	_writeY(dest+5,  cdata[3]); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  cdata[4]); \
	_writeY(dest+9,  cdata[5]); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_8sM(u32 *dest, u32 *data) {
	_UNPACK_V2_8M(s8);
}

void UNPACK_V2_8uM(u32 *dest, u32 *data) {
	_UNPACK_V2_8M(u8);
}

#define _UNPACK_V2_8M_O(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  cdata[2] + _vifRegs->r0); \
	_writeY(dest+5,  cdata[3] + _vifRegs->r1); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  cdata[4] + _vifRegs->r0); \
	_writeY(dest+9,  cdata[5] + _vifRegs->r1); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_8sM_O(u32 *dest, u32 *data) {
	_UNPACK_V2_8M_O(s8);
}

void UNPACK_V2_8uM_O(u32 *dest, u32 *data) {
	_UNPACK_V2_8M_O(u8);
}

#define _UNPACK_V2_8M_D(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  0); \
	_writeW(dest+3,  0); \
	_writeX(dest+4,  cdata[2] + _vifRegs->r0*2); \
	_writeY(dest+5,  cdata[3] + _vifRegs->r1*2); \
	_writeZ(dest+6,  0); \
	_writeW(dest+7,  0); \
	_writeX(dest+8,  cdata[4] + _vifRegs->r0*2); \
	_writeY(dest+9,  cdata[5] + _vifRegs->r1*2); \
	_writeZ(dest+10,  0); \
	_writeW(dest+11,  0);

void UNPACK_V2_8sM_D(u32 *dest, u32 *data) {
	_UNPACK_V2_8M_D(s8);
}

void UNPACK_V2_8uM_D(u32 *dest, u32 *data) {
	_UNPACK_V2_8M_D(u8);
}

#define _UNPACK_V2_8Mpart(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *cdata++); \
					_writeY(dest  , *cdata++); dest+= 3; \
					_writeX(dest++, *cdata++); \
					_writeY(dest  , *cdata++); dest+= 3; ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *cdata++); \
					_writeY(dest  , *cdata  ); );

void UNPACK_V2_8sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart(s8);
}

void UNPACK_V2_8uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart(u8);
}

#define _UNPACK_V2_8Mpart_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *cdata++ + _vifRegs->r0); \
					_writeY(dest  , *cdata++ + _vifRegs->r1); dest+= 3; \
					_writeX(dest++, *cdata++ + _vifRegs->r0); \
					_writeY(dest  , *cdata++ + _vifRegs->r1); dest+= 3; ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *cdata++ + _vifRegs->r0); \
					_writeY(dest  , *cdata   + _vifRegs->r1); );

void UNPACK_V2_8sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart_O(s8);
}

void UNPACK_V2_8uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart_O(u8);
}

#define _UNPACK_V2_8Mpart_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0,  _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
					_writeY(dest  , *cdata++ + _vifRegs->r1*2); dest+= 3; \
					_writeX(dest++, *cdata++ + _vifRegs->r0*2); \
					_writeY(dest  , *cdata++ + _vifRegs->r1*2); dest+= 3; ); \
 \
	_UNPACKpart(1,  _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
					_writeY(dest  , *cdata   + _vifRegs->r1*2); );

void UNPACK_V2_8sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart_D(s8);
}

void UNPACK_V2_8uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V2_8Mpart_D(u8);
}

void UNPACK_V3_32M(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0]);
	_writeY(dest+1,  data[1]);
	_writeZ(dest+2,  data[2]);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[3]);
	_writeY(dest+5,  data[4]);
	_writeZ(dest+6,  data[5]);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[6]);
	_writeY(dest+9,  data[7]);
	_writeZ(dest+10, data[8]);
	_writeW(dest+11, 0);
}

void UNPACK_V3_32M_O(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0);
	_writeY(dest+1,  data[1] + _vifRegs->r1);
	_writeZ(dest+2,  data[2] + _vifRegs->r2);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[3] + _vifRegs->r0);
	_writeY(dest+5,  data[4] + _vifRegs->r1);
	_writeZ(dest+6,  data[5] + _vifRegs->r2);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[6] + _vifRegs->r0);
	_writeY(dest+9,  data[7] + _vifRegs->r1);
	_writeZ(dest+10, data[8] + _vifRegs->r2);
	_writeW(dest+11, 0);
}

void UNPACK_V3_32M_D(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0*2);
	_writeY(dest+1,  data[1] + _vifRegs->r1*2);
	_writeZ(dest+2,  data[2] + _vifRegs->r2*2);
	_writeW(dest+3,  0);
	_writeX(dest+4,  data[3] + _vifRegs->r0*2);
	_writeY(dest+5,  data[4] + _vifRegs->r1*2);
	_writeZ(dest+6,  data[5] + _vifRegs->r2*2);
	_writeW(dest+7,  0);
	_writeX(dest+8,  data[6] + _vifRegs->r0*2);
	_writeY(dest+9,  data[7] + _vifRegs->r1*2);
	_writeZ(dest+10, data[8] + _vifRegs->r2*2);
	_writeW(dest+11, 0);
}

void UNPACK_V3_32Mpart(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++););
	_UNPACKpart(1, _writeY(dest++, *data++););
	_UNPACKpart(2, _writeZ(dest++, *data++););
	_writeW(dest++, 0);
	_UNPACKpart(3, _writeX(dest++, *data++););
	_UNPACKpart(4, _writeY(dest++, *data++););
	_UNPACKpart(5, _writeZ(dest++, *data++););
	_writeW(dest++, 0);
	_UNPACKpart(6, _writeX(dest++, *data++););
	_UNPACKpart(7, _writeY(dest++, *data++););
	_UNPACKpart(8, _writeZ(dest++, *data++););
	_writeW(dest++, 0);
}

void UNPACK_V3_32Mpart_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(1, _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(2, _writeZ(dest++, *data++ + _vifRegs->r2););
	_writeW(dest++, 0);
	_UNPACKpart(3, _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(4, _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(5, _writeZ(dest++, *data++ + _vifRegs->r2););
	_writeW(dest++, 0);
	_UNPACKpart(6, _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(7, _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(8, _writeZ(dest++, *data++ + _vifRegs->r2););
	_writeW(dest++, 0);
}

void UNPACK_V3_32Mpart_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0, _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(1, _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(2, _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_writeW(dest++, 0);
	_UNPACKpart(3, _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(4, _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(5, _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_writeW(dest++, 0);
	_UNPACKpart(6, _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(7, _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(8, _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_writeW(dest++, 0);
}

#define _UNPACK_V3_16M(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0]); \
	_writeY(dest+1,  sdata[1]); \
	_writeZ(dest+2,  sdata[2]); \
    _writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[3]); \
	_writeY(dest+5,  sdata[4]); \
	_writeZ(dest+6,  sdata[5]); \
    _writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[6]); \
	_writeY(dest+9,  sdata[7]); \
	_writeZ(dest+10, sdata[8]); \
	_writeW(dest+11, 0); 

void UNPACK_V3_16sM(u32 *dest, u32 *data) {
	_UNPACK_V3_16M(s16);
}

void UNPACK_V3_16uM(u32 *dest, u32 *data) {
	_UNPACK_V3_16M(u16);
}

#define _UNPACK_V3_16M_O(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  sdata[2] + _vifRegs->r2); \
    _writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[3] + _vifRegs->r0); \
	_writeY(dest+5,  sdata[4] + _vifRegs->r1); \
	_writeZ(dest+6,  sdata[5] + _vifRegs->r2); \
    _writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[6] + _vifRegs->r0); \
	_writeY(dest+9,  sdata[7] + _vifRegs->r1); \
	_writeZ(dest+10, sdata[8] + _vifRegs->r2); \
	_writeW(dest+11, 0); 

void UNPACK_V3_16sM_O(u32 *dest, u32 *data) {
	_UNPACK_V3_16M_O(s16);
}

void UNPACK_V3_16uM_O(u32 *dest, u32 *data) {
	_UNPACK_V3_16M_O(u16);
}

#define _UNPACK_V3_16M_D(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  sdata[2] + _vifRegs->r2*2); \
    _writeW(dest+3,  0); \
	_writeX(dest+4,  sdata[3] + _vifRegs->r0*2); \
	_writeY(dest+5,  sdata[4] + _vifRegs->r1*2); \
	_writeZ(dest+6,  sdata[5] + _vifRegs->r2*2); \
    _writeW(dest+7,  0); \
	_writeX(dest+8,  sdata[6] + _vifRegs->r0*2); \
	_writeY(dest+9,  sdata[7] + _vifRegs->r1*2); \
	_writeZ(dest+10, sdata[8] + _vifRegs->r2*2); \
	_writeW(dest+11, 0); 

void UNPACK_V3_16sM_D(u32 *dest, u32 *data) {
	_UNPACK_V3_16M_D(s16);
}

void UNPACK_V3_16uM_D(u32 *dest, u32 *data) {
	_UNPACK_V3_16M_D(u16);
}

#define _UNPACK_V3_16Mpart(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++); \
				_writeY(dest++, *sdata++);); \
 \
	_UNPACKpart(1, \
                _writeZ(dest  , *sdata++); dest+= 2; \
				_writeX(dest++, *sdata++);); \
 \
	_UNPACKpart(2, \
                _writeY(dest++, *sdata++); \
				_writeZ(dest  , *sdata++); dest+= 2; ); \
 \
	_UNPACKpart(3, \
                _writeX(dest++, *sdata++); \
				_writeY(dest++, *sdata++);); \
 \
	_UNPACKpart(4, \
                _writeZ(dest  , *sdata  ););

void UNPACK_V3_16sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart(s16);
}

void UNPACK_V3_16uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart(u16);
}

#define _UNPACK_V3_16Mpart_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++ + _vifRegs->r0); \
				_writeY(dest++, *sdata++ + _vifRegs->r1);); \
 \
	_UNPACKpart(1, \
                _writeZ(dest  , *sdata++ + _vifRegs->r2); dest+= 2; \
				_writeX(dest++, *sdata++ + _vifRegs->r0);); \
 \
	_UNPACKpart(2, \
                _writeY(dest++, *sdata++ + _vifRegs->r1); \
				_writeZ(dest  , *sdata++ + _vifRegs->r2); dest+= 2; ); \
 \
	_UNPACKpart(3, \
                _writeX(dest++, *sdata++ + _vifRegs->r0); \
				_writeY(dest++, *sdata++ + _vifRegs->r1);); \
 \
	_UNPACKpart(4, \
                _writeZ(dest  , *sdata   + _vifRegs->r2););

void UNPACK_V3_16sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart_O(s16);
}

void UNPACK_V3_16uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart_O(u16);
}

#define _UNPACK_V3_16Mpart_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *sdata++ + _vifRegs->r1*2);); \
 \
	_UNPACKpart(1, \
                _writeZ(dest  , *sdata++ + _vifRegs->r2*2); dest+= 2; \
				_writeX(dest++, *sdata++ + _vifRegs->r0*2);); \
 \
	_UNPACKpart(2, \
                _writeY(dest++, *sdata++ + _vifRegs->r1*2); \
				_writeZ(dest  , *sdata++ + _vifRegs->r2*2); dest+= 2; ); \
 \
	_UNPACKpart(3, \
                _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *sdata++ + _vifRegs->r1*2);); \
 \
	_UNPACKpart(4, \
                _writeZ(dest  , *sdata   + _vifRegs->r2*2););

void UNPACK_V3_16sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart_D(s16);
}

void UNPACK_V3_16uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_16Mpart_D(u16);
}

#define _UNPACK_V3_8M(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0]); \
	_writeY(dest+1,  cdata[1]); \
	_writeZ(dest+2,  cdata[2]); \
    _writeW(dest+3 , 0); \
	_writeX(dest+4,  cdata[3]); \
	_writeY(dest+5,  cdata[4]); \
	_writeZ(dest+6,  cdata[5]); \
    _writeW(dest+7 , 0); \
	_writeX(dest+8,  cdata[6]); \
	_writeY(dest+9,  cdata[7]); \
	_writeZ(dest+10, cdata[8]); \
	_writeW(dest+11, 0); \

void UNPACK_V3_8sM(u32 *dest, u32 *data) {
	_UNPACK_V3_8M(s8);
}

void UNPACK_V3_8uM(u32 *dest, u32 *data) {
	_UNPACK_V3_8M(u8);
}

#define _UNPACK_V3_8M_O(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  cdata[2] + _vifRegs->r2); \
    _writeW(dest+3 , 0); \
	_writeX(dest+4,  cdata[3] + _vifRegs->r0); \
	_writeY(dest+5,  cdata[4] + _vifRegs->r1); \
	_writeZ(dest+6,  cdata[5] + _vifRegs->r2); \
    _writeW(dest+7 , 0); \
	_writeX(dest+8,  cdata[6] + _vifRegs->r0); \
	_writeY(dest+9,  cdata[7] + _vifRegs->r1); \
	_writeZ(dest+10, cdata[8] + _vifRegs->r2); \
	_writeW(dest+11, 0); \

void UNPACK_V3_8sM_O(u32 *dest, u32 *data) {
	_UNPACK_V3_8M_O(s8);
}

void UNPACK_V3_8uM_O(u32 *dest, u32 *data) {
	_UNPACK_V3_8M_O(u8);
}

#define _UNPACK_V3_8M_D(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  cdata[2] + _vifRegs->r2*2); \
    _writeW(dest+3 , 0); \
	_writeX(dest+4,  cdata[3] + _vifRegs->r0*2); \
	_writeY(dest+5,  cdata[4] + _vifRegs->r1*2); \
	_writeZ(dest+6,  cdata[5] + _vifRegs->r2*2); \
    _writeW(dest+7 , 0); \
	_writeX(dest+8,  cdata[6] + _vifRegs->r0*2); \
	_writeY(dest+9,  cdata[7] + _vifRegs->r1*2); \
	_writeZ(dest+10, cdata[8] + _vifRegs->r2*2); \
	_writeW(dest+11, 0); \

void UNPACK_V3_8sM_D(u32 *dest, u32 *data) {
	_UNPACK_V3_8M_D(s8);
}

void UNPACK_V3_8uM_D(u32 *dest, u32 *data) {
	_UNPACK_V3_8M_D(u8);
}

#define _UNPACK_V3_8Mpart(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++); \
				_writeY(dest++, *cdata++); \
				_writeZ(dest  , *cdata++); dest += 2; \
				_writeX(dest++, *cdata++);); \
 \
	_UNPACKpart(1, \
                _writeY(dest++, *cdata++); \
				_writeZ(dest  , *cdata++); dest += 2; \
				_writeX(dest++, *cdata++); \
				_writeY(dest++, *cdata++);); \
 \
	_UNPACKpart(2, \
                _writeZ(dest  , *cdata  ););

void UNPACK_V3_8sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart(s8);
}

void UNPACK_V3_8uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart(u8);
}

#define _UNPACK_V3_8Mpart_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++ + _vifRegs->r0); \
				_writeY(dest++, *cdata++ + _vifRegs->r1); \
				_writeZ(dest  , *cdata++ + _vifRegs->r2); dest += 2; \
				_writeX(dest++, *cdata++ + _vifRegs->r0);); \
 \
	_UNPACKpart(1, \
                _writeY(dest++, *cdata++ + _vifRegs->r1); \
				_writeZ(dest  , *cdata++ + _vifRegs->r2); dest += 2; \
				_writeX(dest++, *cdata++ + _vifRegs->r0); \
				_writeY(dest++, *cdata++ + _vifRegs->r1);); \
 \
	_UNPACKpart(2, \
                _writeZ(dest  , *cdata   + _vifRegs->r2););

void UNPACK_V3_8sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart_O(s8);
}

void UNPACK_V3_8uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart_O(u8);
}

#define _UNPACK_V3_8Mpart_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *cdata++ + _vifRegs->r1*2); \
				_writeZ(dest  , *cdata++ + _vifRegs->r2*2); dest += 2; \
				_writeX(dest++, *cdata++ + _vifRegs->r0*2);); \
 \
	_UNPACKpart(1, \
                _writeY(dest++, *cdata++ + _vifRegs->r1*2); \
				_writeZ(dest  , *cdata++ + _vifRegs->r2*2); dest += 2; \
				_writeX(dest++, *cdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *cdata++ + _vifRegs->r1*2);); \
 \
	_UNPACKpart(2, \
                _writeZ(dest  , *cdata   + _vifRegs->r2*2););

void UNPACK_V3_8sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart_D(s8);
}

void UNPACK_V3_8uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V3_8Mpart_D(u8);
}

void UNPACK_V4_32M(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0]);
	_writeY(dest+1,  data[1]);
	_writeZ(dest+2,  data[2]);
	_writeW(dest+3,  data[3]);

	_writeX(dest+4,  data[4]);
	_writeY(dest+5,  data[5]);
	_writeZ(dest+6,  data[6]);
	_writeW(dest+7,  data[7]);

	_writeX(dest+8,  data[8]);
	_writeY(dest+9,  data[9]);
	_writeZ(dest+10, data[10]);
	_writeW(dest+11, data[11]);
}

void UNPACK_V4_32M_O(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0);
	_writeY(dest+1,  data[1] + _vifRegs->r1);
	_writeZ(dest+2,  data[2] + _vifRegs->r2);
	_writeW(dest+3,  data[3] + _vifRegs->r3);

	_writeX(dest+4,  data[4] + _vifRegs->r0);
	_writeY(dest+5,  data[5] + _vifRegs->r1);
	_writeZ(dest+6,  data[6] + _vifRegs->r2);
	_writeW(dest+7,  data[7] + _vifRegs->r3);

	_writeX(dest+8,  data[8] + _vifRegs->r0);
	_writeY(dest+9,  data[9] + _vifRegs->r1);
	_writeZ(dest+10, data[10] + _vifRegs->r2);
	_writeW(dest+11, data[11] + _vifRegs->r3);
}

void UNPACK_V4_32M_D(u32 *dest, u32 *data) {
	_writeX(dest+0,  data[0] + _vifRegs->r0*2);
	_writeY(dest+1,  data[1] + _vifRegs->r1*2);
	_writeZ(dest+2,  data[2] + _vifRegs->r2*2);
	_writeW(dest+3,  data[3] + _vifRegs->r3*2);

	_writeX(dest+4,  data[4] + _vifRegs->r0*2);
	_writeY(dest+5,  data[5] + _vifRegs->r1*2);
	_writeZ(dest+6,  data[6] + _vifRegs->r2*2);
	_writeW(dest+7,  data[7] + _vifRegs->r3*2);

	_writeX(dest+8,  data[8] + _vifRegs->r0*2);
	_writeY(dest+9,  data[9] + _vifRegs->r1*2);
	_writeZ(dest+10, data[10] + _vifRegs->r2*2);
	_writeW(dest+11, data[11] + _vifRegs->r3*2);
}

void UNPACK_V4_32Mpart(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data++););
	_UNPACKpart(1,  _writeY(dest++, *data++););
	_UNPACKpart(2,  _writeZ(dest++, *data++););
	_UNPACKpart(3,  _writeW(dest++, *data++););

	_UNPACKpart(4,  _writeX(dest++, *data++););
	_UNPACKpart(5,  _writeY(dest++, *data++););
	_UNPACKpart(6,  _writeZ(dest++, *data++););
	_UNPACKpart(7,  _writeW(dest++, *data++););

	_UNPACKpart(8,  _writeX(dest++, *data++););
	_UNPACKpart(9,  _writeY(dest++, *data++););
	_UNPACKpart(10, _writeZ(dest++, *data++););
	_UNPACKpart(11, _writeW(dest  , *data  ););
}

void UNPACK_V4_32Mpart_O(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(1,  _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(2,  _writeZ(dest++, *data++ + _vifRegs->r2););
	_UNPACKpart(3,  _writeW(dest++, *data++ + _vifRegs->r3););

	_UNPACKpart(4,  _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(5,  _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(6,  _writeZ(dest++, *data++ + _vifRegs->r2););
	_UNPACKpart(7,  _writeW(dest++, *data++ + _vifRegs->r3););

	_UNPACKpart(8,  _writeX(dest++, *data++ + _vifRegs->r0););
	_UNPACKpart(9,  _writeY(dest++, *data++ + _vifRegs->r1););
	_UNPACKpart(10, _writeZ(dest++, *data++ + _vifRegs->r2););
	_UNPACKpart(11, _writeW(dest  , *data   + _vifRegs->r3););
}

void UNPACK_V4_32Mpart_D(u32 *dest, u32 *data, int size) {
	_UNPACKpart(0,  _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(1,  _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(2,  _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_UNPACKpart(3,  _writeW(dest++, *data++ + _vifRegs->r3*2););

	_UNPACKpart(4,  _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(5,  _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(6,  _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_UNPACKpart(7,  _writeW(dest++, *data++ + _vifRegs->r3*2););

	_UNPACKpart(8,  _writeX(dest++, *data++ + _vifRegs->r0*2););
	_UNPACKpart(9,  _writeY(dest++, *data++ + _vifRegs->r1*2););
	_UNPACKpart(10, _writeZ(dest++, *data++ + _vifRegs->r2*2););
	_UNPACKpart(11, _writeW(dest  , *data   + _vifRegs->r3*2););
}

#define _UNPACK_V4_16M(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0]); \
	_writeY(dest+1,  sdata[1]); \
	_writeZ(dest+2,  sdata[2]); \
	_writeW(dest+3,  sdata[3]); \
 \
	_writeX(dest+4,  sdata[4]); \
	_writeY(dest+5,  sdata[5]); \
	_writeZ(dest+6,  sdata[6]); \
	_writeW(dest+7,  sdata[7]); \
 \
	_writeX(dest+8,  sdata[8]); \
	_writeY(dest+9,  sdata[9]); \
	_writeZ(dest+10, sdata[10]); \
	_writeW(dest+11, sdata[11]);

void UNPACK_V4_16sM(u32 *dest, u32 *data) {
	_UNPACK_V4_16M(s16);
}

void UNPACK_V4_16uM(u32 *dest, u32 *data) {
	_UNPACK_V4_16M(u16);
}

#define _UNPACK_V4_16M_O(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  sdata[2] + _vifRegs->r2); \
	_writeW(dest+3,  sdata[3] + _vifRegs->r3); \
 \
	_writeX(dest+4,  sdata[4] + _vifRegs->r0); \
	_writeY(dest+5,  sdata[5] + _vifRegs->r1); \
	_writeZ(dest+6,  sdata[6] + _vifRegs->r2); \
	_writeW(dest+7,  sdata[7] + _vifRegs->r3); \
 \
	_writeX(dest+8,  sdata[8] + _vifRegs->r0); \
	_writeY(dest+9,  sdata[9] + _vifRegs->r1); \
	_writeZ(dest+10, sdata[10] + _vifRegs->r2); \
	_writeW(dest+11, sdata[11] + _vifRegs->r3);

void UNPACK_V4_16sM_O(u32 *dest, u32 *data) {
	_UNPACK_V4_16M_O(s16);
}

void UNPACK_V4_16uM_O(u32 *dest, u32 *data) {
	_UNPACK_V4_16M_O(u16);
}

#define _UNPACK_V4_16M_D(format) \
	format *sdata = (format*)data; \
	_writeX(dest+0,  sdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  sdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  sdata[2] + _vifRegs->r2*2); \
	_writeW(dest+3,  sdata[3] + _vifRegs->r3*2); \
 \
	_writeX(dest+4,  sdata[4] + _vifRegs->r0*2); \
	_writeY(dest+5,  sdata[5] + _vifRegs->r1*2); \
	_writeZ(dest+6,  sdata[6] + _vifRegs->r2*2); \
	_writeW(dest+7,  sdata[7] + _vifRegs->r3*2); \
 \
	_writeX(dest+8,  sdata[8] + _vifRegs->r0*2); \
	_writeY(dest+9,  sdata[9] + _vifRegs->r1*2); \
	_writeZ(dest+10, sdata[10] + _vifRegs->r2*2); \
	_writeW(dest+11, sdata[11] + _vifRegs->r3*2);

void UNPACK_V4_16sM_D(u32 *dest, u32 *data) {
	_UNPACK_V4_16M_D(s16);
}

void UNPACK_V4_16uM_D(u32 *dest, u32 *data) {
	_UNPACK_V4_16M_D(u16);
}

#define _UNPACK_V4_16Mpart(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++); \
				_writeY(dest++, *sdata++);); \
	_UNPACKpart(1, \
                _writeZ(dest++, *sdata++); \
				_writeW(dest++, *sdata++);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *sdata++); \
				_writeY(dest++, *sdata++);); \
	_UNPACKpart(3, \
                _writeZ(dest++, *sdata++); \
				_writeW(dest++, *sdata++);); \
 \
	_UNPACKpart(4, \
                _writeX(dest++, *sdata++); \
				_writeY(dest++, *sdata++);); \
	_UNPACKpart(5, \
                _writeZ(dest++, *sdata++); \
				_writeW(dest  , *sdata  ););

void UNPACK_V4_16sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart(s16);
}

void UNPACK_V4_16uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart(u16);
}

#define _UNPACK_V4_16Mpart_O(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++ + _vifRegs->r0); \
				_writeY(dest++, *sdata++ + _vifRegs->r1);); \
	_UNPACKpart(1, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2); \
				_writeW(dest++, *sdata++ + _vifRegs->r3);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *sdata++ + _vifRegs->r0); \
				_writeY(dest++, *sdata++ + _vifRegs->r1);); \
	_UNPACKpart(3, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2); \
				_writeW(dest++, *sdata++ + _vifRegs->r3);); \
 \
	_UNPACKpart(4, \
                _writeX(dest++, *sdata++ + _vifRegs->r0); \
				_writeY(dest++, *sdata++ + _vifRegs->r1);); \
	_UNPACKpart(5, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2); \
				_writeW(dest  , *sdata   + _vifRegs->r3););

void UNPACK_V4_16sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart_O(s16);
}

void UNPACK_V4_16uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart_O(u16);
}

#define _UNPACK_V4_16Mpart_D(format) \
	format *sdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *sdata++ + _vifRegs->r1*2);); \
	_UNPACKpart(1, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2*2); \
				_writeW(dest++, *sdata++ + _vifRegs->r3*2);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *sdata++ + _vifRegs->r1*2);); \
	_UNPACKpart(3, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2*2); \
				_writeW(dest++, *sdata++ + _vifRegs->r3*2);); \
 \
	_UNPACKpart(4, \
                _writeX(dest++, *sdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *sdata++ + _vifRegs->r1*2);); \
	_UNPACKpart(5, \
                _writeZ(dest++, *sdata++ + _vifRegs->r2*2); \
				_writeW(dest  , *sdata   + _vifRegs->r3*2););

void UNPACK_V4_16sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart_D(s16);
}

void UNPACK_V4_16uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_16Mpart_D(u16);
}

#define _UNPACK_V4_8M(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0]); \
	_writeY(dest+1,  cdata[1]); \
	_writeZ(dest+2,  cdata[2]); \
	_writeW(dest+3,  cdata[3]); \
 \
	_writeX(dest+4,  cdata[4]); \
	_writeY(dest+5,  cdata[5]); \
	_writeZ(dest+6,  cdata[6]); \
	_writeW(dest+7,  cdata[7]); \
 \
	_writeX(dest+8,  cdata[8]); \
	_writeY(dest+9,  cdata[9]); \
	_writeZ(dest+10, cdata[10]); \
	_writeW(dest+11, cdata[11]);

void UNPACK_V4_8sM(u32 *dest, u32 *data) {
	_UNPACK_V4_8M(s8);
}

void UNPACK_V4_8uM(u32 *dest, u32 *data) {
	_UNPACK_V4_8M(u8);
}

#define _UNPACK_V4_8M_O(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1); \
	_writeZ(dest+2,  cdata[2] + _vifRegs->r2); \
	_writeW(dest+3,  cdata[3] + _vifRegs->r3); \
 \
	_writeX(dest+4,  cdata[4] + _vifRegs->r0); \
	_writeY(dest+5,  cdata[5] + _vifRegs->r1); \
	_writeZ(dest+6,  cdata[6] + _vifRegs->r2); \
	_writeW(dest+7,  cdata[7] + _vifRegs->r3); \
 \
	_writeX(dest+8,  cdata[8] + _vifRegs->r0); \
	_writeY(dest+9,  cdata[9] + _vifRegs->r1); \
	_writeZ(dest+10, cdata[10] + _vifRegs->r2); \
	_writeW(dest+11, cdata[11] + _vifRegs->r3);

void UNPACK_V4_8sM_O(u32 *dest, u32 *data) {
	_UNPACK_V4_8M_O(s8);
}

void UNPACK_V4_8uM_O(u32 *dest, u32 *data) {
	_UNPACK_V4_8M_O(u8);
}

#define _UNPACK_V4_8M_D(format) \
	format *cdata = (format*)data; \
	_writeX(dest+0,  cdata[0] + _vifRegs->r0*2); \
	_writeY(dest+1,  cdata[1] + _vifRegs->r1*2); \
	_writeZ(dest+2,  cdata[2] + _vifRegs->r2*2); \
	_writeW(dest+3,  cdata[3] + _vifRegs->r3*2); \
 \
	_writeX(dest+4,  cdata[4] + _vifRegs->r0*2); \
	_writeY(dest+5,  cdata[5] + _vifRegs->r1*2); \
	_writeZ(dest+6,  cdata[6] + _vifRegs->r2*2); \
	_writeW(dest+7,  cdata[7] + _vifRegs->r3*2); \
 \
	_writeX(dest+8,  cdata[8] + _vifRegs->r0*2); \
	_writeY(dest+9,  cdata[9] + _vifRegs->r1*2); \
	_writeZ(dest+10, cdata[10] + _vifRegs->r2*2); \
	_writeW(dest+11, cdata[11] + _vifRegs->r3*2);

void UNPACK_V4_8sM_D(u32 *dest, u32 *data) {
	_UNPACK_V4_8M_D(s8);
}

void UNPACK_V4_8uM_D(u32 *dest, u32 *data) {
	_UNPACK_V4_8M_D(u8);
}

#define _UNPACK_V4_8Mpart(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++); \
				_writeY(dest++, *cdata++); \
				_writeZ(dest++, *cdata++); \
				_writeW(dest++, *cdata++);); \
 \
	_UNPACKpart(1, \
                _writeX(dest++, *cdata++); \
				_writeY(dest++, *cdata++); \
				_writeZ(dest++, *cdata++); \
				_writeW(dest++, *cdata++);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *cdata++); \
				_writeY(dest++, *cdata++); \
				_writeZ(dest++, *cdata++); \
				_writeW(dest  , *cdata  ););

void UNPACK_V4_8sMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart(s8);
}

void UNPACK_V4_8uMpart(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart(u8);
}

#define _UNPACK_V4_8Mpart_O(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++ + _vifRegs->r0); \
				_writeY(dest++, *cdata++ + _vifRegs->r1); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2); \
				_writeW(dest++, *cdata++ + _vifRegs->r3);); \
 \
	_UNPACKpart(1, \
                _writeX(dest++, *cdata++ + _vifRegs->r0); \
				_writeY(dest++, *cdata++ + _vifRegs->r1); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2); \
				_writeW(dest++, *cdata++ + _vifRegs->r3);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *cdata++ + _vifRegs->r0); \
				_writeY(dest++, *cdata++ + _vifRegs->r1); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2); \
				_writeW(dest  , *cdata   + _vifRegs->r3););

void UNPACK_V4_8sMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart_O(s8);
}

void UNPACK_V4_8uMpart_O(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart_O(u8);
}

#define _UNPACK_V4_8Mpart_D(format) \
	format *cdata = (format*)data; \
	_UNPACKpart(0, \
                _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *cdata++ + _vifRegs->r1*2); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2*2); \
				_writeW(dest++, *cdata++ + _vifRegs->r3*2);); \
 \
	_UNPACKpart(1, \
                _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *cdata++ + _vifRegs->r1*2); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2*2); \
				_writeW(dest++, *cdata++ + _vifRegs->r3*2);); \
 \
	_UNPACKpart(2, \
                _writeX(dest++, *cdata++ + _vifRegs->r0*2); \
				_writeY(dest++, *cdata++ + _vifRegs->r1*2); \
				_writeZ(dest++, *cdata++ + _vifRegs->r2*2); \
				_writeW(dest  , *cdata   + _vifRegs->r3*2););

void UNPACK_V4_8sMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart_D(s8);
}

void UNPACK_V4_8uMpart_D(u32 *dest, u32 *data, int size) {
	_UNPACK_V4_8Mpart_D(u8);
}

void UNPACK_V4_5M(u32 *dest, u32 *data) {
	u32 rgba;

	SysPrintf("*PCSX2* UNPACK_V4_5M\n");
	rgba = data[0] & 0xffff; /* RGBA1 */
	dest[0] = (rgba & 0x1f) << 3;
	dest[1] = ((rgba >>  5) & 0x1f) << 3;
	dest[2] = ((rgba >> 10) & 0x1f) << 3;
	dest[3] = ((rgba >> 15) & 0x1) << 7;

	rgba = (data[0] >> 16); /* RGBA2 */
	dest[4] = (rgba & 0x1f) << 3;
	dest[5] = ((rgba >>  5) & 0x1f) << 3;
	dest[6] = ((rgba >> 10) & 0x1f) << 3;
	dest[7] = ((rgba >> 15) & 0x1) << 7;

	rgba = data[1] & 0xffff; /* RGBA3 */
	dest[8]  = (rgba & 0x1f) << 3;
	dest[9]  = ((rgba >>  5) & 0x1f) << 3;
	dest[10] = ((rgba >> 10) & 0x1f) << 3;
	dest[11] = ((rgba >> 15) & 0x1) << 7;
}

void UNPACK_V4_5Mpart(u32 *dest, u32 *data, int size) {
	u32 rgba;

	SysPrintf("*PCSX2* UNPACK_V4_5Mpart\n");
	_UNPACKpart(0, 
                rgba = *data & 0xffff; /* RGBA1 */
                dest[0] = (rgba & 0x1f) << 3;
                dest[1] = ((rgba >>  5) & 0x1f) << 3;
                dest[2] = ((rgba >> 10) & 0x1f) << 3;
                dest[3] = ((rgba >> 15) & 0x1) << 7;

                rgba = (*data++ >> 16); /* RGBA2 */
                dest[4] = (rgba & 0x1f) << 3;
                dest[5] = ((rgba >>  5) & 0x1f) << 3;
                dest[6] = ((rgba >> 10) & 0x1f) << 3;
                dest[7] = ((rgba >> 15) & 0x1) << 7;
                dest += 8; );

	_UNPACKpart(1, 
                rgba = *data & 0xffff; /* RGBA3 */
                dest[8]  = (rgba & 0x1f) << 3;
                dest[9]  = ((rgba >>  5) & 0x1f) << 3;
                dest[10] = ((rgba >> 10) & 0x1f) << 3;
                dest[11] = ((rgba >> 15) & 0x1) << 7; );
}


__inline void VIF1chain(DMACh * dma) {
	if (dma->qwc) { 
		u32 * pMem;
		_dmaGetAddr(pMem, dma->madr, 8); 
		VIF1transfer(pMem, dma->qwc*4); 
	}
}

void MFIFO_VIF1transfer(DMACh *dma, u32 sadr, int qwc) {
	u32 *ptag;
	int id;

#ifdef VIF_LOG
	VIF_LOG("MFIFO_VIF1transfer\n");
#endif

	while (qwc > 0) {
		ptag = (u32*)&psS[sadr & 0x3ff0]; qwc--;
	
		id        = (ptag[0] >> 28) & 0x7;
		dma->qwc  = (ptag[0] & 0xffff);
		dma->madr = ptag[1];

		if (dma->chcr & 0x40) VIF1transfer(ptag+2, 2);

		dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

#ifdef VIF_LOG
		VIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], dma->qwc, id, dma->madr);
#endif
		
		switch (id) {
			case 0: // refe
				VIF1chain(dma);
				dma->madr += dma->qwc * 16;
				dma->qwc = 0;
				qwc = 0;
				break;

			case 1: // cnt
				dma->madr = sadr + 16;
				VIF1chain(dma);
				// Set the taddr to the next tag
				sadr += 16 + (dma->qwc * 16);
				qwc-= dma->qwc;
				break;

			case 3: // ref
			case 4: // refs
				VIF1chain(dma);
				sadr += 16;
				qwc-= dma->qwc;
				break;

			case 7: // end
				dma->madr = sadr + 16;
				VIF1chain(dma);
				dma->madr += dma->qwc * 16;
				dma->qwc = 0;
				qwc = 0;
				break;
		}
	}
}

